/*
 * Copyright (c) 1999, 2000, 2001 The University of Utah and the Flux Group.
 * All rights reserved.
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */

/*
 * EE to test stdin reading of ANEP packets
 */

#define RUNFOR	5
#define NBUFS	100

#include <stdio.h>
#include <string.h>
#include <assert.h>

#include <nodeos/an_impl.h>
#include <nodeos/an_flow.h>
#include <nodeos/an_thread.h>
#include <nodeos/an_chan.h>
#include <nodeos/an_pbuf.h>
#include <nodeos/an_event.h>
#include <nodeos/lib/utils.h>

#include <netinet/in.h>	/* XXX ntohs */

static void ee_init(void *arg);
an_flowinit_func_t InitEE = { ee_init };

static void setup_channels(void);
static void ee_aneppacket(an_pbuf_t *pbuf, void *arg, int msize);
static void ee_die(void *event, void *arg);
static void anep_dump(void *data, int size);

static an_chan_t *anepchan, anepchan_data;
static an_pbuf_t *anepbuf[NBUFS], anepbuf_data[NBUFS];

static struct {
	an_chanaddr_t addr;
	an_chanproto_t proto;
	an_cpuspec_t cpu;
} attr;
static an_flow_t *flow;

int
ee_main(int argc, const char **argv)
{
	an_error_t err;

	flow = an_flow_current();
	printf("%p/%p: Dumping ANEP packets from stdin...\n",
	       flow, an_thread_current());

	/*
	 * Setup channels
	 */
	setup_channels();

	/*
	 * Setup an event to time us out in RUNFOR seconds
	 */
	err = an_event_schedule(ee_die, 0, RUNFOR * 1000000);
	assert(err == 0);
	printf("Timer started, running for %d seconds\n", RUNFOR);

	/*
	 * No longer need this thread
	 */
	return 0;
}

static void
setup_channels(void)
{
	int i;

	memset(&attr, 0, sizeof attr);

	/*
	 * Create packet buffers
	 */
	for (i = 0; i < NBUFS; i++) {
		anepbuf[i] =
			an_pbuf_create(&anepbuf_data[i], AN_PBUF_STDIO, 0, 0);
		assert(anepbuf[i] != 0);
	}

	/*
	 * and channel
	 */
	strcpy(attr.proto.data, "stdio:anep");
	attr.proto.len = strlen(attr.proto.data);
	/* XXX no addr info */
	attr.cpu.ttype = AN_TH_IMPLICIT;
	attr.cpu.threads.im.maxthreads = 1;
	anepchan = an_inchan_create(&anepchan_data, flow, 0,
				    &attr.proto, &attr.addr, anepbuf, NBUFS,
				    &attr.cpu, ee_aneppacket, 0);
	assert(anepchan != 0);
}

static void
ee_aneppacket(an_pbuf_t *pbuf, void *arg, int msize)
{
	anep_dump(an_pbuf_data(pbuf), an_pbuf_size(pbuf));
	an_inchan_addpbuf(anepchan, pbuf);
}

static void
ee_die(void *event, void *arg)
{
	printf("All done!\n");

	moab_chan_printstats(anepchan, 1, "Request", RUNFOR, 0);

	if (anepchan)
		an_inchan_destroy(anepchan);
	an_flow_destroy(an_flow_current());
}

static void
ee_init(void *arg)
{
	int status;
	int argc = 1;
	const char *argv[] = { "ChantestEE" }; 

	status = ee_main(argc, argv);
}

static void
anep_dump(void *data, int size)
{
	struct anephdr {
		char ah_version;
		char ah_flags;           
		short ah_typeid;
		short ah_hlen;
		short ah_plen;
	} *ah = data;
	static int pnum;

	if (size < sizeof *ah) {
		printf("*** ANEP packet too small (%d bytes), ignored\n", size);
		return;
	}

	printf("%d: version=%d, flags=0x%x, type=%d, hlen=%d, plen=%d\n",
	       ++pnum, ah->ah_version, ah->ah_flags,
	       ntohs(ah->ah_typeid), ntohs(ah->ah_hlen), ntohs(ah->ah_plen));
	if (ah->ah_version != 1)
		printf("  *** bad version number (%d), should be 1\n",
		       ah->ah_version);
}
