/*
 * Copyright (c) 1999, 2000 The University of Utah and the Flux Group.
 * All rights reserved.
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */

/*
 * EE to test CPU resource controls
 */

#include <stdio.h>
#include <assert.h>
#include <string.h>
#include <pthread.h>

#include <nodeos/an_impl.h>
#include <nodeos/an_cred.h>
#include <nodeos/an_flow.h>
#include <nodeos/an_thread.h>
#include <nodeos/an_mutex.h>
#include <nodeos/an_cond.h>
#include <nodeos/an_event.h>
#include <nodeos/an_info.h>
#include <nodeos/lib/utils.h>

#define RUNFOR		10

#define NUMFLOWS	3
#define F1THREADS	1
#define F2THREADS	2
#define F3THREADS	4
#define NUMTHREADS	7

static void ee_init(void *arg);
static void ee_death(an_flow_t *flow);
an_flowinit_func_t InitEE = { ee_init };

static void ee_main(void *arg);
static void ee_timeout(void *event, void *arg);
static void ee_thread(void *arg);
static void ee_startflow(int fix, int numthreads, int cps);

static an_mutex_t *lock, lock_data;
static an_cond_t *forever, forever_data;

static an_mutex_t *startlock, startlock_data;
static an_cond_t *startcond, startcond_data;
static an_mutex_t *golock, golock_data;
static an_cond_t *gocond, gocond_data;
static int threadscreated, threadsrunning;
static volatile int threadscounting;
static an_flow_t *myflows[NUMFLOWS];
static an_flow_t flow_data[NUMFLOWS];
static an_thread_t *mythreads[NUMTHREADS];
static an_thread_t thread_data[NUMTHREADS];
static unsigned int mycount[NUMTHREADS];

static unsigned int availcps;

void
ee_main(void *arg)
{
	an_error_t err;
	unsigned int cps;

	printf("CPU test EE started\n");

	lock = an_mutex_create(&lock_data);
	forever = an_cond_create(&forever_data);

	startlock = an_mutex_create(&startlock_data);
	startcond = an_cond_create(&startcond_data);
	golock = an_mutex_create(&golock_data);
	gocond = an_cond_create(&gocond_data);
	
	an_mutex_lock(golock);

	cps = availcps / 2;
	printf("Creating high priority flow #%d (%d cps)\n", 0, cps);
	ee_startflow(0, F1THREADS, cps);

	cps /= 2;
	printf("Creating medium priority flow #%d (%d cps)\n", 1, cps);
	ee_startflow(1, F2THREADS, cps);

	cps /= 2;
	printf("Creating low priority flow #%d (%d cps)\n", 2, cps);
	ee_startflow(2, F3THREADS, cps);

	/*
	 * Wait for all the threads to fire up
	 */
	an_mutex_lock(startlock);
	while (threadsrunning != NUMTHREADS)
		an_cond_wait(startcond, startlock);
	an_mutex_unlock(startlock);

	/*
	 * Now launch them and wait awhile
	 */
	printf("Running %d threads for %d seconds...\n",
	       threadsrunning, RUNFOR);
	threadscounting = 1;
	an_cond_broadcast(gocond);
	an_mutex_unlock(golock);
	err = an_event_schedule(ee_timeout, 0, RUNFOR * 1000000);
	assert(err == 0);
}

static void
ee_timeout(void *event, void *arg)
{
	unsigned long long total;
	int i;

	/*
	 * Stop them and wait for them to all exit
	 */
	printf("Times up!  shutting down %d threads\n", threadsrunning);
	threadscounting = 0;
	an_mutex_lock(startlock);
	while (threadsrunning != 0)
		an_cond_wait(startcond, startlock);
	an_mutex_unlock(startlock);

	/*
	 * Finally, announce the results
	 */
	total = 0;
	for (i = 0; i < NUMTHREADS; i++)
		total += mycount[i];
	for (i = 0; i < NUMTHREADS; i++)
		printf("thread %d (pri: %dx flow): counted to %d (%d%%)\n", i,
		       (i < F1THREADS) ? 4 : (i < F2THREADS+F1THREADS) ? 2 : 1,
		       mycount[i],
		       (int)((unsigned long long)mycount[i] * 100 / total));

	for (i = 0; i < NUMFLOWS; i++) {
		char buf[16];
		snprintf(buf, 16, "Flow%d", i);
		moab_flow_printstats(myflows[i], 0, buf, RUNFOR, 1);
	}
	moab_flow_printstats(0, 0, "Total", RUNFOR, 0);

	an_flow_destroy(an_flow_current());
}

static void
ee_startflow(int fix, int numthreads, int cps)
{
	an_cpuspec_t cs;
	an_memspec_t ms;
	an_error_t err;
	int tix, i;
	void *stackptr;
	size_t stacksize;

	/*
	 * Create some threads for the flow
	 */
	tix = threadscreated;
	threadscreated += numthreads;
	stackptr = 0;
	stacksize = 20 * 1024;
	for (i = 0; i < numthreads; i++) {
		mythreads[tix+i] = an_thread_create(&thread_data[tix+i],
						    stackptr, stacksize);
		assert(mythreads[tix+i] != 0);
	}

	/*
	 * Now setup the resources
	 */
	memset(&cs, 0, sizeof cs);
	cs.ttype = AN_TH_EXPLICIT;
	cs.threads.ex.nthreads = numthreads;
	cs.threads.ex.threads = &mythreads[tix];
#if 0
	cs.rate = 10000;
#else
	cs.rate = cps;
#endif
	memset(&ms, 0, sizeof ms);
	ms.nchunks = numthreads * (128*1024) / NODEOS_MEMPOOL_CHUNKSIZE;
	ms.pool = an_mempool_current();

	/*
	 * Create the flow
	 */
	myflows[fix] = an_flow_create(&flow_data[fix], an_cred_current(),
				      &cs, &ms, 0, 0, 0);
	assert(myflows[fix] == &flow_data[fix]);

	/*
	 * Fire up the threads
	 */
	for (i = 0; i < numthreads; i++) {
		err = an_flow_startthread(myflows[fix],
					  ee_thread, (void *)(tix+i), 0);
		assert(err == 0);
	}
}

static void
ee_thread(void *arg)
{
	int tid = (int)(arg);

	an_mutex_lock(startlock);
	threadsrunning++;
#ifdef VERBOSE
	printf("%p/%p: thread %d running\n",
	       an_flow_current(), an_thread_current(), tid);
#endif
	an_cond_signal(startcond);
	an_mutex_unlock(startlock);

	an_mutex_lock(golock);
	while (threadscounting == 0)
		an_cond_wait(gocond, golock);
	an_mutex_unlock(golock);

	while (threadscounting != 0)
		mycount[tid]++;

	an_mutex_lock(startlock);
#ifdef VERBOSE
	printf("%p/%p: thread %d exiting\n",
	       an_flow_current(), an_thread_current(), tid);
#endif
	threadsrunning--;
	an_cond_signal(startcond);
	an_mutex_unlock(startlock);
}

static void
ee_init(void *arg)
{
	an_cpuspec_t cs;
	an_memspec_t ms;
	an_flow_t *flow;
	an_infocpu_t info;

	an_info_getcpu(&info);
	availcps = info.cps;

	/*
	 * Fire up a really high priority flow to do all the real work.
	 * Otherwise we will likely never get to run again after we start
	 * the test threads as they will probably have higher priority
	 * than the default flow we were started in.
	 */
	memset(&cs, 0, sizeof cs);
	cs.ttype = AN_TH_IMPLICIT;
	cs.threads.im.maxthreads = 1;
	cs.threads.im.stacksize = 20 * 1024;
	cs.rate = availcps / 2;
	availcps /= 2;
	memset(&ms, 0, sizeof ms);
	ms.nchunks = (128*1024) / NODEOS_MEMPOOL_CHUNKSIZE;
	ms.pool = an_mempool_current();
	flow = an_flow_create(0, an_cred_current(), &cs, &ms,
			      ee_main, 0, ee_death);
	assert(flow != 0);
}

static void
ee_death(an_flow_t *flow)
{
	an_flow_destroy(an_flow_current());
}
