/*
 * Copyright (c) 2000, 2001 The University of Utah and the Flux Group.
 * All rights reserved.
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */

/*
 * Control the demo. Very simple.
 */
#undef  DEBUG

#include <stdio.h>
#include <assert.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <sys/types.h>

#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netinet/in_systm.h>
#include <netinet/if_ether.h>
#include <netinet/ip.h>
#include <netinet/udp.h>
#include "demo.h"

#define DSTIP		"155.99.214.138"
#define CTRLPORT	9999

static struct in_addr		dstip;
static int			ctrlport = CTRLPORT;
static int			sock;
static struct sockaddr_in	dstsockaddr;
static int			send_command(void *buf, int bufsize);
static void			fatal(const char *fmt, ...);

#ifdef DEBUG
#define DPRINTF(fmt, args... )	printf(__FUNCTION__  ":" fmt , ## args)
#else
#define DPRINTF(fmt, args... )
#endif
#define FATAL(fmt, args... )	fatal(__FUNCTION__  ":" fmt , ## args)

/*
 * An array of commands.
 */
static void	dofoo(int argc, char **argv);
static void	donethog(int argc, char **argv);
static void	docpuhog(int argc, char **argv);
static void	domemhog(int argc, char **argv);
static void	dopq(int argc, char **argv);
static void	docpusched(int argc, char **argv);
static void	domemlimit(int argc, char **argv);
static void	dostats(int argc, char **argv);
static void	doverbose(int argc, char **argv);
static void	doexit(int argc, char **argv);

struct command {
	char	*name;
	void   (*func)(int argc, char **argv);
	char    *usage;
} command_array[] = {
	{ DEMOCTRL_FOO,    dofoo, NULL },
	{ DEMOCTRL_NETHOG, donethog, NULL },
	{ DEMOCTRL_CPUHOG, docpuhog, NULL },
	{ DEMOCTRL_MEMHOG, domemhog, NULL },
	{ DEMOCTRL_PQ,     dopq, NULL  },
	{ DEMOCTRL_CPUSCHED, docpusched, NULL },
	{ DEMOCTRL_MEMLIMIT, domemlimit, NULL },
	{ DEMOCTRL_STATS, dostats, NULL },
	{ DEMOCTRL_VERBOSE, doverbose, NULL },
	{ DEMOCTRL_EXIT,   doexit, NULL },
};
static int numcommands = sizeof(command_array)/sizeof(struct command);

int
main(int argc, char **argv)
{
	char		*bp;
	int		i, data;
 
	if ((bp = getenv("DSTIP")) == NULL)
		bp = DSTIP;
	inet_aton(bp, &dstip);
	if ((bp = getenv("CTRLPORT")) != NULL)
		ctrlport = atoi(bp);

	if (argc < 2) {
		printf("%s <command> [<arg0> ... <argn>]\n", argv[0]);
		printf("\tavailable commands: ");
		for (i = 0;
		     i < sizeof(command_array) / sizeof(command_array[0]);
		     i++) {
			printf(" `%s'", command_array[i].name);
		}
		printf("\n");
		printf("\tsends control messages to: %s/%d\n",
		       bp, CTRLPORT);
		exit(1);
	}

	/* Create socket on which to send. */
	sock = socket(AF_INET, SOCK_DGRAM, 0);
	if (sock < 0) {
		perror("opening datagram socket");
		exit(1);
	}

	data = 0x20000;
	if (setsockopt(sock, SOL_SOCKET, SO_SNDBUF, &data, sizeof(data)) < 0) {
		perror("setsockopt");
		exit(1);
	}

	/*
	 * Construct name, with no wildcards, of the socket to send to.
	 */
	dstsockaddr.sin_addr.s_addr = dstip.s_addr;
	dstsockaddr.sin_family = AF_INET;
	dstsockaddr.sin_port = htons(ctrlport);

	/*
	 * Figure out what command was given.
	 */
	for (i = 0; i < numcommands; i++) {
		if (strcmp(argv[1], command_array[i].name) == 0) {
			command_array[i].func(argc-2, argv+2);
			exit(0);
		}
	}
	fprintf(stderr, "No such command: %s\n", argv[1]);
	exit(1);
}

static void
dofoo(int argc, char **argv)
{
	ctrl_packet_t	packet;
	
	if (argc != 1)
		FATAL("Wrong number of arguments");
	
	DPRINTF("argc:%d arg0:%s\n", argc, argv[0]);

	packet.opcode    = htonl(CTRLPKTOP_FOO);
	packet.subopcode = htonl(0);
	packet.data[0]   = atoi(argv[0]);

	send_command((void *) &packet, sizeof(packet));
}

/*
 * Control the NETHOG stream.
 */
static void
donethog(int argc, char **argv)
{
	ctrl_packet_t	packet;
	
	if (argc < 1 || argc > 2)
		FATAL("usage: nethog [on|off|bandw <int>]");
	
	DPRINTF("argc:%d arg0:%s\n", argc, argv[0]);

	packet.opcode = htonl(CTRLPKTOP_NETHOG);

	if (strcmp(argv[0], "on") == 0) {
		packet.subopcode = htonl(NETHOGCTRL_ON);
		packet.data[0]   = 1;
	}
	else if (strcmp(argv[0], "off") == 0) {
		packet.subopcode = htonl(NETHOGCTRL_OFF);
		packet.data[0]   = 0;
	}
	else if (strcmp(argv[0], "bandw") == 0) {
		packet.subopcode = htonl(NETHOGCTRL_BANDW);
		packet.data[0]   = atoi(argv[1]);
	}
	else {
		FATAL("Invalid command: %s", argv[0]);
	}

	send_command((void *) &packet, sizeof(packet));
}

/*
 * Control the CPUHOG stream.
 */
static void
docpuhog(int argc, char **argv)
{
	ctrl_packet_t	packet;
	
	if (argc < 1 || argc > 2)
		FATAL("usage: cpuhog [on|off|rate <int>]");
	
	DPRINTF("argc:%d arg0:%s\n", argc, argv[0]);

	packet.opcode = htonl(CTRLPKTOP_CPUHOG);

	if (strcmp(argv[0], "on") == 0) {
		packet.subopcode = htonl(CPUHOGCTRL_ON);
		packet.data[0]   = 1;
	}
	else if (strcmp(argv[0], "off") == 0) {
		packet.subopcode = htonl(CPUHOGCTRL_OFF);
		packet.data[0]   = 0;
	}
	else if (strcmp(argv[0], "rate") == 0) {
		packet.subopcode = htonl(CPUHOGCTRL_RATE);
		packet.data[0]   = atoi(argv[1]);
	}
	else {
		FATAL("Invalid command: %s", argv[0]);
	}

	send_command((void *) &packet, sizeof(packet));
}

static void
dogenericcontrol(int argc, char** argv,
		 const char* command,
		 int pktop,
		 int onval,
		 int offval)
{
	ctrl_packet_t	packet;
	
	if (argc != 1)
		FATAL("usage: %s [on|off]", command);
	
	DPRINTF("argc:%d arg0:%s\n", argc, argv[0]);

	packet.opcode = htonl(pktop);

	if (strcmp(argv[0], "on") == 0) {
		packet.subopcode = htonl(onval);
		packet.data[0]   = 1;
	}
	else if (strcmp(argv[0], "off") == 0) {
		packet.subopcode = htonl(offval);
		packet.data[0]   = 0;
	}
	else {
		FATAL("Invalid command: %s", argv[0]);
	}

	send_command((void *) &packet, sizeof(packet));
}


/*
 * Control the MEMHOG domain.
 */
static void
domemhog(int argc, char **argv)
{
	dogenericcontrol(argc, argv,
			 DEMOCTRL_MEMHOG,
			 CTRLPKTOP_MEMHOG,
			 MEMHOGCTRL_ON,
			 MEMHOGCTRL_OFF);
}

/*
 * Control the CPU scheduler.
 */
static void
docpusched(int argc, char **argv)
{
	dogenericcontrol(argc, argv,
			 DEMOCTRL_CPUSCHED,
			 CTRLPKTOP_CPUSCHED,
			 CPUSCHEDCTRL_ON,
			 CPUSCHEDCTRL_OFF);
}

/*
 * Control enforcement of memory limits
 */
static void
domemlimit(int argc, char **argv)
{
	dogenericcontrol(argc, argv,
			 DEMOCTRL_MEMLIMIT,
			 CTRLPKTOP_MEMLIMIT,
			 MEMLIMITCTRL_ON,
			 MEMLIMITCTRL_OFF);
}

/*
 * Control the Packet Scheduler
 */
static void
dopq(int argc, char **argv)
{
	dogenericcontrol(argc, argv,
			 DEMOCTRL_PQ,
			 CTRLPKTOP_PQ,
			 PQCTRL_ON,
			 PQCTRL_OFF);
}

/*
 * Control the verbosity
 */
static void
doverbose(int argc, char **argv)
{
	dogenericcontrol(argc, argv,
			 DEMOCTRL_VERBOSE,
			 CTRLPKTOP_VERBOSE,
			 VERBOSECTRL_ON,
			 VERBOSECTRL_OFF);
}

/*
 * Print stats
 */
static void
dostats(int argc, char **argv)
{
	ctrl_packet_t	packet;
	
	if (argc != 1)
		FATAL("usage: stats [total|delta]");
	
	DPRINTF("argc:%d arg0:%s\n", argc, argv[0]);

	packet.opcode = htonl(CTRLPKTOP_STATS);

	if (strcmp(argv[0], "total") == 0) {
		packet.subopcode = htonl(STATSCTRL_TOTAL);
		packet.data[0]   = 1;
	}
	else if (strcmp(argv[0], "delta") == 0) {
		packet.subopcode = htonl(STATSCTRL_DELTA);
		packet.data[0]   = 0;
	}
	else {
		FATAL("Invalid command: %s", argv[0]);
	}

	send_command((void *) &packet, sizeof(packet));
}

/*
 * Exit
 */
static void
doexit(int argc, char **argv)
{
	ctrl_packet_t	packet;
	
	if (argc != 0)
		FATAL("usage: exit");
	
	DPRINTF("argc:%d arg0:%s\n", argc, argv[0]);

	packet.opcode = htonl(CTRLPKTOP_EXIT);
	packet.subopcode = 0;

	send_command((void *) &packet, sizeof(packet));
}


static int
send_command(void *buf, int bufsize)
{
	int	retval;

	retval = sendto(sock, buf, bufsize, 0,
			(struct sockaddr *) &dstsockaddr, sizeof(dstsockaddr));

	if( retval < 0 )
		perror("Sendto failed: ");
	return	retval;
}

static void
fatal(const char *fmt, ...)
{
	va_list args;

	va_start(args, fmt);
	vfprintf(stderr, fmt, args);
	va_end(args);

	fprintf(stderr, "\n");
	exit(1);
}

