/*
 * Copyright (c) 1999, 2000 The University of Utah and the Flux Group.
 * All rights reserved.
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */

/*
 * EE to test flow termination.
 */

#define TRYSETJMP

#include <stdio.h>
#include <assert.h>
#include <string.h>
#include <pthread.h>
#ifdef TRYSETJMP
#include <setjmp.h>
#endif

#include <nodeos/an_impl.h>
#include <nodeos/an_cred.h>
#include <nodeos/an_flow.h>
#include <nodeos/an_thread.h>
#include <nodeos/an_mutex.h>
#include <nodeos/an_cond.h>

#define USE_EXPLICIT_THREADS
#undef NODEOS_KILLS_THREADS
#define TIMEOUT

#define NUMFLOWS	10
#define NUMTHREADS	3

#ifdef TIMEOUT
#define RUNFOR		(30)
static void ee_starttimer(void);
#endif

static void ee_init(void *arg);
an_flowinit_func_t InitEE = { ee_init };

static void ee_spin(void *arg);
static void ee_syscall(void *arg);
static void ee_block(void *arg);
static void ee_child(void *arg);
static void ee_death(an_flow_t *flow);

int level = 0;
static an_flow_t *ee_startchild(void);
static an_mutex_t *lock, lock_data;
static an_cond_t *forever, forever_data;

static an_mutex_t *threadlock, threadlock_data;
static an_cond_t *threadcond, threadcond_data;
static int nthread;
static an_thread_t *mythreads[NUMFLOWS*NUMTHREADS];
#ifdef USE_EXPLICIT_THREADS
static an_thread_t *initthreads[NUMFLOWS];
static an_thread_t thread_data[NUMFLOWS*(NUMTHREADS+1)];
#endif

int
ee_main(int argc, const char **argv)
{
	an_flow_t *iflow;
	an_timespec_t sltime;

	printf("%p/%p: Flow test EE started\n",
	       an_flow_current(), an_thread_current());

	lock = an_mutex_create(&lock_data);
	forever = an_cond_create(&forever_data);

	threadlock = an_mutex_create(&threadlock_data);
	threadcond = an_cond_create(&threadcond_data);
	
#ifdef TIMEOUT
	ee_starttimer();
#endif

	printf("%p/%p: Creating child flow...\n",
	       an_flow_current(), an_thread_current());
	iflow = ee_startchild();

	an_mutex_lock(threadlock);
	while (nthread != NUMTHREADS * NUMFLOWS)
		an_cond_wait(threadcond, threadlock);
	an_mutex_unlock(threadlock);

	sltime.tv_sec = 3;
	sltime.tv_nsec = 0;
	an_thread_sleep(&sltime);

	printf("%p/%p: Killing child flow %p!\n",
	       an_flow_current(), an_thread_current(), iflow);
	an_flow_destroy(iflow);
#ifdef NODEOS_KILLS_THREADS
	printf("%p/%p: Child flow %p dead!\n",
	       an_flow_current(), an_thread_current(), iflow);
#ifdef USE_EXPLICIT_THREADS
	{
		int i;

		for (i = 0; i < NUMFLOWS * (NUMTHREADS+1); i++)
			if (an_thread_destroy(&thread_data[i]) != 0)
				printf("Destroy of thread %p (#%d) failed\n",
				       &thread_data[i], i);
	}
#endif
#endif

	return 0;
}

static an_flow_t *
ee_startchild(void)
{
	an_cpuspec_t cs;
	an_memspec_t ms;
	an_flow_t *flow;
#ifdef USE_EXPLICIT_THREADS
	int tix, i;
	void *stackptr;
	size_t stacksize;
#endif

	flow = an_mempool_alloc(an_mempool_current(), sizeof *flow);
	assert(flow != 0);

	ms.nchunks = (128*1024) / NODEOS_MEMPOOL_CHUNKSIZE;
	ms.pool = an_mempool_current();
	memset(&cs, 0, sizeof cs);
#ifdef USE_EXPLICIT_THREADS
	/* Create the threads */
	stackptr = 0;
	stacksize = 20 * 1024;
	tix = level * NUMTHREADS;
	for (i = 0; i < NUMTHREADS; i++) {
		mythreads[tix+i] = an_thread_create(&thread_data[tix+i],
						    stackptr, stacksize);
		assert(mythreads[tix+i] != 0);
	}
	initthreads[level] =
		an_thread_create(&thread_data[NUMFLOWS*NUMTHREADS+level],
				 stackptr, stacksize);
	assert(initthreads[level] != 0);
	cs.ttype = AN_TH_EXPLICIT;
	cs.threads.ex.nthreads = 1;
	cs.threads.ex.threads = &initthreads[level];
#else
	cs.ttype = AN_TH_IMPLICIT;
	cs.threads.im.maxthreads = NUMTHREADS+1;
	cs.threads.im.stacksize = 20 * 1024;
#endif

	flow = an_flow_create(flow, an_cred_current(), &cs, &ms,
			      ee_child, 0, ee_death);
	assert(flow != 0);

	return flow;
}

static void
ee_child(void *arg)
{
	int tix, err;

	assert(arg == 0);

	an_mutex_lock(threadlock);
	printf("%p/%p: Firing off my infinite loop thread...\n",
	       an_flow_current(), an_thread_current());
	tix = level * NUMTHREADS;
#ifdef USE_EXPLICIT_THREADS
	err = an_thread_start(mythreads[tix+0], an_flow_current(),
			      ee_spin, 0);
#else
	err = an_flow_startthread(an_flow_current(), ee_spin, 0,
				  &mythreads[tix+0]);
#endif
	assert(err == 0);

	printf("%p/%p: Firing off my syscall thread...\n",
	       an_flow_current(), an_thread_current());
#ifdef USE_EXPLICIT_THREADS
	err = an_thread_start(mythreads[tix+1], an_flow_current(),
			      ee_syscall, 0);
#else
	err = an_flow_startthread(an_flow_current(), ee_syscall, 0,
				  &mythreads[tix+1]);
#endif
	assert(err == 0);

	printf("%p/%p: Firing off my mutex-wait thread...\n",
	       an_flow_current(), an_thread_current());
#ifdef USE_EXPLICIT_THREADS
	err = an_thread_start(mythreads[tix+2], an_flow_current(),
			      ee_block, 0);
#else
	err = an_flow_startthread(an_flow_current(), ee_block, 0,
				  &mythreads[tix+2]);
#endif
	assert(err == 0);

	printf("%p/%p: threads %p/%p/%p started\n",
	       an_flow_current(), an_thread_current(),
	       mythreads[tix+0], mythreads[tix+1], mythreads[tix+2]);
	fflush(stdout);
	an_mutex_unlock(threadlock);

	if (++level < NUMFLOWS) {
		printf("%p/%p: Firing off a sub flow...\n",
		       an_flow_current(), an_thread_current());
		(void) ee_startchild();
	}
#if 0
	printf("%p/%p: done\n",
	       an_flow_current(), an_thread_current());
	fflush(stdout);
#endif
}

static void
ee_ihandler(int itype, void *info)
{
	an_mutex_lock(threadlock);
	printf("%p/%p: %s, exiting\n", an_flow_current(), an_thread_current(),
	       itype == AN_THREAD_INTR ? "INTR" : "TRAP");
	nthread--;
	an_cond_signal(threadcond);
	an_mutex_unlock(threadlock);
	an_thread_exit();
}

static void
ee_ihandler2(int itype, void *info)
{
	an_mutex_lock(threadlock);
	if (an_mutex_trylock(lock) == 0)
		printf("%p/%p: cond_wait inthandler: mutex not locked!\n",
		       an_flow_current(), an_thread_current());
#ifdef TRYSETJMP
	{
		sigjmp_buf *sjb = an_thread_getdata();

		printf("%p/%p: %s(cond_wait), longjmp'ing via %p\n",
		       an_flow_current(), an_thread_current(),
		       itype == AN_THREAD_INTR ? "INTR" : "TRAP", sjb);
		siglongjmp(*sjb, 1);
	}
#else
	printf("%p/%p: %s(cond_wait), exiting\n",
	       an_flow_current(), an_thread_current(),
	       itype == AN_THREAD_INTR ? "INTR" : "TRAP");
	nthread--;
	an_cond_signal(threadcond);
	an_mutex_unlock(threadlock);
	an_mutex_unlock(lock);
	an_thread_exit();
#endif
}

static void
ee_spin(void *arg)
{
	an_mutex_lock(threadlock);
	nthread++;
	printf("%p/%p: SPIN thread (%d)\n",
	       an_flow_current(), an_thread_current(), nthread);
	an_thread_sethandler(ee_ihandler);
	an_cond_signal(threadcond);
	an_mutex_unlock(threadlock);

	while (1)
		;
}

static void
ee_syscall(void *arg)
{
	an_mutex_lock(threadlock);
	nthread++;
	printf("%p/%p: SYSCALL thread (%d)\n",
	       an_flow_current(), an_thread_current(), nthread);
	an_thread_sethandler(ee_ihandler);
	an_cond_signal(threadcond);
	an_mutex_unlock(threadlock);

	while (1)
		(void) an_flow_current();
}

static void
ee_block(void *arg)
{
#ifdef TRYSETJMP
	sigjmp_buf sjb;
#endif

	an_mutex_lock(threadlock);
	nthread++;
	printf("%p/%p: BLOCKED thread (%d)\n",
	       an_flow_current(), an_thread_current(), nthread);
#ifdef TRYSETJMP
	if (sigsetjmp(sjb, 0) != 0) {
		printf("%p/%p: BLOCKED thread, exiting\n",
		       an_flow_current(), an_thread_current());
		nthread--;
		an_cond_signal(threadcond);
		an_mutex_unlock(threadlock);
		an_mutex_unlock(lock);
		an_thread_exit();
	}
	an_thread_setdata(&sjb);
#endif
	an_thread_sethandler(ee_ihandler2);
	an_cond_signal(threadcond);
	an_mutex_unlock(threadlock);

	an_mutex_lock(lock);
	while (1)
		an_cond_wait(forever, lock);
}

static void
ee_death(an_flow_t *flow)
{
#ifndef NODEOS_KILLS_THREADS
	int i;

	printf("%p/%p: Flowtest: flow %p killed, interrupting %d threads\n",
	       an_flow_current(), an_thread_current(), flow, nthread);
	an_mutex_lock(threadlock);
	for (i = 0; i < nthread; i++)
		an_thread_intr(mythreads[i]);
	while (nthread > 0)
		an_cond_wait(threadcond, threadlock);
	an_mutex_unlock(threadlock);
	printf("%p/%p: NullEE: all threads gone!\n",
	       an_flow_current(), an_thread_current());
#ifdef USE_EXPLICIT_THREADS
	for (i = 0; i < NUMFLOWS * (NUMTHREADS+1); i++)
		if (an_thread_destroy(&thread_data[i]) != 0)
			printf("Destroy of thread %p (#%d) failed\n",
			       &thread_data[i], i);
#endif
#endif
}

static void
ee_init(void *arg)
{
	int status;
	int argc = 1;
	const char *argv[] = { "FlowtestEE" }; 
	an_flow_t *flow = an_flow_current();

	status = ee_main(argc, argv);
	an_flow_destroy(flow);
}

#ifdef TIMEOUT

#include <nodeos/an_event.h>
static void ee_timeout(void *event, void *arg);

static void
ee_starttimer(void)
{
	int err;

	err = an_event_schedule(ee_timeout, 0, RUNFOR * 1000000);
	assert(err == 0);
	printf("Timer started, running for %d seconds\n", RUNFOR);
}

static
void ee_timeout(void *event, void *arg)
{
	printf("%p/%p: Flowtest: hung?\n",
	       an_flow_current(), an_thread_current());
	an_flow_destroy(an_flow_current());
}
#endif
