/*
 * Copyright (c) 1999, 2000 The University of Utah and the Flux Group.
 * All rights reserved.
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */

/*
 * Print info about any received packets on any interface to stdout.
 */
#include <stdio.h>
#include <assert.h>
#include <string.h>
#include <sys/types.h>

#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netinet/in_systm.h>
#include <netinet/if_ether.h>
#include <netinet/ip.h>
#include <netinet/ip_icmp.h>

#include <nodeos/an_impl.h>
#include <nodeos/an_flow.h>
#include <nodeos/an_thread.h>
#include <nodeos/an_chan.h>
#include <nodeos/an_mutex.h>
#include <nodeos/an_pbuf.h>
#include <nodeos/an_info.h>
#include <nodeos/an_event.h>
#include <nodeos/lib/arplib.h>
#include <nodeos/lib/pbcache.h>

#include <nodeos/impl/an_gprof.h>	/* XXX */

#define RUNFOR		60	/* run for this many seconds */
#define MAXTHREADS	2	/* max threads handling packets at once */
#define MAXBUFS		32	/* max packet buffers for those threads */

/*
 * Per-channel info
 */
struct chaninfo {
	an_chan_t inchan;
	an_demuxkey_t dk;
	an_chanaddr_t addr;
	an_chanproto_t proto;
	an_cpuspec_t cpu;
	an_pbuf_t *bufs[MAXBUFS];
	int id;
} *chans;

static pbcache_t *pbufcache;

static void ee_init(void *arg);
an_flowinit_func_t InitEE = { ee_init };

static void ee_die(void *event, void *arg);
static void packet_receive(an_pbuf_t *pbuf, void *arg, int msize);
static void setup_channels(void);

static int dying;

int
ee_main(int argc, const char **argv)
{
	assert(argc >= 1);

	printf("%p/%p: %s initialization.\n",
	       an_flow_current(), an_thread_current(), argv[0]);

	/*
	 * Setup channels
	 */
	pbufcache = pbcache_create(MAXBUFS, 0);
	setup_channels();

	printf("%s: all channels are 'GO'.\n", argv[0]);

	an_gprof_start(argv[0]);
#ifdef OSKIT
	{
		extern int memlogging;
		memlogging = 1;
	}
#endif        
	return 0;
}

static void
setup_channels(void)
{
	an_error_t err;
	an_flow_t *flow;
	an_infoif_t info[8];
	int nif = 8;
	int i;
	int b;

	flow = an_flow_current();

	/*
	 * Get info on all the installed interfaces.
	 */
	err = an_info_getifs(info, &nif);
	assert(err == 0);
	assert(nif > 0);
	assert(nif <= 8);

	/*
	 * Open one channel for each interface.
	 */
	chans = an_mempool_alloc(an_mempool_current(), sizeof(*chans) * nif);
	memset(chans, 0, sizeof(*chans) * nif);
	for (i = 0; i < nif; i++)
	{
		an_chan_t *ich;

		/* Assume ethernet (not ATM, FDDI, etc). */
		assert(info[i].type == ANIF_ETH);

		chans[i].id = i;

		/*
		 * raw protocol, all wildcard addresses
		 */
		snprintf(chans[i].proto.data, 5, "eth%d", i);
		chans[i].proto.len = strlen(chans[i].proto.data);
		/* XXX wildcard address/type is just '0' */
		memset(&(chans[i].addr.eth.daddr), 0, sizeof(chans[i].addr.eth.daddr));
		memset(&(chans[i].addr.eth.saddr), 0, sizeof(chans[i].addr.eth.daddr));
		chans[i].addr.eth.type = 0;

		/*
		 * Create an empty key that matches everything.
		 */
		an_demuxkey_create(&(chans[i].dk), flow);
		
		/*
		 * Two threads, take two buffers from cache
		 */
		chans[i].cpu.ttype = AN_TH_IMPLICIT;
		chans[i].cpu.threads.im.maxthreads = MAXTHREADS;
		for (b = 0; b < MAXTHREADS; b++)
			chans[i].bufs[b] = pbcache_alloc(pbufcache, 0);

		/*
		 * Create the inchan
		 */
		ich = an_inchan_create(&(chans[i].inchan),
				       flow,
				       &(chans[i].dk),
				       &(chans[i].proto),
				       &(chans[i].addr),
				       chans[i].bufs, MAXTHREADS, &(chans[i].cpu),
				       packet_receive, &(chans[i]));
		assert(ich != 0);

		assert(info[i].type == ANIF_ETH);
		/* XXX print the ethernet address for each channel, too. */
		printf("%p: created inchannel for %s\n",
		       ich, (char*)(&(chans[i].proto.data)));
	}

	/*
	 * Setup an event to time us out in RUNFOR seconds
	 */
	err = an_event_schedule(ee_die, 0, RUNFOR * 1000000);
	assert(err == 0);
	printf("Timer started, running for %d seconds\n", RUNFOR);
}

/*
 * handler for packets.
 *
 * Just print the first part of the packet
 */
static void
packet_receive(an_pbuf_t *pbuf, void *arg, int msize)
{
	int psize = an_pbuf_size(pbuf);
	char *data = an_pbuf_data(pbuf);
	struct chaninfo *cinfo;
	int i;

	if (dying)
		goto done;
	
	cinfo = (struct chaninfo*)(arg);

	/* replace the current pbuf with a new one for new packets. */
	an_inchan_addpbuf(&(cinfo->inchan), pbcache_alloc(pbufcache, 0));

	printf("%d [%4d]: ", cinfo->id, psize);
	for (i = 0; i < psize; i++)
	{
		printf("%02x ", (data[i] & 0xff));

		if (i == 16)
		{
			printf("...");
			break;
		}
	}
	printf("\n");

done:
	pbcache_free(pbufcache, pbuf);
}

static void
ee_die(void *event, void *arg)
{
	dying = 1;
	an_gprof_stop();

	/* XXX print stats on the inchans */

	/* XXX destroy the inchans */

	printf("dying.\n");

	pbcache_destroy(pbufcache, 0);
	an_flow_destroy(an_flow_current());
}

static void
ee_init(void *arg)
{
	int status;
	int argc = 1;
	const char *argv[] = { "packetdump" }; 

	status = ee_main(argc, argv);
}
