/*
 * Copyright (c) 1999, 2000, 2001 The University of Utah and the Flux Group.
 * All rights reserved.
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */

#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <string.h>
#include <sys/types.h>

#include <sys/socket.h>
#include <netinet/in.h>

#include <nodeos/an_impl.h>
#include <nodeos/an_flow.h>
#include <nodeos/an_thread.h>
#include <nodeos/an_chan.h>
#include <nodeos/an_pbuf.h>

#include <nodeos/impl/an_gprof.h>	/* XXX */

#define RUNFOR		10	/* run for this many seconds */
#define MAXBUFS		32

#define TTCPPORT	5001	/* default ttcp port */

static void ee_init(void *arg);
an_flowinit_func_t InitEE = { ee_init };

static void ee_sendudp(void);
static void ee_die(an_pbuf_t *packet, void *arg, int msize);
static void setup_channels(int blow, int blowhostip, int runfor);

static an_chan_t *die, die_data;
static an_pbuf_t diebuf_data;
static an_chan_t *udpoutchan, udpoutchan_data;
static void ee_udp(an_pbuf_t *packet, void *arg, int msize);
static void ee_ttcp(an_pbuf_t *packet, void *arg, int msize);

static an_chan_t *udpinchan, udpinchan_data, *ttcpinchan, ttcpinchan_data;
static an_pbuf_t udpbuf_data, ttcpbufs_data[MAXBUFS];
static struct {
	int off;
	int len;
	char data[8];
} ttcppattern = { 0, 8, { 040, 041, 042, 043, 044, 045, 046, 047 } };
static an_demuxkey_t ttcpdemux_data, *ttcpdemux;

static int npackets, ttcppackets, chandead;
static unsigned long nbytes;

static void
ee_init(void *arg)
{
	an_initargs_t *initargs;
	int blow = 0;
	int blowhostip = 0x9b63d658;	/* moab.cs.utah.edu */
	
	/*
	 * Parse command line arguments.
	 *    socktest [blow [blowhostip]]
	 */
	initargs = (an_initargs_t*)arg;
	assert(initargs);
	if (initargs->argc > 2) 
	{
		blow = atoi(initargs->argv[1]);
		sscanf(initargs->argv[2], "%i", &blowhostip);
	}
	else if (initargs->argc == 2)
	{
		blow = atoi(initargs->argv[1]);
	}

	/*
	 * Print a banner.
	 */
	if (blow) {
		printf("%p/%p: socktest started (sending to %d.%d.%d.%d)\n",
		       an_flow_current(), an_thread_current(),
		       (blowhostip>>24)&0xFF, (blowhostip>>16)&0xFF,
		       (blowhostip>>8)&0xFF, blowhostip&0xFF);
	}
	else 
	{
		printf("%p/%p: socktest started (recv)\n",
		       an_flow_current(), an_thread_current());
	}

	/*
	 * Setup channels
	 */
	setup_channels(blow, blowhostip, RUNFOR);

	if (blow)
	{
		printf("sending\n");
		ee_sendudp();
	}
	else
	{
		printf("waiting\n");
	}

	an_gprof_start(initargs->argv[0]);
#if 0
#ifdef OSKIT
	{
		extern int memlogging;
		memlogging = 1;
	}
#endif
#endif
	return;
}

/*
 * blow is 1 or 0.  blowhostip is the host to blow packets at, if blow
 * is 1.  Runfor is the timeout to set the clock channel at.  We'll
 * close down after the clock channel fires.
 */
static void
setup_channels(int blow, int blowhostip, int runfor)
{
	struct {
		an_chanaddr_t addr;
		an_chanproto_t proto;
		an_cpuspec_t cpu;
		an_pbuf_t *bufs[MAXBUFS];
	} attr;
	an_error_t err;
	an_flow_t *flow = an_flow_current();
	int i;

	memset(&attr, 0, sizeof attr);

	/*
	 * Setup a clock channel to time us out in 'runfor' seconds
	 */
	strcpy(attr.proto.data, "clock");
	attr.proto.len = strlen(attr.proto.data);
	attr.addr.clock.freq = (runfor * 100);
	attr.cpu.ttype = AN_TH_IMPLICIT;
	attr.cpu.threads.im.maxthreads = 2;
	attr.bufs[0] = an_pbuf_create(&diebuf_data, AN_PBUF_CLOCK, 0, 0);
	assert(attr.bufs[0] != 0);
	die = an_inchan_create(&die_data, flow, 0, &attr.proto, &attr.addr,
			       attr.bufs, 1, &attr.cpu, ee_die, 0);
	assert(die != 0);
	printf("%p: clockchannel for %d seconds\n", die, runfor);
	
	if (blow)
	{
		/*
		 * Setup out channel for UDP
		 */
		strcpy(attr.proto.data, "socket:if/ipv4/udp");
		attr.proto.len = strlen(attr.proto.data);
		attr.addr.udp4.daddr = htonl(blowhostip);
		attr.addr.udp4.dport = htons(TTCPPORT);
		udpoutchan = an_outchan_create(&udpoutchan_data, flow,
					       &attr.proto, &attr.addr, 0);
		assert(udpoutchan != 0);
		printf("%p: outchannel for port=%d\n", udpoutchan, TTCPPORT);
	}
	else
	{
		/*
		 * Setup in channels for UDP and UDP-based TTCP
		 */
		strcpy(attr.proto.data, "socket:if/ipv4/udp");
		attr.proto.len = strlen(attr.proto.data);
		attr.addr.udp4.daddr = htonl(INADDR_ANY);
		attr.addr.udp4.dport = htons(TTCPPORT);
		attr.cpu.ttype = AN_TH_IMPLICIT;
		attr.cpu.threads.im.maxthreads = 10;
		attr.bufs[0] = an_pbuf_create(&udpbuf_data, AN_PBUF_SOCK, 0, 0);
		assert(attr.bufs[0] != 0);
		udpinchan = an_inchan_create(&udpinchan_data, flow, 0,
					     &attr.proto, &attr.addr,
					     attr.bufs, 1, &attr.cpu, ee_udp, 0);
		if (udpinchan == 0)
			printf("WARNING! Could not create UDP socket channel.\n"
			       "         Probably configured without "
			       "--enable-chan_socket.\n");
		assert(udpinchan != 0);
		printf("%p: inchannel for TTCP traffic\n", udpinchan);

		ttcpdemux = an_demuxkey_create(&ttcpdemux_data, an_flow_current());
		assert(ttcpdemux != 0);
		err = an_demuxkey_add(ttcpdemux, ttcppattern.off, ttcppattern.len,
				      ttcppattern.data, 0);
		assert(err == 0);
		
		strcpy(attr.proto.data, "socket:if/ipv4/udp");
		attr.proto.len = strlen(attr.proto.data);
		attr.addr.udp4.daddr = htonl(INADDR_ANY);
		attr.addr.udp4.dport = htons(TTCPPORT);
		attr.cpu.ttype = AN_TH_IMPLICIT;
		attr.cpu.threads.im.maxthreads = 10;
		for (i = 0; i < MAXBUFS; i++) {
			attr.bufs[i] = an_pbuf_create(&ttcpbufs_data[i], AN_PBUF_SOCK,
						      0, 0);
			assert(attr.bufs[i] != 0);
		}
		ttcpinchan = an_inchan_create(&ttcpinchan_data, flow, ttcpdemux,
					      &attr.proto, &attr.addr,
					      attr.bufs, MAXBUFS, &attr.cpu,
					      ee_ttcp, 0);
		assert(ttcpinchan != 0);
		printf("%p: inchannel for TTCP traffic (default pattern)\n",
		       ttcpinchan);
	}
}

#include <nodeos/an_mutex.h>
#include <nodeos/an_cond.h>

#ifdef OSKIT
#include <oskit/machine/proc_reg.h>
#else
#define get_tsc() (0)
#endif
unsigned long long sstamp, estamp;
unsigned tsc_freq = 200;

static void
ee_sendudp(void)
{
	an_pbuf_t *packet;
	char datums[2048];
	an_error_t err;

	memset(datums, 0x69, sizeof datums);
	packet = an_pbuf_create(datums, AN_PBUF_SOCK, 0, sizeof datums);
	sstamp = get_tsc();
	while (chandead == 0) {
		err = an_outchan_send(udpoutchan, packet);
		assert(err == 0);
		npackets++;
		nbytes += sizeof datums;
		estamp = get_tsc();
	}
}

static void
ee_udp(an_pbuf_t *pbuf, void *arg, int msize)
{
	int psize = an_pbuf_size(pbuf);

	if (chandead)
		return;

	if (npackets == 0)
		sstamp = get_tsc();

	npackets++;
	nbytes += psize;
	an_inchan_addpbuf(udpinchan, pbuf);

	estamp = get_tsc();
}

static void
ee_ttcp(an_pbuf_t *pbuf, void *arg, int msize)
{
	int psize = an_pbuf_size(pbuf);

	if (chandead)
		return;

	if (npackets == 0)
		sstamp = get_tsc();

	npackets++;
	ttcppackets++;
	nbytes += psize;
	an_inchan_addpbuf(ttcpinchan, pbuf);

	estamp = get_tsc();
}

static void
ee_die(an_pbuf_t *pbuf, void *arg, int msize)
{
	chandead = 1;

#if 0
	an_gprof_stop();
#endif

	if (sstamp == estamp)
		printf("Done.  Probably didn't get/send any packets at all.\n");
	else 
		printf("Done: %lu bytes, %d packets in %u usec (%u bytes/sec)\n",
		       nbytes, npackets, (unsigned)((estamp - sstamp)/tsc_freq),
		       (unsigned)(((unsigned long long)nbytes*1000000) /
				  ((estamp - sstamp)/tsc_freq)));
	if (ttcppackets)
		printf("      %d default TTCP packets\n", ttcppackets);

#if 0
	an_chan_destroy(ttcpinchan);
	an_chan_destroy(udpinchan);
#endif
	an_flow_destroy(an_flow_current());
}

