/*
 * Copyright (c) 1999, 2000, 2001 The University of Utah and the Flux Group.
 * All rights reserved.
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */

#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <string.h>
#include <sys/types.h>

#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netinet/in_systm.h>
#include <netinet/if_ether.h>

#include <nodeos/an_impl.h>
#include <nodeos/an_flow.h>
#include <nodeos/an_thread.h>
#include <nodeos/an_chan.h>
#include <nodeos/an_pbuf.h>
#include <nodeos/an_event.h>
#include <nodeos/an_info.h>
#include <nodeos/lib/utils.h>

#include <nodeos/impl/an_gprof.h>	/* XXX */

#define RUNFOR		30	/* run for this many seconds */
#define INIF		0	/* interface to use for input; i.e. "if<INIF>" */
#define OUTIF		0	/* ditto for output */
#define MAXBUFS		64

static void ee_init(void *arg);
an_flowinit_func_t InitEE = { ee_init };

static void ee_die(void *event, void *arg);
static void setup_channels(int blow, int blowhostip, int runfor);

/* source enet addr on INIF (filled in by nodeos) */
an_ethaddrspec_t myaddr = {
	{0x00, 0x00, 0x00, 0x00, 0x00, 0x00}, {0}, ETHERTYPE_IP
};

/* source UDP/IP info (dport, daddr) */
an_udp4addrspec_t myudp = {
	0, 0,
	6969,		/* dport */
#if 1
	0x9b63d697	/* daddr: moab */
//	0x9b63d670	/* daddr: tbpc12 */
#else
	0xcea3991a	/* daddr: poohbear */
#endif
};

/* target enet addr on OUTIF */
an_ethaddrspec_t toaddr = {
#if 1
	{0x00, 0x60, 0xf5, 0x04, 0x00, 0x53}, {0}, ETHERTYPE_IP	/* moab, qfe3 */
#else
	{0x00, 0xa0, 0xc9, 0xf1, 0x09, 0xcb}, {0}, ETHERTYPE_IP /* Golden */
#endif
};

/* target UDP/IP info (dport, daddr) */
an_udp4addrspec_t toudp = {
	0, 0,
	6969,		/* dport */
#if 1
	0x9b63d658	/* daddr: moab */
#else
	0xcea39919	/* daddr: golden */
#endif
};

static an_pbuf_t	*ethbufs[MAXBUFS], ethbufs_data[MAXBUFS];
static an_pbuf_t	*udpbufs[MAXBUFS], udpbufs_data[MAXBUFS];

static void ee_udprecv(an_pbuf_t *packet, void *arg, int msize);

static an_chan_t *udpinchan, udpinchan_data, *ethinchan, ethinchan_data;
static an_chan_t *udpoutchan, udpoutchan_data, *ethoutchan, ethoutchan_data;

static int npackets, chandead;
static unsigned long nbytes;

static void
ee_init(void *arg)
{
	an_initargs_t *initargs;
	int blow = 0;
	int blowhostip = 0x9b63d658;	/* moab.cs.utah.edu */
	
	/*
	 * Parse command line arguments.
	 *    socktest [blow [blowhostip]]
	 */
	initargs = (an_initargs_t*)arg;
	assert(initargs);
	if (initargs->argc > 2) 
	{
		blow = atoi(initargs->argv[1]);
		sscanf(initargs->argv[2], "%i", &blowhostip);
	}
	else if (initargs->argc == 2)
	{
		blow = atoi(initargs->argv[1]);
	}

	/*
	 * Print a banner.
	 */
	if (blow) {
		printf("%p/%p: udptest started (sending to %d.%d.%d.%d)\n",
		       an_flow_current(), an_thread_current(),
		       (blowhostip>>24)&0xFF, (blowhostip>>16)&0xFF,
		       (blowhostip>>8)&0xFF, blowhostip&0xFF);
	}
	else 
	{
		printf("%p/%p: udptest started (recv)\n",
		       an_flow_current(), an_thread_current());
	}

	/*
	 * Setup channels
	 */
	setup_channels(blow, blowhostip, RUNFOR);

	if (blow)
	{
		printf("sending\n");
/*		ee_sendudp(); */
	}
	else
	{
		printf("waiting\n");
	}

	an_gprof_start(initargs->argv[0]);
#if 0
#ifdef OSKIT
	{
		extern int memlogging;
		memlogging = 1;
	}
#endif
#endif
	return;
}

/*
 * Create a Click based UDP/IP in channel.
 * We explicitly provide an interface-level channel (i.e., an "if" channel)
 * from which the Click graph will receive input packets.  In other words,
 * we first create an "if" interface and then pass that to the channel create
 * call for the ultimate udp/ip channel.
 *
 * Note that this is an interim interface.  Eventually we will pass in a
 * click graph description along with some number of "if" based in/out
 * channels which serve as the inputs and outputs of the graph.
 */
static void
setup_inchannel(char *ifname, an_ethaddrspec_t *ifaddr,
		an_pbuf_t *ifbufs[], int numifbufs,
		an_udp4addrspec_t *udpaddr, an_pbuf_t *ubufs[], int numubufs)
{
	struct {
		an_chanaddr_t addr;
		an_chanproto_t proto;
		an_cpuspec_t cpu;
	} attr;

	memset(&attr, 0, sizeof attr);

	/*
	 * Create an interface-level channel which selects the interface(s)
	 * that UDP/IP packets are to be received from.
	 *
	 * Protocol should be "ifN", not "ethN" as the Click code assumes that
	 * any ethernet header has been stripped.
	 *
	 * The address spec should identify the destination MAC addr (and
	 * optionally the source MAC addr and ethernet type) for packets to
	 * be fed into the Click UDP/IP graph.  Fields with value 0 mean "any".
	 *
	 * The demux provided here is applied to the packets as they come
	 * off the wire.  This can be used to refine the packet selection.
	 *
	 *	Note: basic UDP/IP demuxing is implied when this channel is
	 *	composed with a "udp/ip" channel, so this user provided portion
	 *	should not conflict with that; i.e., select TCP packets.
	 *
	 *	Note: the demux specified with the "udp/ip" channel can be
	 *	used to select packets based on the payload of the UDP packet,
	 *	this demux is intended to match against header info.
	 *
	 *	Note: this demux is somewhat limited in that it must match
	 *	all the pieces of a fragmented UDP packet.
	 *
 	 * The thread limit means nothing here.  A single thread is used to
	 * run packets through the Click UDP/IP graph.
	 *
	 * The pbufs given here are used to buffer packets as they come off
	 * the interface but before they enter the Click graph.
	 *
	 * The receive function is ignored.
	 */
	strcpy(attr.proto.data, ifname);
	attr.proto.len = strlen(attr.proto.data);
	attr.addr.eth = *ifaddr;
	attr.cpu.ttype = AN_TH_IMPLICIT;
	attr.cpu.threads.im.maxthreads = 1;
	ethinchan = an_inchan_create(&ethinchan_data, an_flow_current(),
				     0,		/* demux */
				     &attr.proto, &attr.addr,
				     ifbufs, numifbufs, &attr.cpu,
				     0, 0);	/* function & arg */
	assert(ethinchan != 0);

	/*
	 * Now create the Click-based UDP/IP channel by "composing" the
	 * "if" inchan and a Click graph.  That is, packets coming in via
	 * the "if" chan are input to the Click UDP/IP processing graph.
	 * The resulting channel looks much like a socket:udp/ip channel.
	 *
	 * Protocol should be "udp/ip".
	 *
	 * The address spec should identify the destination IPv4 address
	 * and port number (also optionally source addr/pnum?)
	 *
	 * The demux provided here is applied to the UDP packet payload.
	 *
 	 * The thread limit here is the maximum number of simultaneous
	 * invocations of the packet receive function.
	 *
	 * The pbufs given here are used to buffer packets as they come
	 * out of the Click graph.
	 *
	 * The receive function is the function to be called with every
	 * (header-stripped) UDP packet.
	 */
	strcpy(attr.proto.data, "udp/ip");
	attr.proto.len = strlen(attr.proto.data);
	attr.addr.udp4 = *udpaddr;
	attr.cpu.ttype = AN_TH_IMPLICIT;
	attr.cpu.threads.im.maxthreads = 10;
	udpinchan = an_clickinchan_create(&udpinchan_data, an_flow_current(),
					  0,		/* demux */
					  &attr.proto, &attr.addr,
					  ubufs, numubufs, &attr.cpu,
					  ee_udprecv, 0, /* func & arg */
					  ethinchan);
	assert(udpinchan);
}

/*
 * Create a Click based UDP/IP out channel.
 *
 * Analogous to the in channel creation, we explicitly provide an
 * interface-level channel (i.e., an "if" channel) to which the Click
 * graph will output its processed packets.  In other words, we first
 * create an "if" interface and then pass that to the channel create
 * call for the ultimate udp/ip channel.
 *
 * Again, note that this is an interim interface.
 */
static void
setup_outchannel(char *ifname, an_ethaddrspec_t *ifaddr,
		 unsigned int ifbandwidth, an_udp4addrspec_t *udpaddr)
{
	struct {
		an_chanaddr_t addr;
		an_chanproto_t proto;
		an_netspec_t net;
	} attr;

	/*
	 * Create an interface-level channel which selects the interface(s)
	 * that UDP/IP packets are to be sent out on.
	 *
	 * Protocol should be "ifN", not "ethN" as the Click code does not
	 * prepend an ethernet header.
	 *
	 * The address spec should identify the destination MAC addr (and
	 * optionally the source MAC addr and ethernet type) for packets to be
	 * put out on the indicated interface.  Fields with value 0 mean "any".
	 *
	 * The network resource specification limits the amount of interface
	 * bandwidth used for our UDP/IP packets.
	 */
	memset(&attr, 0, sizeof attr);
	strcpy(attr.proto.data, ifname);
	attr.proto.len = strlen(attr.proto.data);
	attr.addr.eth = *ifaddr;
	attr.net.bandwidth = ifbandwidth;
	ethoutchan = an_outchan_create(&ethoutchan_data, an_flow_current(),
				     &attr.proto, &attr.addr, &attr.net);
	assert(ethoutchan != 0);

	/*
	 * Now create the Click-based UDP/IP channel by "composing" the
	 * Click graph with the "if" outchan.  That is, packets processed
	 * by the Click UDP/IP processing graph are fed into (outchan_send)
	 * the "if" channel.  The resulting channel looks much like a
	 * socket:udp/ip channel.
	 *
	 * Protocol should be "udp/ip".
	 *
	 * The address spec should identify the destination IPv4 address
	 * and port number (also optionally source addr/pnum?)
	 *
	 * The network resource specification is ignored.
	 */
	memset(&attr, 0, sizeof attr);
	strcpy(attr.proto.data, "udp/ip");
	attr.proto.len = strlen(attr.proto.data);
	attr.addr.udp4 = *udpaddr;
	udpoutchan = an_clickoutchan_create(&udpoutchan_data,
					    an_flow_current(),
					    &attr.proto, &attr.addr, &attr.net,
					    ethoutchan);
	assert(udpoutchan);
}

static void
setup_channels(int blow, int blowhostip, int runfor)
{
	int	i;
	an_error_t err;
	an_infoif_t info[8];
	int nif = 8;
	char inifname[8], outifname[8];

	err = an_info_getifs(info, &nif);
	assert(err == 0);

	/*
	 * Find my ETH interface address
	 */
	assert(nif >= INIF);
	assert(info[INIF].type == ANIF_ETH);
	memcpy(myaddr.daddr, info[INIF].addr.eth.daddr, sizeof myaddr.daddr);
	myaddr.type = htons(myaddr.type);
	sprintf(inifname, "if%d", INIF);
#if INIF == OUTIF
	strcpy(outifname, inifname);
#else
	assert(nif >= OUTIF);
	assert(info[OUTIF].type == ANIF_ETH);
	sprintf(outifname, "if%d", OUTIF);
#endif

	/*
	 * Dest addr
	 */
	toaddr.type = htons(toaddr.type);

	/*
	 * Setup an event to time us out in RUNFOR seconds
	 */
	err = an_event_schedule(ee_die, 0, RUNFOR * 1000000);
	assert(err == 0);
	printf("Timer started, running for %d seconds\n", RUNFOR);

	/*
	 * Buffers for input side.
	 */
	for (i = 0; i < MAXBUFS; i++) {
		ethbufs[i] = an_pbuf_create(&ethbufs_data[i], AN_PBUF_ETH,
					    0, 0);
		assert(ethbufs[i] != 0);

		udpbufs[i] = an_pbuf_create(&udpbufs_data[i], AN_PBUF_DEFAULT,
					    0, 0);
		assert(udpbufs[i] != 0);
	}

	/*
	 * UDP specs.
	 */
	myudp.sport = htons(myudp.sport);
	myudp.saddr = htonl(myudp.saddr);
	myudp.dport = htons(myudp.dport);
	myudp.daddr = htonl(myudp.daddr);

	toudp.sport = myudp.dport;
	toudp.saddr = myudp.daddr;
	toudp.dport = htons(toudp.dport);
	toudp.daddr = htonl(toudp.daddr);

	setup_inchannel(inifname, &myaddr, ethbufs, MAXBUFS,
			&myudp, udpbufs, MAXBUFS);
	printf("receiving UDP/IP for %d.%d.%d.%d/%d (IP/port) "
	       "on %s to %02x:%02x:%02x:%02x:%02x:%02x\n",
	       (myudp.daddr>>0)&0xFF, (myudp.daddr>>8)&0xFF,
	       (myudp.daddr>>16)&0xFF, (myudp.daddr>>24)&0xFF,
	       ntohs(myudp.dport),
	       inifname,
	       myaddr.daddr[0], myaddr.daddr[1], myaddr.daddr[2],
	       myaddr.daddr[3], myaddr.daddr[4], myaddr.daddr[5]);

	setup_outchannel(outifname, &toaddr, 15, &toudp);
	printf("sending UDP/IP to %d.%d.%d.%d/%d (IP/port) "
	       "on %s to %02x:%02x:%02x:%02x:%02x:%02x\n",
	       (toudp.daddr>>0)&0xFF, (toudp.daddr>>8)&0xFF,
	       (toudp.daddr>>16)&0xFF, (toudp.daddr>>24)&0xFF,
	       ntohs(toudp.dport),
	       outifname,
	       toaddr.daddr[0], toaddr.daddr[1], toaddr.daddr[2],
	       toaddr.daddr[3], toaddr.daddr[4], toaddr.daddr[5]);

}

static void
ee_udprecv(an_pbuf_t *pbuf, void *arg, int msize)
{
	int err;
	int psize = an_pbuf_size(pbuf);

	if (chandead)
		return;

	npackets++;
	nbytes += psize;

	err = an_outchan_send(udpoutchan, pbuf);
	assert(err == 0);

	an_inchan_addpbuf(udpinchan, pbuf);
}

static void
ee_die(void *event, void *arg)
{
	chandead = 1;

	printf("Done: %lu bytes, %d packets\n", nbytes, npackets);

	moab_chan_printstats(udpinchan, 1, "UDP in", RUNFOR, 0);
	moab_chan_printstats(udpoutchan, 1, "UDP out", RUNFOR, 0);
	moab_chan_printstats(ethinchan, 1, "ETH in", RUNFOR, 0);
	moab_chan_printstats(ethoutchan, 1, "ETH out", RUNFOR, 0);

#if 0
	an_chan_destroy(udpinchan);
	an_chan_destroy(udpoutchan);
	an_chan_destroy(ethinchan);
	an_chan_destroy(ethoutchan);
#endif

	an_flow_destroy(an_flow_current());
}
