/*
 * Copyright (c) 1999, 2000, 2001 The University of Utah and the Flux Group.
 * All rights reserved.
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */

#undef VERBOSE
#define ARPCACHE

/*
 * Arp support. We don't cache anything, but issue a new request each time.
 * Yep, rather simple but it will do. 
 */
#include <stdio.h>
#include <assert.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>

#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <net/if.h>
#include <netinet/in_systm.h>
#include <netinet/if_ether.h>
#include <netinet/ip.h>
#include <netinet/ip_icmp.h>

#include <nodeos/an_impl.h>
#include <nodeos/an_flow.h>
#include <nodeos/an_thread.h>
#include <nodeos/an_chan.h>
#include <nodeos/an_cond.h>
#include <nodeos/an_pbuf.h>
#include <nodeos/an_info.h>
#include <nodeos/an_event.h>
#include <nodeos/lib/arplib.h>

#ifndef ETHER_ADDR_LEN
#define ETHER_ADDR_LEN	6
#endif

#define MAXIFACES	8
#define MAXCENTRIES	4

/* Stuff for arp channels */
static an_infoif_t	ifinfo[MAXIFACES];
static int		nif = MAXIFACES;
static u_char		bcastaddr[6] = {0xff, 0xff, 0xff, 0xff, 0xff, 0xff};
static int		arpready;

#ifdef ARPCACHE
/*
 * Simple ARP cache.
 * Entries are never implicitly removed.
 * Only used for hardwiring entries.
 */
struct arpcacheentry {
	struct in_addr	ipaddr;	/* Note: network order! */
	unsigned char	ethaddr[ETHER_ADDR_LEN];
	unsigned short	valid;
};
#endif

/*
 * Record interface address and ip address for each interface.
 */
struct arpentry {
	int		initialized;
	unsigned char	ethaddr[ETHER_ADDR_LEN];
	struct in_addr  ipaddr;
	an_chan_t	*repinchan, repinchan_data;	/* Incoming Replies */
	an_pbuf_t	*arprepbuf, arprepbuf_data;
	an_demuxkey_t	repdemux_data;
	an_chan_t	*reqinchan, reqinchan_data;	/* Incoming Requests */
	an_pbuf_t	*arpreqbuf, arpreqbuf_data;
	an_demuxkey_t	reqdemux_data;
	an_chan_t	*dreqinchan, dreqinchan_data;	/* Direct requests */
	an_pbuf_t	*arpdreqbuf, arpdreqbuf_data;
	an_demuxkey_t	dreqdemux_data;
	an_chan_t	*reqchan, reqchan_data;		/* Outgoing Requests */
	an_chan_t	*outchan, outchan_data;		/* Outgoing Replies */
	unsigned char   outaddr[ETHER_ADDR_LEN];	/* Address */
	an_mutex_t	*arplock, arplock_data;
	an_mutex_t	*waitlock, waitlock_data;
	an_cond_t	*waitcond, waitcond_data;
	volatile int	waiting_for_reply;
	unsigned char	arpethaddr[ETHER_ADDR_LEN];
#ifdef ARPCACHE
	struct arpcacheentry cache[MAXCENTRIES];
	int		cachelastused;
#endif
};
struct arpentry arptab[MAXIFACES];

/* Forward decls */
static void		ee_arprequest(an_pbuf_t *pbuf, void *arg, int msize);
static void		ee_arpreply(an_pbuf_t *pbuf, void *arg, int msize);

int
arp_init(void)
{
	an_error_t	err;
	int		i;

	if (arpready)
		return 0;

	err = an_info_getifs(ifinfo, &nif);
	assert(err == 0);

	for (i = 0; i < nif; i++) {
		memcpy(arptab[i].ethaddr,
		       ifinfo[i].addr.eth.daddr, sizeof arptab[i].ethaddr);
	}
	arpready = 1;

	return 0;
}

/*
 * Clean up before exit.
 */
int
arp_exit(void)
{
	int		i;

	if (!arpready)
		return 0;

	for (i = 0; i < nif; i++) {
		struct arpentry	*arpentry = &arptab[i];

		if (! arpentry->initialized)
			continue;

		an_outchan_destroy(arpentry->reqchan);
		if (arpentry->outchan)
			an_outchan_destroy(arpentry->outchan);
		
		an_inchan_destroy(arpentry->repinchan);
		an_inchan_destroy(arpentry->reqinchan);
	}
	return 0;
}

int
arp_config_iface(int iface, struct in_addr *ipaddr)
{
	an_demuxkey_t		*arpdemux;
	struct arpentry		*arpentry = &arptab[iface];
	an_pbuf_t		*pbuf;
	an_chan_t		*chan;
	struct ether_arp	key, mask;
	an_error_t		err;
	an_flow_t		*flow = an_flow_current();
	struct {
		an_chanaddr_t addr;
		an_chanproto_t proto;
		an_cpuspec_t cpu;
		an_pbuf_t *bufs[1];
	} attr;

	if (!arpready) {
		printf(__FUNCTION__ ": Arp module not initialized\n");
		return -1;
	}

	if (arpentry->initialized) {
		/*
		 * XXX We should allow for reconfiguration, but currently
		 * running the code a second time causes it to spaz out and
		 * the network gets flooded.
		 */
		return 0;
	}

	assert(iface < MAXIFACES);

	/*
	 * This is who we will be!
	 */
	arpentry->ipaddr.s_addr = ipaddr->s_addr;

	/*
	 * Cuncurrency not allowed yet!
	 */
	arpentry->arplock  = an_mutex_create(&arpentry->arplock_data);
	arpentry->waitlock = an_mutex_create(&arpentry->waitlock_data);
	arpentry->waitcond = an_cond_create(&arpentry->waitcond_data);

	/*
	 * Setup in channel for broadcast (arp) packets. These are packets
	 * we have to reply to.
	 * First, create a demux to match ARP requests.
	 */
	memset(&key, 0, sizeof key);
	memset(&mask, 0, sizeof mask);
	key.arp_hrd = htons(ARPHRD_ETHER);
	mask.arp_hrd = ~0;
	key.arp_pro = htons(ETHERTYPE_IP);
	mask.arp_pro = ~0;
	key.arp_op = htons(ARPOP_REQUEST);
	mask.arp_op = ~0;
	memcpy(&key.arp_tpa, &ipaddr->s_addr, sizeof key.arp_tpa);
	memset(&mask.arp_tpa, ~0, sizeof mask.arp_tpa);

	arpdemux = an_demuxkey_create(&arpentry->reqdemux_data, flow);
	assert(arpdemux != 0);
	err = an_demuxkey_add(arpdemux, 0, sizeof key, &key, &mask);
	assert(err == 0);

	pbuf = an_pbuf_create(&arpentry->arpreqbuf_data, AN_PBUF_ETH, 0, 0);
	assert(pbuf != 0);
	arpentry->arpreqbuf = pbuf;

	memset(&attr, 0, sizeof attr);
	sprintf(attr.proto.data, "if%d", ifinfo[iface].id);
	attr.proto.len = strlen(attr.proto.data);
	memcpy(attr.addr.eth.daddr, bcastaddr, sizeof(bcastaddr));
	attr.addr.eth.type = htons(ETHERTYPE_ARP);
	attr.cpu.ttype = AN_TH_IMPLICIT;
	attr.cpu.threads.im.maxthreads = 1;
	attr.bufs[0] = pbuf;
	chan = an_inchan_create(&arpentry->reqinchan_data, flow, arpdemux,
				&attr.proto, &attr.addr, attr.bufs, 1,
				&attr.cpu, ee_arprequest, (void *) iface);
	assert(chan != 0);
	arpentry->reqinchan = chan;
#ifdef  VERBOSE
	printf("%p: ARP Request inchannel for %s\n", chan, attr.proto.data);
#endif

	/*
	 * Setup in channel for direct (arp) packets. These are packets
	 * we have to reply to.
	 * First, create a demux to match ARP requests.
	 */
	memset(&key, 0, sizeof key);
	memset(&mask, 0, sizeof mask);
	key.arp_hrd = htons(ARPHRD_ETHER);
	mask.arp_hrd = ~0;
	key.arp_pro = htons(ETHERTYPE_IP);
	mask.arp_pro = ~0;
	key.arp_op = htons(ARPOP_REQUEST);
	mask.arp_op = ~0;
	memcpy(&key.arp_tpa, &ipaddr->s_addr, sizeof key.arp_tpa);
	memset(&mask.arp_tpa, ~0, sizeof mask.arp_tpa);

	arpdemux = an_demuxkey_create(&arpentry->dreqdemux_data, flow);
	assert(arpdemux != 0);
	err = an_demuxkey_add(arpdemux, 0, sizeof key, &key, &mask);
	assert(err == 0);
	
	pbuf = an_pbuf_create(&arpentry->arpdreqbuf_data, AN_PBUF_ETH, 0, 0);
	assert(pbuf != 0);
	arpentry->arpdreqbuf = pbuf;

	memset(&attr, 0, sizeof attr);
	sprintf(attr.proto.data, "if%d", ifinfo[iface].id);
	attr.proto.len = strlen(attr.proto.data);
	memcpy(attr.addr.eth.daddr,
	       arpentry->ethaddr, sizeof(arpentry->ethaddr));
	attr.addr.eth.type = htons(ETHERTYPE_ARP);
	attr.cpu.ttype = AN_TH_IMPLICIT;
	attr.cpu.threads.im.maxthreads = 1;
	attr.bufs[0] = pbuf;
	chan = an_inchan_create(&arpentry->dreqinchan_data, flow, arpdemux,
				&attr.proto, &attr.addr, attr.bufs, 1,
				&attr.cpu, ee_arprequest, (void *) iface);
	assert(chan != 0);
	arpentry->dreqinchan = chan;
#ifdef  VERBOSE
	printf("%p: ARP Direct request inchannel for %s\n",
	       chan, attr.proto.data);
#endif
	
	/*
	 * Setup a broadcast output channel to do ARP requests.
	 */
	memset(&attr, 0, sizeof attr);
	sprintf(attr.proto.data, "if%d", ifinfo[iface].id);
	attr.proto.len = strlen(attr.proto.data);

	memcpy(attr.addr.eth.daddr, bcastaddr, sizeof(bcastaddr));
	attr.addr.eth.type = htons(ETHERTYPE_ARP);
	
	chan = an_outchan_create(&arpentry->reqchan_data,
				 an_flow_current(),
				 &attr.proto, &attr.addr, 0);
	assert(chan != 0);
	arpentry->reqchan = chan;
#ifdef  VERBOSE
	printf("%p: ARP broadcast outchannel for %s\n", chan, attr.proto.data);
#endif

	/*
	 * Setup a wildcard input channel to capture arp replies.
	 */
	memset(&key, 0, sizeof key);
	memset(&mask, 0, sizeof mask);
	key.arp_hrd = htons(ARPHRD_ETHER);
	mask.arp_hrd = ~0;
	key.arp_pro = htons(ETHERTYPE_IP);
	mask.arp_pro = ~0;
	key.arp_op = htons(ARPOP_REPLY);
	mask.arp_op = ~0;

	arpdemux = an_demuxkey_create(&arpentry->repdemux_data, flow);
	assert(arpdemux != 0);
	err = an_demuxkey_add(arpdemux, 0, sizeof key, &key, &mask);
	assert(err == 0);

	pbuf = an_pbuf_create(&arpentry->arprepbuf_data, AN_PBUF_ETH, 0, 0);
	assert(pbuf != 0);
	arpentry->arprepbuf = pbuf;

	memset(&attr, 0, sizeof attr);
	sprintf(attr.proto.data, "if%d", ifinfo[iface].id);
	attr.proto.len = strlen(attr.proto.data);
	memcpy(attr.addr.eth.daddr,
	       arpentry->ethaddr, sizeof(arpentry->ethaddr));
	attr.addr.eth.type = htons(ETHERTYPE_ARP);
	attr.cpu.ttype = AN_TH_IMPLICIT;
	attr.cpu.threads.im.maxthreads = 1;
	attr.bufs[0] = pbuf;
	chan = an_inchan_create(&arpentry->repinchan_data, flow, arpdemux,
				&attr.proto, &attr.addr, attr.bufs, 1,
				&attr.cpu, ee_arpreply, (void *) iface);
	assert(chan != 0);
	arpentry->repinchan = chan;
#ifdef  VERBOSE
	printf("%p: ARP Reply inchannel for %s\n", chan, attr.proto.data);
#endif

	arpentry->initialized = 1;
	return 0;
}

/*
 * Arp for an IP address on a specific interface.
 */
int
arp_arpfor(int iface, struct in_addr *ipaddr, unsigned char *ethaddr)
{
	struct arpentry	*arpentry = &arptab[iface];
	an_pbuf_t	*pbuf;
	struct ether_arp *arp;
	int		i;
	
	if (!arpready) {
		printf(__FUNCTION__ ": Arp module not initialized\n");
		return -1;
	}

	if (!arpentry->initialized) {
		printf(__FUNCTION__ ": IFACE %d not configured\n", iface);
		return -1;
	}

	/*
	 * Only one per interface for now!
	 */
	an_mutex_lock(arpentry->arplock);

	/*
	 * Construct an arp request packet and send it on the
	 * broadcast outchannel for this interface.
	 */
	pbuf = an_pbuf_create(0, AN_PBUF_ETH, 0, 0);
	assert(pbuf);
	arp = (struct ether_arp *) an_pbuf_data(pbuf);

	arp->ea_hdr.ar_hrd = htons(ARPHRD_ETHER);
	arp->ea_hdr.ar_pro = htons(ETHERTYPE_IP);
	arp->ea_hdr.ar_hln = ETHER_ADDR_LEN;
	arp->ea_hdr.ar_pln = sizeof (struct in_addr);
	arp->ea_hdr.ar_op  = htons(ARPOP_REQUEST);

	/* Him */
	memset(arp->arp_tha, 0, ETHER_ADDR_LEN);
	memcpy(arp->arp_tpa, ipaddr, sizeof(struct in_addr));
	/* Me */
	memcpy(arp->arp_sha, arpentry->ethaddr, ETHER_ADDR_LEN);
	memcpy(arp->arp_spa, &arpentry->ipaddr, sizeof (struct in_addr));
	an_pbuf_setsize(pbuf, sizeof(struct ether_arp), 0);

	for (i = 0; i < 5; i++) {
		an_timespec_t ts;

		/*
		 * Wait for reply (max of 10ms)
		 */
		an_gettime(&ts);
		if ((ts.tv_nsec += 10000000) >= 1000000000) {
			ts.tv_nsec -= 1000000000;
			ts.tv_sec++;
		}

		an_mutex_lock(arpentry->waitlock);
		memset(arpentry->arpethaddr, 0, sizeof(arpentry->arpethaddr));
		arpentry->waiting_for_reply = 1;

		an_outchan_send(arpentry->reqchan, pbuf);

		an_cond_timedwait(arpentry->waitcond, arpentry->waitlock, &ts);
		if (!arpentry->waiting_for_reply) {
			an_mutex_unlock(arpentry->waitlock);
			memcpy(ethaddr, arpentry->arpethaddr,
			       sizeof(arpentry->arpethaddr));
			an_pbuf_destroy(pbuf);
			an_mutex_unlock(arpentry->arplock);
			return 0;
		}
		an_mutex_unlock(arpentry->waitlock);

		/* paranoia */
		while (an_pbuf_isbusy(pbuf))
			;
	}

	an_pbuf_destroy(pbuf);
	an_mutex_unlock(arpentry->arplock);
	return -1;
}

static void
swapn(unsigned char a[], unsigned char b[], int n)
{
	int i;
	unsigned char t;

	for (i = 0; i < n; i++) {
		t = a[i];
		a[i] = b[i];
		b[i] = t;
	}
}

#ifdef VERBOSE
/* The address we are passed is in network order. */
static char *
iptoa(unsigned long x)
{
	static char buf[4*4];		/* nnn.nnn.nnn.nnnz */
	unsigned char *p = (unsigned char *)&x;

	sprintf(buf, "%d.%d.%d.%d", p[0], p[1], p[2], p[3]);
	return buf;
}
#endif

/*
 * Handle an ARP request.
 */
static void
ee_arprequest(an_pbuf_t *pbuf, void *arg, int msize)
{
	int iface = (int) arg;
	struct arpentry	*arpentry = &arptab[iface];
	int psize = an_pbuf_size(pbuf);
	void *data = an_pbuf_data(pbuf);
	struct ether_arp *arp;
	struct in_addr ip;
	an_error_t err;

	if (psize < sizeof *arp) {
#ifdef VERBOSE
		printf("yow! ARP packet too small (%d < %d)\n",
		       psize, sizeof *arp);
#endif
		goto done;
	}

	arp = (struct ether_arp *)data;
	ip.s_addr = *(unsigned long *)arp->arp_tpa;
	assert(ntohs(arp->arp_hrd) == ARPHRD_ETHER);
	assert(ntohs(arp->arp_pro) == ETHERTYPE_IP);
	assert(ntohs(arp->arp_op) == ARPOP_REQUEST);
	assert(ip.s_addr == arpentry->ipaddr.s_addr);

	/* Send the reply. */
	arp->arp_op = htons(ARPOP_REPLY);
	swapn(arp->arp_spa, arp->arp_tpa, sizeof arp->arp_tpa);
	swapn(arp->arp_sha, arp->arp_tha, sizeof arp->arp_tha);
	memcpy(arp->arp_sha, arpentry->ethaddr, sizeof arpentry->ethaddr);

#ifdef VERBOSE
	printf("sending ARP reply to %s (%02x:%02x:%02x:%02x:%02x:%02x)\n",
	       iptoa(*(unsigned int *)arp->arp_tpa),
	       arp->arp_tha[0], arp->arp_tha[1], arp->arp_tha[2],
	       arp->arp_tha[3], arp->arp_tha[4], arp->arp_tha[5]);
#endif

	/*
	 * Got a legit packet to send.
	 */
	if (memcmp(arpentry->outaddr, arp->arp_tha,
		   sizeof arpentry->outaddr) != 0) {
		struct {
			an_chanaddr_t addr;
			an_chanproto_t proto;
		} attr;

		if (arpentry->outchan)
			an_outchan_destroy(arpentry->outchan);

		memset(&attr, 0, sizeof attr);
		sprintf(attr.proto.data, "if%d", ifinfo[iface].id);
		attr.proto.len = strlen(attr.proto.data);

		memcpy(arpentry->outaddr, arp->arp_tha,
		       sizeof arpentry->outaddr);

		memcpy(attr.addr.eth.daddr, arp->arp_tha,
		       sizeof arpentry->outaddr);
		attr.addr.eth.type = htons(ETHERTYPE_ARP);
		
		arpentry->outchan =
			an_outchan_create(&arpentry->outchan_data,
					  an_flow_current(),
					  &attr.proto, &attr.addr, 0);
		assert(arpentry->outchan != 0);
	}
	err = an_outchan_send(arpentry->outchan, pbuf);
	assert(err == 0);
done:
	an_pbuf_return(pbuf);
}

/*
 * Handle a reply to an ARP request we sent out.
 */
static void
ee_arpreply(an_pbuf_t *pbuf, void *arg, int msize)
{
	int iface = (int) arg;
	struct arpentry	*arpentry = &arptab[iface];
	int psize = an_pbuf_size(pbuf);
	void *data = an_pbuf_data(pbuf);
	struct ether_arp *arp;
	struct in_addr ip;

	if (psize < sizeof *arp) {
#ifdef VERBOSE
		printf("yow! ARP packet too small (%d < %d)\n",
		       psize, sizeof *arp);
#endif
		goto done;
	}

	arp = (struct ether_arp *)data;
	ip.s_addr = *(unsigned long *)arp->arp_spa;
	assert(ntohs(arp->arp_hrd) == ARPHRD_ETHER);
	assert(ntohs(arp->arp_pro) == ETHERTYPE_IP);
	assert(ntohs(arp->arp_op) == ARPOP_REPLY);

	/*
	 * Rather simple locking protocol!
	 */
	an_mutex_lock(arpentry->waitlock);
	if (arpentry->waiting_for_reply) {
		memcpy(arpentry->arpethaddr,
		       arp->arp_sha, sizeof(arpentry->arpethaddr));
	}
	arpentry->waiting_for_reply = 0;
	an_mutex_unlock(arpentry->waitlock);

#ifdef VERBOSE
	printf("ARP reply from %s (%02x:%02x:%02x:%02x:%02x:%02x)\n",
	       iptoa(*(unsigned int *)arp->arp_spa),
	       arp->arp_sha[0], arp->arp_sha[1], arp->arp_sha[2],
	       arp->arp_sha[3], arp->arp_sha[4], arp->arp_sha[5]);
#endif

 done:
	an_pbuf_return(pbuf);
}

int
arp_find_ethaddr(int iface, an_ethaddrspec_t *myaddr)
{
	if (!arpready) {
		printf(__FUNCTION__ ": Arp module not initialized\n");
		return -1;
	}

	/*
	 * Find my address.
	 * XXX assumes first interface is ethernet.
	 */
	assert(nif >= iface);
	assert(ifinfo[iface].type == ANIF_ETH);
	memcpy(myaddr->daddr,
	       ifinfo[iface].addr.eth.daddr, sizeof myaddr->daddr);

	return 0;
}

#ifdef ARPCACHE
/*
 * Lookup an IP address in the cache.
 * Returns 1 if found, cacheix will point to entry.
 * Returns 0 if not found, cacheix points to most available entry.
 */
static int
arp_cache_find(int iface, struct in_addr *ipaddr, unsigned char *ethaddr,
	       int *cacheix)
{
	int ix, fix, useix = -1;
	struct arpentry *at;
	struct arpcacheentry *ae;

	at = &arptab[iface];
	fix = ix = at->cachelastused;
	ae = at->cache;
	do {
		if (ae->valid &&
		    memcmp(&ae->ipaddr, ipaddr, sizeof(struct in_addr)) == 0 &&
		    (ethaddr == 0 ||
		     memcmp(ae->ethaddr, ethaddr, ETHER_ADDR_LEN) == 0)) {
			*cacheix = ix;
			return 1;
		}
		if (useix < 0 && !ae->valid)
			useix = ix;

		if (++ix == MAXCENTRIES) {
			ix = 0;
			ae = at->cache;
		}
	} while (ix != fix);

	if (useix < 0)
		useix = (++fix == MAXCENTRIES) ? 0 : fix;
	*cacheix = useix;
	return 0;
}

int
arp_cache_add(int iface, struct in_addr *ipaddr, unsigned char *ethaddr)
{
	struct arpentry *at = &arptab[iface];
	struct arpcacheentry *ae;
	int i;

	an_mutex_lock(at->arplock);
	if (arp_cache_find(iface, ipaddr, ethaddr, &i)) {
		an_mutex_unlock(at->arplock);
		return 0;
	}

	ae = &at->cache[i];
	memcpy(&ae->ipaddr, ipaddr, sizeof(struct in_addr));
	memcpy(ae->ethaddr, ethaddr, ETHER_ADDR_LEN);
	ae->valid = 1;

	at->cachelastused = i;
	an_mutex_unlock(at->arplock);
	return 0;
}

int
arp_cache_remove(int iface, struct in_addr *ipaddr, unsigned char *ethaddr)
{
	struct arpentry *at = &arptab[iface];
	int i;

	an_mutex_lock(at->arplock);
	if (!arp_cache_find(iface, ipaddr, ethaddr, &i)) {
		an_mutex_unlock(at->arplock);
		return -1;
	}

	at->cache[i].valid = 0;

	at->cachelastused = i;
	an_mutex_unlock(at->arplock);
	return 0;
}

int
arp_cache_lookup(int iface, struct in_addr *ipaddr, unsigned char *ethaddr)
{
	struct arpentry *at = &arptab[iface];
	int i;

	an_mutex_lock(at->arplock);
	if (!arp_cache_find(iface, ipaddr, 0, &i)) {
		an_mutex_unlock(at->arplock);
		return -1;
	}

	memcpy(ethaddr, at->cache[i].ethaddr, ETHER_ADDR_LEN);

	at->cachelastused = i;
	an_mutex_unlock(at->arplock);
	return 0;
}
#endif
