/*
 * Copyright (c) 1999, 2000 The University of Utah and the Flux Group.
 * All rights reserved.
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */
#include <stdio.h>
#include <nodeos/an_flow.h>

static an_flowstats_t totstats;
static void print_stats(an_flowstats_t *stats,
			int verbose, const char *str, int interval);
static void diff_stats(an_flowstats_t *new, an_flowstats_t *old,
		       an_flowstats_t *diff);
static void add_stats(an_flowstats_t *stats);

an_error_t
moab_flow_printstats(an_flow_t *flow, int verbose,
		     const char *str, int interval, int accum)
{
	an_flowstats_t stats;
	an_error_t err;

	
	if (flow == 0) {
		print_stats(&totstats, verbose, str, interval);
		memset(&totstats, 0, sizeof totstats);
		return 0;
	}

	err = an_flow_stats(flow, &stats);
	if (err == 0) {
		print_stats(&stats, verbose, str, interval);
		if (accum)
			add_stats(&stats);
	}

	return err;
}

an_error_t
moab_flow_printstatsdelta(an_flow_t *flow, int verbose, const char *str,
			  int interval, void **cookie)
{
	an_flowstats_t *statp, stats, dstats;

	if (cookie == 0)
		return -1;

	statp = *cookie;
	if (statp == 0) {
		statp = an_mempool_alloc(an_mempool_current(), sizeof *statp);
		if (statp == 0)
			return -1;
		memset(statp, 0, sizeof *statp);
		*cookie = statp;
	}
	if (an_flow_stats(flow, &stats) != 0)
		return -1;
	diff_stats(&stats, statp, &dstats);
	print_stats(&dstats, verbose, str, interval);
	*statp = stats;

	return 0;
}

static void
print_stats(an_flowstats_t *stats, int verbose, const char *str, int interval)
{
	if (stats->cpuused == 0) {
		if (verbose)
			printf("%s: idle\n", str);
		return;
	}

	if (stats->cpuunits == NODEOS_CPU_CYCLES)
		printf("%s: %qu cycles of CPU", str, stats->cpuused);
	else
		printf("%s: %qu microseconds of CPU", str, stats->cpuused);
	if (interval > 0)
		printf(" (%u cycles/sec)",
		       (unsigned)(stats->cpuused / interval));
	if (stats->memused)
		printf(", %qu bytes of memory", stats->memused);
	printf("\n");

	if (verbose)
		printf("  %d thread activations (%d current, %d max)\n",
		       (int)stats->activations, (int)stats->curthreads,
		       (int)stats->maxthreads);
}

static void
diff_stats(an_flowstats_t *new, an_flowstats_t *old, an_flowstats_t *diff)
{
	diff->curthreads = new->curthreads;
	diff->maxthreads = new->maxthreads;
	diff->cpuunits = new->cpuunits;
	diff->activations = new->activations - old->activations;
	diff->cpuused = new->cpuused - old->cpuused;
	diff->memused = new->memused;
}

static void
add_stats(an_flowstats_t *stats)
{
	totstats.curthreads = stats->curthreads;
	totstats.maxthreads = stats->maxthreads;
	totstats.cpuunits = stats->cpuunits;
	totstats.activations += stats->activations;
	totstats.cpuused += stats->cpuused;
	totstats.memused = stats->memused;
}
