/*
 * Copyright (c) 1999, 2000 The University of Utah and the Flux Group.
 * All rights reserved.
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */

/*
 * Pbuffer cache.
 *
 * Keeps a list of available pbuffers for quick allocation.
 */
#include <nodeos/lib/pbcache.h>
#include <nodeos/an_mempool.h>
#include <nodeos/an_thread.h>

/*
 * Create a cache with the indicated number of pbuffers.
 */
pbcache_t *
pbcache_create(int nbufs, void (*init)(an_pbuf_t *))
{
	pbcache_t *cache;
	an_mutex_t *lock;
	an_cond_t *cond;
	an_pbuf_t *bufptr;
	size_t size;
	int i;

	size = sizeof *cache + nbufs * sizeof(struct pbcache_buf);
	cache = an_mempool_alloc(an_mempool_current(), size);
	if (cache == 0)
		return 0;

	memset(cache, 0, size);

	lock = an_mutex_create(&cache->lock);
	if (lock == 0) {
		an_mempool_free(an_mempool_current(), cache, size);
		return 0;
	}
	assert(lock == &cache->lock);
	cond = an_cond_create(&cache->cond);
	if (cond == 0) {
		an_mutex_destroy(&cache->lock);
		an_mempool_free(an_mempool_current(), cache, size);
		return 0;
	}
	assert(cond == &cache->cond);

	an_mutex_lock(&cache->lock);
	cache->memsize = size;
	for (i = 0; i < nbufs; i++) {
		bufptr = an_pbuf_create(&cache->list[i].buf,
					AN_PBUF_ETH, 0, 0);
		if (bufptr == 0) {
			an_mutex_unlock(&cache->lock);
			pbcache_destroy(cache, 0);
			return 0;
		}
		assert(bufptr == &cache->list[i].buf);

		if (cache->tail == 0)
			cache->tail = &cache->list[i];
		cache->list[i].next = cache->head;
		cache->head = &cache->list[i];
		cache->nbufs++;
	}

	if (init != 0) {
		for (i = 0; i < nbufs; i++)
			init(&cache->list[i].buf);
	}

	an_mutex_unlock(&cache->lock);
	return cache;
}

/*
 * Destroy a cache.  Returns the number of elements not destroyed.
 * A return value of zero indicates all pbufs were destroyed as well
 * as the cache object.  Any non-zero return value indicates that the
 * cache was not destroyed.  Repeated calls can be made on the same
 * cache until a zero value is returned.
 */
int
pbcache_destroy(pbcache_t *cache, void (*deinit)(an_pbuf_t *))
{
	int destroyed;
	struct pbcache_buf *bp, **bpp;

	an_mutex_lock(&cache->lock);
	destroyed = 0;
	bpp = &cache->head;
	while ((bp = *bpp) != 0) {
		if (an_pbuf_isbusy(&bp->buf)) {
			bpp = &bp->next;
			continue;
		}
		if (deinit != 0)
			deinit(&bp->buf);
		an_pbuf_destroy(&bp->buf);
		*bpp = bp->next;
		destroyed++;
	}
	if (destroyed != cache->nbufs) {
		cache->nbufs -= destroyed;
		an_mutex_unlock(&cache->lock);
		return cache->nbufs;
	}
	an_mutex_unlock(&cache->lock);
	an_mutex_destroy(&cache->lock);
	an_mempool_free(an_mempool_current(), cache, cache->memsize);
	return 0;
}

an_pbuf_t *
pbcache_alloc(pbcache_t *cache, int dowait)
{
	an_pbuf_t *pbuf = 0;
	struct pbcache_buf *this, **prev;

	an_mutex_lock(&cache->lock);
	cache->stats.gets++;
	if (cache->head == 0)
		cache->stats.empty++;
	for (prev = &cache->head; (this = *prev) != 0; prev = &this->next) {
		cache->stats.looks++;
		if (!an_pbuf_isbusy(&this->buf)) {
			*prev = this->next;
			if (cache->tail == this)
				cache->tail = this->next;
			pbuf = &this->buf;
			cache->stats.gots++;
			break;
		}
	}
	if (pbuf == 0 && dowait) {
		cache->stats.waits++;
		while ((this = cache->head) == 0) {
			cache->iswaiting = 1;
			an_cond_wait(&cache->cond, &cache->lock);
		}
		cache->head = this->next;
		if (cache->tail == this)
			cache->tail = this->next;

		pbuf = &this->buf;
		an_mutex_unlock(&cache->lock);
#ifdef notyet
		an_pbuf_wait(pbuf);
		assert(!an_pbuf_isbusy(pbuf));
#else
		while (an_pbuf_isbusy(pbuf))
			an_thread_yield();
#endif
	} else
		an_mutex_unlock(&cache->lock);

	return pbuf;
}

void
pbcache_free(pbcache_t *cache, an_pbuf_t *pbuf)
{
	struct pbcache_buf *bl = (struct pbcache_buf *)pbuf;

	an_mutex_lock(&cache->lock);
	bl->next = 0;
	if (cache->head == 0) {
		cache->head = bl;
		if (cache->iswaiting) {
			an_cond_broadcast(&cache->cond);
			cache->iswaiting = 0;
		}
	} else
		cache->tail->next = bl;
	cache->tail = bl;
	an_mutex_unlock(&cache->lock);
}

#include <stdio.h>
void
pbcache_dump(pbcache_t *cache)
{
	printf("%d pbufs: ", cache->nbufs);
	printf("%d requests, %d scanned, %d waits, %d failed, %d cache empty\n",
	       cache->stats.gets, cache->stats.looks, cache->stats.waits,
	       cache->stats.gets - cache->stats.gots, cache->stats.empty);
}

