/*
 * Copyright (c) 2000 The University of Utah and the Flux Group.
 * All rights reserved.
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */

/*
 * NodeOS Filesystem API
 */
#ifndef _NODEOS_FS_H_
#define _NODEOS_FS_H_

/*
 * Not implemented: 
 *      an_fileMMap()
 *      an_fileMUnmap()
 *      an_fileMSync()
 *      an_fileShmOpen()
 *      an_fileShmUnlink()
 */

#ifdef NODEOS_SPEC_COMPAT
#define an_File an_file_t
#define an_Offset an_offset_t
#define an_Whence an_whence_t
#define an_ACD an_acd_t
#define an_Size an_size_t
#define an_Stat an_stat_t
#define an_Prot an_prot_t
#define an_fileClose an_file_close
#define an_fileFsync an_file_fsync
#define an_fileTruncate an_file_truncate
#define an_fileLseek an_file_lseek
#define an_fileMkDir an_file_mkdir
#define an_fileRead an_file_read
#define an_fileRename an_file_rename
#define an_fileWrite an_file_write
#define an_fileUnlink an_file_unlink
#define an_fileOpen an_file_open
#define an_fileStat an_file_stat
#define an_fileFstat an_file_fstat
#endif

#include <nodeos/an_error.h>
#include <nodeos/an_time.h>
#include <nodeos/impl/an_fs.h>


/*
 * Close the open file referenced by 'file'. 
 */
an_error_t an_file_close(an_file_t *file);

/*
 * Cause all modified data and attributes of 'file' to be
 * moved to persistent storage.  
 */
an_error_t an_file_fsync(an_file_t *file);

/*
 * Extend or truncate to 'length' the file referenced
 * by 'file'.
 */
an_error_t an_file_truncate(an_file_t *file, an_offset_t offset);

/*
 * Seek pointer for the file referenced by 'file' to 'offset' 
 * according to the directive 'whence'.  The allowed values
 * for 'whence' are: 
 *
 * AN_SEEK_SET - Offset is set to 'offset' bytes 
 * AN_SEEK_CUR - Offset is set to the current position plus 'offset'
 *               bytes.
 * AN_SEEK_END - Offset is set to the size of the file plus 'offset' 
 *               bytes.  
 */
an_error_t an_file_lseek(an_file_t *file, an_offset_t length,
			 an_whence_t whence, an_offset_t *retoff);

/*
 * Create the directory given by 'path' with the access permissions
 * specified by 'acd'.
 */
an_error_t an_file_mkdir(const char *path, an_acd_t *acd);

/*
 * Read 'nbytes' of data from the file referenced by 'file' into 
 * the buffer pointed to by 'buf'. The actual number of bytes read is
 * returned in 'retlen'. 
 */
an_error_t an_file_read(an_file_t *file, an_size_t nbytes, 
			void *buf, an_size_t *retlen);

/*
 * Rename the file named 'from' to new name 'to'.
 */
an_error_t an_file_rename(const char *from, const char *to);


/*
 * Write 'nbytes of data to the file referenced by 'file' form
 * the buffer pointed to by 'buf'. The actual number of bytes written 
 * is returned in 'retval'.
 */
an_error_t an_file_write(an_file_t *file, an_size_t length, void *buf, 
			 an_size_t *retval);


/*
 * Remove the file named by 'path' from its directory. In a departure from
 * POSIX semantics, if 'pasth' is an empty directory, the directory
 * is removed. 
 */
an_error_t an_file_unlink(const char *path);


/*
 * Open the file specified by 'path' for reading and/or writing
 * according to 'flags'. The new file object is returned in the 
 * location specified by 'file'. The flags are specified by or-ing 
 * an appropriate subset of the following values:
 * 
 * AN_O_RDONLY - open for readine only
 * AN_O_WRONLY - open for writing only
 * AN_O_RDWR   - open for reading and writing 
 * AN_O_APPEND - append on each write 
 * AN_O_CREAT  - create 'file' if it does not exist
 * AN_O_TRUNC  - truncate size to 0 bytes 
 * AN_O_EXCL   - error if create and file exists
 *
 */
an_error_t an_file_open(const char *path, int flags, an_acd_t *acd, 
			an_file_t *retval);


/*
 * Obtain information about the file named by 'path' and store
 * that information in the buffer pointed to by 'retval'. 
 */
an_error_t an_file_stat(const char *path, an_stat_t *retval);


/*
 * Obtain information about the open file referenced by 'file' and
 * store that infomration in the buffer pointed to by 'retval'. 
 * Otherwise, fstat() behaves identically to the 'stat' function.
 */
an_error_t an_file_fstat(an_file_t *file, an_stat_t *retval);

#endif 
