/*
 * Copyright (c) 1999, 2000 The University of Utah and the Flux Group.
 * All rights reserved.
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */

/*
 * Informational NodeOS API calls
 */
#ifndef _NODEOS_INFO_H_
#define _NODEOS_INFO_H_

#ifdef NODEOS_SPEC_COMPAT
#define an_InfoInterface	an_infoif_t
#define an_InfoIPv4		an_infoipv4_t
#define an_InfoGetIfs		an_info_getifs
#define an_InfoGetHostname	an_info_gethostname
#define an_InfoSetHostname	an_info_sethostname
#endif

#include <nodeos/an_impl.h>
#include <nodeos/an_chan.h>
#include <nodeos/impl/an_info.h>

#define MAX_IFNAME_LEN	8

/*
 * Network interface descriptor.
 * One per hardware interface.
 *
 * ``type'' is the interface type; e.g. ethernet, ATM,
 * ``id'' is a NodeOS assigned unique logical interface number,
 * ``pname'' is the implementation-specific string identifying the interface,
 * ``addr'' identifies this node on the particular interface,
 * ``bandwidth'' is the available bandwidth for the interface in bits/second.
 *
 * The format of address information is interface-specific but will always
 * include a unique address for the node.
 *
 * When creating a channel with proto-spec "if<id>" (e.g., "if0"), the
 * returned address information can be used directly to receive all packets
 * destined for this node on the given interface.  Such packets will have
 * their interface-specific headers stripped before delivery to the channel
 * receive function.  Hence an EE can receive raw packet payloads without
 * needing to know the underlying address format for the hardware.
 * 
 * When creating a channel with proto-spec given by the interface pname
 * (e.g., "eth0"), the channel by default will deliver all packets destined
 * for the node complete with all headers intact.
 */
typedef enum { ANIF_ANY, ANIF_ETH, ANIF_ATM } an_infoiftype_t;

typedef struct an_infoif {
	an_infoiftype_t type;		/* interface type */
	int id;				/* logical interface ID */
	char pname[NODEOS_MAXIFNAME];	/* "physical" name; e.g. eth0, atm1 */
	an_chanaddr_t addr;		/* address info */
	unsigned int bandwidth;		/* available bandwidth */
} an_infoif_t;

typedef struct an_infoipv4 {
	int ifid;		/* logical interface ID */
	int ipaddr;		/* IPv4 */
	/* XXX other ipv4 per-interface information? */
} an_infoipv4_t;

typedef struct an_infocpu {
	int ncpu;		/* number of CPUs */
	unsigned int cps;	/* normalized cycles/second per CPU */
} an_infocpu_t;

/*
 * Retrieve information about network interfaces.
 *
 * ``info'' is an array of an_ifinfo structs to be filled in
 * ``count'' the size of the array on call, the number filled in on return
 *
 * Returns 0 on success, an error otherwise.
 */
an_error_t an_info_getifs(an_infoif_t info[], int *count);

/*
 * Retrieve information about CPUs on the node.
 *
 * ``cpuinfo'' is a pointer to a infocpu_t struct to fill in.
 *
 * Returns 0 on success, an error otherwise.
 */
an_error_t an_info_getcpu(an_infocpu_t *cpuinfo);

/*
 * Return the hostname of this node.  Returns a read-only constant.
 */
const char* an_info_gethostname(void);

const char* an_info_getoptlist(void);

/*
 * Set the hostname of this node.  The nodeos will
 * make a copy of the argument.
 */
an_error_t an_info_sethostname(const char* name);



#endif /* _NODEOS_INFO_H_ */
