/*
 * Copyright (c) 1999, 2000 The University of Utah and the Flux Group.
 * All rights reserved.
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */

/*
 * NodeOS Thread API
 */
#ifndef _NODEOS_THREAD_H_
#define _NODEOS_THREAD_H_

#ifdef NODEOS_SPEC_COMPAT
#define an_Thread an_thread_t
#define an_threadId an_thread_current
#define an_threadStart an_thread_start
#define an_threadExit an_thread_exit
#define an_threadYield an_thread_yield
#define an_threadSleep an_thread_sleep
#define an_threadSetData an_thread_setdata
#define an_threadGetData an_thread_getdata
#define an_threadInterrupt an_thread_intr
#define an_threadBlockInterrupts an_thread_blockintr
#define an_threadUnblockInterrupts an_thread_unblockintr
#define an_threadSetPrio an_thread_setprio
#define an_threadGetPrio an_thread_getprio
#endif

#include <nodeos/an_error.h>
#include <nodeos/an_impl.h>
#include <nodeos/an_flow.h>
#include <nodeos/impl/an_thread.h>

/*
 * Thread priorities are scheduler specific but for general use
 * we define three distinguished values.
 *
 * EXCLUSIVE effectively disables preemption on the current processor,
 * allowing the thread to run til it volentarily yields or blocks.
 *
 * NORMAL is a catch-all class in which all threads are considered
 * equal by the scheduler w.r.t. its algorithm (FIFO, RR, whatever).
 *
 * IDLE will run the thread only if there is nothing else to run.
 */
#define	AN_THREAD_PRIO_EXCLUSIVE	0x0000FFFF
#define AN_THREAD_PRIO_IDLE		0
#define AN_THREAD_PRIO_NORMAL		(AN_THREAD_PRIO_EXCLUSIVE/2)

/*
 * Return the identity of the currently running thread.
 */
an_thread_t *an_thread_current(void);

/*
 * Create a thread object in the indicated memory.
 * The new thread's stack is described by the stackptr/stacksize params.
 * Until the thread object is associated with a flow, no operation other
 * than thread_destroy can be performed on it.
 */
an_thread_t *an_thread_create(void *mem, void *stackptr, size_t stacksize);

/*
 * Destroys a thread object.
 * This call will fail if the thread is associated with a flow and running;
 * i.e., this is *not* a thread_kill call!  This call will free the thread's
 * resources.  After the call returns, the memory previously in use
 * by the thread object and the thread stack can be reused or freed.
 */
an_error_t an_thread_destroy(an_thread_t *thread);

/*
 * Start a specific thread running in the context of the given flow.
 * The new thread starts in the indicated function with the indicated argument.
 * The thread passed should not currently be associated with any flow.
 * Returns zero if the thread was launched, an error otherwise.
 */
an_error_t an_thread_start(an_thread_t *thread, an_flow_t *flow,
			   void (*f)(void *arg), void *arg);

/*
 * Exit the current thread.
 * Doesn't return (duh!)
 */
void an_thread_exit(void);

/*
 * Yield the processor to any other runnable thread.
 */
void an_thread_yield(void);

/*
 * Sleep for the indicated amount of time.
 * Currently blocks uninterruptibly.  This is bad.
 *
 * XXX it isn't clear what we should do after a thread interrupt handler
 * has been run.  Should the thread return to the sleep for the duration
 * of the indicated delay?  Should thread sleep return an EINTR sort of
 * error requiring the caller to recompute the delay and try again?
 * The latter case could be simplified by using an absolute time for the
 * sleep ala pthread_cond_timedwait.  This way the thread could loop on
 * the sleep until it returns EINVAL indicating that the sleep time has
 * expired.
 */
void an_thread_sleep(an_timespec_t *delay);

/*
 * Get/set thread specific data for the current thread.
 * Only one item per thread.  Subsequent setdata calls overwrite the
 * previous value.
 */
void *an_thread_getdata(void);
void an_thread_setdata(void *data);

/*
 * Make a thread run an interrupt handler.
 * If the specified thread is running, sleeping or blocked on a condition
 * variable and the thread allows interrupts, the thread is redirected to
 * run an interrupt handler.  How the interrupt handler is setup, what
 * its arguments are, and what the thread can do in the handler context
 * are all implementation specific.
 */
void an_thread_intr(an_thread_t *thread);

/*
 * Block/unblock thread interrupts for the current thread.
 */
void an_thread_blockintr(void);
void an_thread_unblockintr(void);

/*
 * Get/set CPU priority for an arbitrary thread.
 */
an_error_t an_thread_getprio(an_thread_t *thread, int *out_prio);
an_error_t an_thread_setprio(an_thread_t *thread, int prio);

/*
 * Find a specific thread's stack state.  
 * Must work even on "dead" threads.
 * XXX probably should be in impl.h?
 */
an_error_t an_thread_stackinfo(an_thread_t *thread, void **base, size_t *size,
			       void **ptr);

#endif /* _NODEOS_THREAD_H_ */
