/*
 * Copyright (c) 1999, 2000, 2001 The University of Utah and the Flux Group.
 * All rights reserved.
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */

/*
 * Implementation specific types and constants for the Utah OSKit-based NodeOS.
 *
 * Note that these are not machine dependencies, this file is used for both
 * the x86 and ARM versions.
 *
 * Note that this file also currently contains "OS" specific constants and
 * types; i.e., for Linux, FreeBSD and Solaris as well as for OSKit.
 */
#ifndef _NODEOS_IMPL_IMPL_H_
#define _NODEOS_IMPL_IMPL_H_

#ifdef NODEOS_SPEC_COMPAT
#define an_NotifyFunc an_flowterm_func_t
#define an_FlowNotifyFunc an_flowterm_func_t
#define an_FlowStartSpec an_flowinit_t
#define an_MempoolNotifyFunc an_mempoolfull_func_t
#define an_ChannelAddress an_chanaddr_t
#define an_ChannelProtocol an_chanproto_t
#define an_DeliverFunc an_chanrecvfunc_t
#define an_EventHandler an_eventfunc_t
#define an_Schedule an_sched_t
#define an_ThreadNotifyFunc an_threadintr_func_t
#define an_threadSetHandler an_thread_sethandler
#endif

#include <sys/types.h>

#include <nodeos/impl/decls.h>
#include <nodeos/impl/an_stats.h>

/*
 * XXX hack for solaris
 */
#ifdef __svr4__
typedef uint8_t u_int8_t;
typedef uint16_t u_int16_t;
typedef uint32_t u_int32_t;
#endif

/*
 * XXX hack global constants that should be specified in other ways.
 */
#define NODEOS_PAGESIZE		4096	/* memory page size */
#define NODEOS_CHAN_PROTOSIZE	128	/* protocol spec size */
#define NODEOS_CHAN_ADDRSIZE	128	/* peer addressing info size */
#define NODEOS_MAXIFS		16	/* maximum network interfaces */
#define NODEOS_MAXIFNAME	8	/* maximum interface pname length */

#define NODEOS_CHAN_BUFSIZE	2048	/* channel buffer size */
#define NODEOS_CHAN_COPYSIZE	128	/* data size under which copied */

/*
 * Granularity of the event timer in microseconds.
 * Event times are truncated to the nearest such tick.
 */
#define NODEOS_EVENT_TICK	10000

/*
 * Size of memory ``chunks'' managed by mempools.
 */
#define NODEOS_MEMPOOL_CHUNKSIZE	4096

/*
 * Implementation-specific, API-exposed types.
 */

/*
 * An ethernet-style channel address spec.
 * Note: all values should be in network byte order (Big Endian)
 */
typedef struct {
	u_int8_t daddr[6];
	u_int8_t saddr[6];
	u_int16_t type;
} an_ethaddrspec_t;

/*
 * A UDP/IPv4 channel address spec
 * Note: all values should be in network byte order (Big Endian)
 */
typedef struct {
	u_int16_t sport;
	u_int32_t saddr;
	u_int16_t dport;
	u_int32_t daddr;
} an_udp4addrspec_t;

/*
 * A cheezy-hack-clock-style channel address spec
 */
typedef struct {
	u_int32_t freq;
} an_clockaddrspec_t;

/*
 * Implementation-specific, API-exposed types.
 */

/*
 * Flow initialization and termination functions.
 */
struct an_flow;
typedef void (*an_flowinit_func_t)(void *);
typedef void (*an_flowterm_func_t)(struct an_flow *);

/*
 * Called when a mempool is over limit
 */
struct an_mempool;
typedef void (*an_mempoolfull_func_t)(struct an_mempool *, int, int);

/*
 * Channel protocol processing information.
 * Currently implemented as a string; e.g. "ETH/IP".
 */
typedef struct {
	int len;
	char data[NODEOS_CHAN_PROTOSIZE-sizeof(int)];
} an_chanproto_t;

/*
 * Channel peer address information.
 * Channel-type specific.
 *
 * XXX rename to an_chanaddrspec_t 
 */
typedef union {
	an_ethaddrspec_t eth;
	an_udp4addrspec_t udp4;
	an_clockaddrspec_t clock;
	char data[NODEOS_CHAN_ADDRSIZE];
} an_chanaddr_t;

/*
 * Called when a packet arrives on a channel.
 * First argument is the packet received.
 * Second argument is a user-specified argument,
 *   supplied at channel creation time.
 * Third argument is an offset in the packet,
 *   one byte beyond the last byte matched by the channel demuxkey.
 */
struct an_pbuf;
typedef void (*an_chanrecvfunc_t)(struct an_pbuf *, void *, int);

/*
 * Called when an event is triggered.
 * First argument is the event triggered (XXX not yet).
 * Second argument is a user-specified argment,
 *   supplied when the event is created.
 */
typedef void (*an_eventfunc_t)(void *, void *);

/*
 * Not yet
 */
typedef void *an_sched_t;

/*
 * For argc, argv passed to default main.c implementation
 */
typedef struct {
	int argc;
	char **argv;
} an_initargs_t;

/*
 * Thread interrupt handling
 * Includes API call to set the handler and signature for handler.
 */

typedef void (*an_threadintr_func_t)(int, void *);

/* types passed to the notify function */
#define AN_THREAD_TRAP_SEGV 0x11
#define AN_THREAD_TRAP_FPU  0x12  
#define AN_THREAD_INTR	    0x21

void an_thread_sethandler(an_threadintr_func_t handler);

#endif /* _NODEOS_IMPL_IMPL_H_ */
