/*
 * Copyright (c) 1999, 2000 The University of Utah and the Flux Group.
 * All rights reserved.
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */

#ifndef _FLOW_IMPL_H_
#define _FLOW_IMPL_H_

#include <nodeos/an_impl.h>
#include <nodeos/an_resource.h>

#include <nodeos/impl/ani_hnode.h>
#include <nodeos/impl/ani_resource.h>
#include <nodeos/impl/ani_ilock.h>
#include <nodeos/impl/ani_threadpool.h>

struct ani_cred;
struct ani_chan;
struct ani_mempool;

/*
 * The flow: rendezvous point for threads and resources.
 *
 * Every flow is part of the flow hierarchy as well as possibly associated
 * with a mempool.
 */

struct ani_flow {
	struct hnode		me;	/* flow hier link (must be first) */
	struct ani_cred		*cred;		/* credentials */
	struct ani_mempool	*mpool;		/* mempool */
	struct ani_chan		*chanlist;	/* list of channels */
	struct ani_threadpool	tpool;		/* thread pool */
	int			flags;		/* various flags */
	an_flowterm_func_t	termfunc;	/* termination function */
	ani_resource_t		res;		/* reserved resources */
	int			refs;		/* reference count */
	ani_ilock_t		lock;		/* mutual exclusion */
	an_flowstats_t		stats;		/* statistics */
};
typedef struct ani_flow ani_flow_t;

#define FLOW_INUPCALL	1
#define FLOW_DYING	2

#define ani_flow_mempool(f)	((f)->mpool)
#define ani_flow_threadpool(f)	(&(f)->tpool)
#define ani_flow_lock(f)	ani_ilock_lock(&(f)->lock)
#define ani_flow_unlock(f)	ani_ilock_unlock(&(f)->lock)

#define ani_flow_parent(f)	((ani_flow_t *)hnode_parent(&(f)->me))
#define ani_flow_child(f)	((ani_flow_t *)hnode_child(&(f)->me))
#define ani_flow_sibling(f)	((ani_flow_t *)hnode_sibling(&(f)->me))

#define ani_flow_reference(f)	((f)->refs++)
#define ani_flow_release(f)	if (--(f)->refs == 0) ani_flow_cleanup(f)

#ifdef MEMDEBUG
#define assert_ani_flow_valid(f) \
	assert((f) != 0 && (f) != (void *)0xdddddddd), \
	assert(*(unsigned int *)(f) != 0xdddddddd)
#else
#define assert_ani_flow_valid(f)	assert((f) != 0)
#endif

/*
 * Boot-time initialization of the flow component.
 */
void ani_flow_setup(ani_flow_t *rflow);

ani_error_t ani_rootflow_init(ani_flow_t *rflow, struct ani_cred *rcred,
			      struct ani_mempool *rmempool,
			      int nthreads, int stacksize,
			      void **root_threads, void **root_threadstacks,
			      an_flowinit_func_t ifunc, void *initarg);

void ani_rootflow_deinit(ani_flow_t *rflow);

ani_error_t ani_rootflow_startthread(ani_thread_t *thread,
				     ani_thread_func_t func,
				     void *farg1, void *farg2, void *farg3);

/*
 * Return reference to current flow context.
 */
ani_flow_t *ani_flow_current(void);

/*
 * Create a new flow.
 * The new flow is initialized with
 * the indicated parent flow, memory pool and thread scheduler.
 */
ani_error_t ani_flow_init(ani_flow_t *flow, struct ani_cred *cred,
			  an_cpuspec_t *cspec, an_memspec_t *mspec,
			  an_flowinit_func_t initfunc, void *initarg,
			  an_flowterm_func_t termfunc);

/*
 * Destroy the indicated flow.
 * Detach from the flow hierarchy, destroy all child flows and channels.
 * Flow resources are not freed til all flow threads are gone.
 */
ani_error_t ani_flow_deinit(ani_flow_t *flow);

/*
 * Final flow destruction.
 * Removes its memory resources from the associated memory pool
 * and frees the data structures.
 */
void ani_flow_cleanup(ani_flow_t *flow);

void ani_flow_addchannel(ani_flow_t *flow, struct ani_chan *chan);
void ani_flow_removechannel(ani_flow_t *flow, struct ani_chan *chan);

/*
 * Fire up a thread from the flow's threadpool at the specified function.
 */
ani_error_t ani_flow_startthread(ani_flow_t *flow, int inkern,
				 void (*startfunc)(void *, void *),
				 void *farg1, void *farg2);

an_flowstats_t *ani_flow_stats(ani_flow_t *flow);

#endif /* _FLOW_IMPL_H_ */

