/*
 * Copyright (c) 2000 The University of Utah and the Flux Group.
 * All rights reserved.
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */

#ifndef _NODEOS_IMPL_ANI_FS_H_
#define _NODEOS_IMPL_ANI_FS_H_

#include <unistd.h>
#include <stdio.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/uio.h>

#include <nodeos/an_error.h>
#include <nodeos/an_time.h>

typedef union an_acd {
  mode_t st_mode;                       /* protection mode */
} an_acd_t;

typedef size_t an_size_t;
typedef int    an_file_t;
typedef off_t  an_offset_t;
typedef int    an_whence_t;

typedef enum {ANFILE_UNKNOWN, ANFILE_FILE, ANFILE_DIR } an_filetype_t;

typedef struct an_stat {
  an_timespec_t st_atimespec;     /* time of last access */
  an_timespec_t st_mtimespec;     /* time of last data modification */
  an_timespec_t st_ctimespec;     /* time of last file status change */
  an_filetype_t st_type;          /* directory or regular file */
  an_acd_t      st_acd;           /* file access protection mode */
  an_size_t     st_size;          /* file size, in bytes */ 
} an_stat_t;

#define AN_O_RDONLY O_RDONLY
#define AN_O_WRONLY O_WRONLY 
#define AN_O_RDWR   O_RDWR
#define AN_O_APPEND O_APPEND
#define AN_O_CREAT  O_CREAT
#define AN_O_TRUNC  O_TRUNC
#define AN_O_EXCL   O_EXCL


/* inline functions */

NODEOS_FUNC_INLINE an_error_t 
ani_file_close(an_file_t *file) 
{
        return (close(*file) == -1) ? errno : 0;
}

NODEOS_FUNC_INLINE an_error_t 
ani_file_fsync(an_file_t *file)
{
        return (fsync(*file) == -1) ? errno : 0;
}

NODEOS_FUNC_INLINE an_error_t 
ani_file_truncate(an_file_t *file, an_offset_t offset)
{
        return (ftruncate(*file, offset) == -1) ? errno : 0;
}

NODEOS_FUNC_INLINE an_error_t 
ani_file_lseek(an_file_t *file, an_offset_t length, an_whence_t whence,
	       an_offset_t *retoff)
{
	return ((*retoff = lseek(*file, length, whence)) == -1) ? errno : 0;
}

NODEOS_FUNC_INLINE an_error_t 
ani_file_mkdir(const char *path, an_acd_t *acd) 
{
	return (mkdir(path, acd->st_mode) == -1) ? errno : 0;
}

NODEOS_FUNC_INLINE an_error_t 
ani_file_read(an_file_t *file, an_size_t nbytes, void *buf, an_size_t *retlen)
{
	return ((*retlen = read(*file, buf, nbytes)) == -1) ? errno : 0;
}

NODEOS_FUNC_INLINE an_error_t 
ani_file_rename(const char *from, const char *to)
{
        return (rename(from, to) == -1) ? errno : 0;
}

NODEOS_FUNC_INLINE an_error_t 
ani_file_write(an_file_t *file, an_size_t length, void *buf, an_size_t *retval)
{
	return ((*retval = write(*file, buf, length)) == -1) ? errno : 0;
}

NODEOS_FUNC_INLINE an_error_t 
ani_file_unlink(const char *path)
{
	return (unlink(path) == -1) ? errno : 0;
}

NODEOS_FUNC_INLINE an_error_t 
ani_file_open(const char *path, int flags, an_acd_t *acd, an_file_t *retval)
{
         *retval = flags & O_CREAT ? open(path, flags, acd->st_mode) 
	                           : open(path, flags);
	 return (*retval == -1) ? errno : 0;
}

/* Prototypes */

ani_error_t 
ani_file_stat(const char *path, an_stat_t *anstat);

ani_error_t 
ani_file_fstat(an_file_t *file, an_stat_t *anstat);

#endif
