/*
 * Copyright (c) 1999, 2000 The University of Utah and the Flux Group.
 * All rights reserved.
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */

#ifndef _NODEOS_IMPL_ANI_PBUFFER_H_
#define _NODEOS_IMPL_ANI_PBUFFER_H_

#include <assert.h>
#include <sys/types.h>
#include <nodeos/impl/decls.h>
#include <nodeos/impl/ani_threadwork.h>
#include <nodeos/impl/ani_queue.h>

#ifdef OSKIT
#include <oskit/io/skbufio.h>
#include <oskit/machine/atomic.h>

/*
 * NodeOS skbufio implemenation.
 *
 * We provide a wrapper skbufio implementation so that we can interpose on the
 * release method and thus know when a driver is done transmitting a packet
 * (indicating that the pbuf can be reused).
 */
typedef struct ani_pbuf_skbufio {
	oskit_skbufio_t	ourif;
	atomic_t refs;
	oskit_skbufio_t *realif;
	void (*func)(void *);
	void *farg;
} ani_pbuf_skbufio_t;
#else
typedef struct ani_pbuf_data_buffer {
	int refs;
	size_t off;
	size_t size;
	size_t bufsize;
	void *bufdata;
	char bufmem[0];
} ani_pbuf_data_buffer_t;
#endif

/*
 * A packet buffer is a contiguous range of memory described by a pointer
 * (bufdata) and size (bufsize).  Within the buffer is a "valid" region
 * likewise described by a pointer (data) and size (size).  The valid region
 * bounds the valid data in the buffer.  It may be explicitly expanded or
 * contracted at either end up to the limits of the buffer itself.
 */

struct ani_pbuf;
struct ani_chan;
struct ani_flow;

typedef void (*ani_pbuf_free_t)(struct ani_pbuf *pbuf, void *arg);

typedef struct ani_pbuf {
#ifdef OSKIT
	oskit_skbufio_t *skbufio;/* skbufio object */
	struct sk_buff *skb;	/* sk_buff struct */
	ani_pbuf_skbufio_t sendio; /* bufio associated with pbuf during send */
	int popoff;		/* XXX amount to pop off after send done */
#else
	ani_pbuf_data_buffer_t *dbuffer;
#endif
	int refcount;		/* reference count */
	ani_pbuf_free_t ffunc;	/* free function, called when refcount==0 */
	void *farg;		/* free function argument */

	struct ani_flow *flow;	/* associated flow */
	struct ani_chan *chan;	/* associated channel */
	ani_threadwork_t work;	/* work record for thread processing this buf */
	volatile int isbusy;	/* non-zero if pbuffer should not be tweaked */

	ani_queue_head_t link;	/* channel list linkage */
} ani_pbuf_t;

/* field accessors */
#ifdef OSKIT
#define ani_pbuf_offset(pb)	((pb)->skb->data - (pb)->skb->head)
#define ani_pbuf_data(pb)	((void *)(pb)->skb->data)
#define ani_pbuf_size(pb)	((pb)->skb->len)
#define ani_pbuf_bufdata(pb)	((pb)->skb->head)
#define ani_pbuf_bufsize(pb)	((pb)->skb->truesize)
#define ani_pbuf_isvalid(pb)	((pb)->skbufio != 0) 
#else
#define ani_pbuf_offset(pb)	((pb)->dbuffer->off)
#define ani_pbuf_data(pb)	((void *)((int)(pb)->dbuffer->bufdata + (pb)->dbuffer->off))
#define ani_pbuf_size(pb)	((pb)->dbuffer->size)
#define ani_pbuf_bufdata(pb)	((pb)->dbuffer->bufdata)
#define ani_pbuf_bufsize(pb)	((pb)->dbuffer->bufsize)
#define ani_pbuf_isvalid(pb)	((pb)->dbuffer->bufsize != 0)
#endif
#define ani_pbuf_isbusy(pb)	((pb)->isbusy)
#define ani_pbuf_setbusy(pb)	((pb)->isbusy = 1)
#define ani_pbuf_clrbusy(pb)	((pb)->isbusy = 0)
#define ani_pbuf_chan(pb)	((pb)->chan)
#define ani_pbuf_setchan(pb,c)	((pb)->chan = (c))
 
/* reference counting */
#define ani_pbuf_reference(pb)	((pb)->refcount++)
#define ani_pbuf_release(pb)	\
	if ((--(pb)->refcount) == 0) (pb)->ffunc((pb), (pb)->farg)

/* wrapper goo for pbuf lists */
typedef ani_queue_head_t ani_pbuflist_t;

#define ani_pbuflist_init(l)		ani_queue_init(l)

#ifdef DEBUG
#include <stdio.h>
#define ani_pbuflist_dump(s, l) { \
	ani_pbuf_t *_pb; \
	printf("%s%p: ", (s), (l)); \
	ani_queue_iterate((l), _pb, ani_pbuf_t *, link) { \
		if (_pb == 0) { printf("<NULL>"); break; } \
		printf("P%p->", _pb); \
	} \
	printf("\n"); \
}
#else
#define ani_pbuflist_dump(s, l)
#endif

#define ani_pbuflist_isempty(l) \
	ani_queue_empty(l)

#define ani_pbuflist_addhead(l, b) { \
	ani_queue_enter_first((l), (b), ani_pbuf_t *, link); \
}

#define ani_pbuflist_addtail(l, b) { \
	ani_queue_enter((l), (b), ani_pbuf_t *, link); \
}

#define ani_pbuflist_remhead(l) ({ \
	ani_pbuf_t *_pb_ = 0; \
	if (!ani_queue_empty(l)) { \
		ani_queue_remove_first((l), _pb_, ani_pbuf_t *, link); \
		_pb_->link.next = _pb_->link.prev = 0; \
	} \
	_pb_; \
})

#define ani_pbuflist_rempbuf(l, p) { \
	ani_queue_remove((l), (p), ani_pbuf_t *, link); \
	(p)->link.next = (p)->link.prev = 0; \
}

#define ani_pbuflist_append(l1, l2) { \
	ani_queue_merge((l1), (l2), ani_pbuf_t *, link); \
}

#define ani_pbuf_onlist(p)	((p)->link.next != 0)

/* prototypes */

ani_error_t ani_pbuf_init(ani_pbuf_t *pbuf, struct ani_flow *flow,
			  struct ani_chan *chan, size_t bsize, size_t off,
			  size_t size);
void ani_pbuf_init_external(ani_pbuf_t *pbuf, struct ani_flow *flow,
			    struct ani_chan *chan,
#ifdef OSKIT
			    struct sk_buff *skb
#else
			    ani_pbuf_data_buffer_t *dbuf,
			    ani_pbuf_free_t ffunc, void *farg
#endif
			    );
void ani_pbuf_deinit(ani_pbuf_t *pbuf);
ani_error_t ani_pbuf_clone(ani_pbuf_t *pbuf, ani_pbuf_t *newpbuf);
ani_error_t ani_pbuf_copy(ani_pbuf_t *pbuf, ani_pbuf_t **newpbuf);
ani_error_t ani_pbuf_setoffset(ani_pbuf_t *pbuf, size_t newoff,
			       void **out_data);
ani_error_t ani_pbuf_setsize(ani_pbuf_t *pbuf, size_t newsize,
			     void **out_data);
void *ani_pbuf_setmax(ani_pbuf_t *pbuf);
void ani_pbuf_swap(ani_pbuf_t *pbuf1, ani_pbuf_t *pbuf2);
void ani_pbuf_return(ani_pbuf_t *pbuf);

#ifdef OSKIT
struct oskit_netio;
struct oskit_bufio;

ani_error_t ani_pbuf_frombufio(struct ani_flow *flow, struct ani_chan *chan,
			       struct oskit_bufio *bufio, size_t off,
			       size_t size, ani_pbuf_t **out_pbuf);
ani_error_t ani_pbuf_tobufio(ani_pbuf_t *pbuf, void *hdr, size_t hsize,
			     struct oskit_netio *nio,
			     struct oskit_bufio **out_bufio, int *didcopy);
#ifdef OSKIT
void ani_os_wrap_skbufio(ani_pbuf_skbufio_t *newio, oskit_skbufio_t *io,
			 void (*func)(void *), void *arg);
#endif
#else
ani_error_t ani_pbuf_tomem(ani_pbuf_t *pbuf, void *hdr, size_t hsize,
			   void **out_buf, size_t *out_size, int *didcopy);
void ani_pbuf_relmem(void *buf, size_t bufsize);
#endif

#endif /* _NODEOS_IMPL_ANI_PBUFFER_H_ */
