/*
 * Copyright (c) 1999, 2000 The University of Utah and the Flux Group.
 * All rights reserved.
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */

/*
 * Pbuffer cache.
 *
 * Keeps a list of available pbuffers for quick allocation.
 */
#include <nodeos/an_pbuf.h>
#include <nodeos/an_mutex.h>
#include <nodeos/an_cond.h>

struct pbcache_buf {
	an_pbuf_t buf;	/* must be first */
	struct pbcache_buf *next;
};

typedef struct pbcache {
	struct pbcache_buf *head;
	struct pbcache_buf *tail;
	an_mutex_t lock;
	an_cond_t cond;
	int iswaiting;
	int memsize;
	int nbufs;
	struct {
		int gets;
		int looks;
		int gots;
		int waits;
		int empty;
	} stats;
	struct pbcache_buf list[0];
} pbcache_t;

/*
 * Create a cache with the indicated number of pbufs.
 * If non-zero, the specified init function is called for each pbuf after
 * it is created.
 */
pbcache_t *pbcache_create(int nbufs, void (*init)(an_pbuf_t *));

/*
 * Destroy a cache.
 * If non-zero, the specified deinit function is called for each non-busy
 * pbuf before it is destroyed.  Returns the number of elements not destroyed.
 * A return value of zero indicates all pbufs were destroyed as well
 * as the cache object.  Any non-zero return value indicates that the
 * cache was not destroyed.  Repeated calls can be made on the same
 * cache until a zero value is returned.
 */
int pbcache_destroy(pbcache_t *cache, void (*deinit)(an_pbuf_t *));

/*
 * Allocate an available (!busy) pbuf from the indicated cache.
 * If dowait is non-zero, does not return until a buffer is available.
 * Otherwise, returns zero if no buffer is available.
 */
an_pbuf_t *pbcache_alloc(pbcache_t *cache, int dowait);

/*
 * Return a possibly busy pbuf to the indicated cache.
 */
void pbcache_free(pbcache_t *cache, an_pbuf_t *pbuf);

/*
 * Debug: print cache stats
 */
void pbcache_dump(pbcache_t *cache);
