/*
 * Copyright (c) 1999, 2000, 2001 The University of Utah and the Flux Group.
 * All rights reserved.
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */

/*
 * The Packet API
 */
#include <nodeos/an_error.h>
#include <nodeos/an_demuxkey.h>
#include <nodeos/impl/ani_nodeos.h>

#include <nodeos/impl/ani_flow.h>
#include <nodeos/impl/ani_mempool.h>
#include <nodeos/impl/ani_demuxkey.h>
#include <nodeos/impl/ani_error.h>

an_demuxkey_t *
an_demuxkey_create(void *mem, an_flow_t *flow)
{
	ani_demuxkey_t *ikey;
	ani_flow_t *iflow;
	ani_error_t err;

	NODEOS_ENTER;
	ikey = NODEOS_INTERN_DEMUXKEY(mem);
	iflow = NODEOS_INTERN_FLOW(flow);
	if (ikey == 0 || iflow == 0)
		err = ani_error_set(AN_ERR_INVALIDARG,
				    "invalid an_demuxkey argument");
	else {
		assert(iflow == ani_flow_current());
		err = ani_demuxkey_init(ikey);
	}
	NODEOS_LEAVE;
	return ani_error_occured(err) ? 0 : NODEOS_EXTERN_DEMUXKEY(ikey);
}

an_error_t
an_demuxkey_destroy(an_demuxkey_t *key)
{
	ani_error_t err;

	NODEOS_ENTER;
	err = ani_demuxkey_deinit(NODEOS_INTERN_DEMUXKEY(key));
	NODEOS_LEAVE;
	return NODEOS_EXTERN_ERROR(err);
}

an_demuxkey_t *
an_demuxkey_clone(void *mem, an_demuxkey_t *orig)
{
	ani_demuxkey_t *inew, *iorig;
	ani_error_t err;

	NODEOS_ENTER;
	inew = NODEOS_INTERN_DEMUXKEY(mem);
	iorig = NODEOS_INTERN_DEMUXKEY(orig);
	if (inew == 0 || iorig == 0)
		err = ani_error_set(AN_ERR_INVALIDARG,
				    "invalid an_demuxkey argument");
	else
		err = ani_demuxkey_clone(inew, iorig);
	NODEOS_LEAVE;
	return ani_error_occured(err) ? 0 : NODEOS_EXTERN_DEMUXKEY(inew);
}

an_error_t
an_demuxkey_add(an_demuxkey_t *key, int off, int len, void *datap,
		void *maskp)
{
	ani_demuxkey_t *ikey;
	ani_demuxkeyseg_t *seg;
	ani_error_t err;

	NODEOS_ENTER;

	ikey = NODEOS_INTERN_DEMUXKEY(key);
	if (ikey == 0) {
		err = ani_error_set(AN_ERR_INVALIDARG,
				    "invalid an_demuxkey argument");
		goto done;
	}

	/*
	 * Create a tuple segment with indicated key/mask info.
	 *
	 * N.B. no copyin needed in our implementation.
	 */
	seg = ani_demuxkeyseg_tuple(off, len, datap, maskp);
	if (seg == 0) {
		err = ani_error_set(AN_ERR_NOMEMORY,
				    "no memory for demuxkey segment");
		goto done;
	}

	err = ani_demuxkey_append(ikey, seg);
	if (ani_error_occured(err))
		ani_demuxkeyseg_destroy(seg);

done:
	NODEOS_LEAVE;

	return NODEOS_EXTERN_ERROR(err);
}

an_error_t
an_demuxkey_addeth(an_demuxkey_t *key, int off, int flags,
		   u_int8_t src[6], u_int8_t dst[6], u_int16_t etype)
{
	ani_demuxkey_t *ikey;
	ani_demuxkeyseg_t *seg;
	ani_error_t err;

	NODEOS_ENTER;

	ikey = NODEOS_INTERN_DEMUXKEY(key);
	if (ikey == 0) {
		err = ani_error_set(AN_ERR_INVALIDARG,
				    "invalid an_demuxkey argument");
		goto done;
	}

	/*
	 * Create a tuple segment with indicated ethernet header info.
	 *
	 * N.B. no copyin needed in our implementation.
	 */
	seg = ani_demuxkeyseg_eth(off, flags, src, dst, etype);
	if (seg == 0) {
		err = ani_error_set(AN_ERR_NOMEMORY,
				    "no memory for demuxkey segment");
		goto done;
	}

	err = ani_demuxkey_append(ikey, seg);
	if (ani_error_occured(err))
		ani_demuxkeyseg_destroy(seg);

done:
	NODEOS_LEAVE;

	return NODEOS_EXTERN_ERROR(err);
}

an_error_t
an_demuxkey_addipv4(an_demuxkey_t *key, int off, int flags,
		    u_int32_t src, u_int32_t srcmask,
		    u_int32_t dst, u_int32_t dstmask,
		    u_int8_t proto)
{
	ani_demuxkey_t *ikey;
	ani_demuxkeyseg_t *seg;
	ani_error_t err;

	NODEOS_ENTER;

	ikey = NODEOS_INTERN_DEMUXKEY(key);
	if (ikey == 0) {
		err = ani_error_set(AN_ERR_INVALIDARG,
				    "invalid an_demuxkey argument");
		goto done;
	}

	/*
	 * Create a segment with indicated IP header info.
	 */
	seg = ani_demuxkeyseg_ipv4(off, flags,
				   src, srcmask, dst, dstmask, proto);
	if (seg == 0) {
		err = ani_error_set(AN_ERR_NOMEMORY,
				    "no memory for demuxkey segment");
		goto done;
	}

	err = ani_demuxkey_append(ikey, seg);
	if (ani_error_occured(err))
		ani_demuxkeyseg_destroy(seg);

done:
	NODEOS_LEAVE;

	return NODEOS_EXTERN_ERROR(err);
}

an_error_t
an_demuxkey_addudp(an_demuxkey_t *key, int off, int flags,
		   u_int16_t src, u_int16_t dst)
{
	ani_demuxkey_t *ikey;
	ani_demuxkeyseg_t *seg;
	ani_error_t err;

	NODEOS_ENTER;

	ikey = NODEOS_INTERN_DEMUXKEY(key);
	if (ikey == 0) {
		err = ani_error_set(AN_ERR_INVALIDARG,
				    "invalid an_demuxkey argument");
		goto done;
	}

	/*
	 * Create a segment with indicated UDP header info.
	 */
	seg = ani_demuxkeyseg_udp(off, flags, src, dst);
	if (seg == 0) {
		err = ani_error_set(AN_ERR_NOMEMORY,
				    "no memory for demuxkey segment");
		goto done;
	}

	err = ani_demuxkey_append(ikey, seg);
	if (ani_error_occured(err))
		ani_demuxkeyseg_destroy(seg);

done:
	NODEOS_LEAVE;

	return NODEOS_EXTERN_ERROR(err);
}

an_error_t
an_demuxkey_addtcp(an_demuxkey_t *key, int off, int flags,
		   u_int16_t src, u_int16_t dst)
{
	ani_demuxkey_t *ikey;
	ani_demuxkeyseg_t *seg;
	ani_error_t err;

	NODEOS_ENTER;

	ikey = NODEOS_INTERN_DEMUXKEY(key);
	if (ikey == 0) {
		err = ani_error_set(AN_ERR_INVALIDARG,
				    "invalid an_demuxkey argument");
		goto done;
	}

	/*
	 * Create a segment with indicated TCP header info.
	 */
	seg = ani_demuxkeyseg_tcp(off, flags, src, dst);
	if (seg == 0) {
		err = ani_error_set(AN_ERR_NOMEMORY,
				    "no memory for demuxkey segment");
		goto done;
	}

	err = ani_demuxkey_append(ikey, seg);
	if (ani_error_occured(err))
		ani_demuxkeyseg_destroy(seg);

done:
	NODEOS_LEAVE;

	return NODEOS_EXTERN_ERROR(err);
}

an_error_t
an_demuxkey_addanep(an_demuxkey_t *key, int off, int flags,
		    u_int8_t version, u_int16_t protoid)
{
	ani_demuxkey_t *ikey;
	ani_demuxkeyseg_t *seg;
	ani_error_t err;

	NODEOS_ENTER;

	ikey = NODEOS_INTERN_DEMUXKEY(key);
	if (ikey == 0) {
		err = ani_error_set(AN_ERR_INVALIDARG,
				    "invalid an_demuxkey argument");
		goto done;
	}

	/*
	 * Create a segment with indicated TCP header info.
	 */
	seg = ani_demuxkeyseg_anep(off, flags, version, protoid);
	if (seg == 0) {
		err = ani_error_set(AN_ERR_NOMEMORY,
				    "no memory for demuxkey segment");
		goto done;
	}

	err = ani_demuxkey_append(ikey, seg);
	if (ani_error_occured(err))
		ani_demuxkeyseg_destroy(seg);

done:
	NODEOS_LEAVE;

	return NODEOS_EXTERN_ERROR(err);
}

an_error_t
an_demuxkey_tostring(an_demuxkey_t *key, char *buf, int buflen)
{
	ani_demuxkey_t *ikey;
	ani_error_t err;

	NODEOS_ENTER;
	
	ikey = NODEOS_INTERN_DEMUXKEY(key);
	if (ikey == 0) {
		err = ani_error_set(AN_ERR_INVALIDARG,
				    "invalid an_demuxkey_t parameter");
	} else if (buf == 0) {
		err = ani_error_set(AN_ERR_INVALIDARG,
				    "invalid char* parameter");
	} else {
		/* XXX no intern of buf, buflen */
		err = ani_demuxkey_tostring(ikey, buf, buflen);
	}
	NODEOS_LEAVE;
	return NODEOS_EXTERN_ERROR(err);
}

int
an_demuxkey_matches(an_demuxkey_t *key, void *data, int datalength)
{
	ani_demuxkey_t *ikey;
	ani_error_t err;
	int offset;
	
	NODEOS_ENTER;

	ikey = NODEOS_INTERN_DEMUXKEY(key);
	if (ikey == 0) {
		err = ani_error_set(AN_ERR_INVALIDARG,
				    "bad key argument "
				    "in an_demuxkey_matches");
		offset = -1;
		goto done;
	}

	if (data == 0 || datalength < 0) {
		err = ani_error_set(AN_ERR_INVALIDARG,
				    "bad data/length argument "
				    "in an_demuxkey_matches");
		offset = -1;
		goto done;
	}
	
	/*
	 * N.B. no copyin of data necessary in our implementation.
	 */
	offset = ani_demuxkey_match(ikey, data, datalength);
	err = ani_error_clear();

 done:
	NODEOS_LEAVE;
	return offset;
}
