/*
 * Copyright (c) 2000 The University of Utah and the Flux Group.
 * All rights reserved.
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */

/*
 * The Information API
 */
#include <nodeos/an_info.h>
#include <nodeos/impl/ani_if.h>
#include <nodeos/impl/ani_moab_setup.h>
#include <nodeos/config.h>
#include "os.h"

/*
 * Retrieve information about network interfaces.
 *
 * ``info'' is an array of an_ifinfo structs to be filled in
 * ``count'' the size of the array on call, the number filled in on return
 *
 * Returns 0 on success, an error otherwise.
 */
an_error_t
an_info_getifs(an_infoif_t info[], int *count)
{
	ani_error_t err;
	int icount, num;
	an_infoif_t *iinfo;

	NODEOS_ENTER;

	/*
	 * Validate arguments.
	 */
	if (count == 0) {
		err = ani_error_set(AN_ERR_INVALIDARG,
				    "invalid count argument");
		goto done; 
	}

	/*
	 * Get desired size.
	 *
	 * N.B. no copyin necessary in our implemenatation.
	 */
	icount = *count;

	/*
	 * Get a pointer to the info
	 */
	num = icount;
	iinfo = ani_if_getinfo(&num);

	/*
	 * Copy out data
	 *
	 * N.B. we can just do a direct copy in our implementation.
	 */
	*count = num;
	if (iinfo != 0)
		memcpy(info, iinfo, sizeof *iinfo * num);
	err = ani_error_clear();

done:
	NODEOS_LEAVE;
	return NODEOS_EXTERN_ERROR(err);
}

/*
 * Retrieve information about CPUs on the node.
 *
 * ``cpuinfo'' is a pointer to a infocpu_t struct to fill in.
 *
 * Returns 0 on success, an error otherwise.
 */
an_error_t
an_info_getcpu(an_infocpu_t *cpuinfo)
{
	ani_resource_t res;
	ani_error_t err;

	NODEOS_ENTER;
	ani_moab_resources(&res);
	cpuinfo->ncpu = 1;	/* XXX */
	cpuinfo->cps = res.cpurate;
	err = ani_error_clear();
	NODEOS_LEAVE;
	return NODEOS_EXTERN_ERROR(err);
}

/*
 * Get options the NodeOS was built with.
 * XXX hack.
 */
const char *
an_info_getoptlist(void)
{
	static char options_str[256];

	LW_NODEOS_ENTER;
	if (options_str[0] == 0) {
#ifdef HAVE_NET
		strcat(options_str, "HAVE_NET ");
#endif
#ifdef HAVE_DISK
		strcat(options_str, "HAVE_DISK ");
#endif
#ifdef CHAN_CLOCK
		strcat(options_str, "CHAN_CLOCK ");
#endif
#ifdef CHAN_ETH
		strcat(options_str, "CHAN_ETH ");
#endif
#ifdef CHAN_IF
		strcat(options_str, "CHAN_IF ");
#endif
#ifdef CHAN_SOCKET
		strcat(options_str, "CHAN_SOCKET ");
#endif
#ifdef CHAN_STDIO
		strcat(options_str, "CHAN_STDIO ");
#endif
#ifdef CHAN_CLICK
		strcat(options_str, "CHAN_CLICK ");
#endif
	}
	LW_NODEOS_LEAVE;

	/*
	 * N.B.: don't need to copyout the string in our implementation.
	 */
	return options_str;
}


/*
 * Return the hostname of this node.  Returns a read-only constant.
 * XXX hack.
 */
const char *
an_info_gethostname(void)
{
	const char *name;

	LW_NODEOS_ENTER;
	name = ani_os_gethostname();
	LW_NODEOS_LEAVE;

	/*
	 * N.B.: don't need to copyout the name in our implementation.
	 */
	return name;
}


/*
 * Set the hostname of this node.  The nodeos will
 * make a copy of the argument.
 */
an_error_t
an_info_sethostname(const char *name)
{
	ani_error_t err;
	int ierr;

	NODEOS_ENTER;

	/*
	 * N.B.: don't need to copyin the name in our implementation.
	 */
	ierr = ani_os_sethostname(name);
	if (ierr != 0)
		err = ani_error_set(ierr, "os_sethostname failed");
	else
		err = ani_error_clear();

	NODEOS_LEAVE;
	return NODEOS_EXTERN_ERROR(err);
}
