/*
 * Copyright (c) 1999, 2000 The University of Utah and the Flux Group.
 * All rights reserved.
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */

/*
 * The Mempool API
 */
#include <nodeos/an_error.h>
#include <nodeos/an_impl.h>
#include <nodeos/an_mempool.h>

#include <nodeos/impl/ani_nodeos.h>
#include <nodeos/impl/ani_mempool.h>

an_mempool_t *
an_mempool_create(void *mem, an_mempool_t *parent, an_mempoolfull_func_t func)
{
	ani_mempool_t *newpool;
	ani_mempool_t *iparent = NULL;
	ani_error_t err;

	NODEOS_ENTER;

	newpool = NODEOS_INTERN_MEMPOOL(mem);
	if (newpool == 0) {
		err = ani_error_set(AN_ERR_INVALIDARG,
				    "invalid mempool argument");
		goto done;
	}

	if (parent) {
		iparent = NODEOS_INTERN_MEMPOOL(parent);
		if (iparent == 0) {
			err = ani_error_set(AN_ERR_INVALIDARG,
					    "invalid parent mempool argument");
			goto done;
		}
	}

	err = ani_mempool_init(newpool, iparent, func);

 done:
	NODEOS_LEAVE;
	return ani_error_occured(err) ? 0 : NODEOS_EXTERN_MEMPOOL(newpool);
}

an_error_t
an_mempool_destroy(an_mempool_t *mpool)
{
	ani_mempool_t *impool;
	ani_error_t err;

	NODEOS_ENTER;
	impool = NODEOS_INTERN_MEMPOOL(mpool);
	if (impool == 0) {
		err = ani_error_set(AN_ERR_INVALIDARG,
				    "invalid mempool argument");
		goto done;
	}
	err = ani_mempool_deinit(impool);
 done:
	NODEOS_LEAVE;
	return NODEOS_EXTERN_ERROR(err);
}

void *
an_mempool_alloc(an_mempool_t *mpool, size_t size)
{
	ani_mempool_t *impool;
	void *ptr = 0;

	NODEOS_ENTER;
	impool = NODEOS_INTERN_MEMPOOL(mpool);
	if (impool != 0)
		ptr = ani_mempool_alloc(impool, size);
	NODEOS_LEAVE;
	return ptr;
}

void
an_mempool_free(an_mempool_t *mpool, void *ptr, size_t size)
{
	ani_mempool_t *impool;

	NODEOS_ENTER;
	impool = NODEOS_INTERN_MEMPOOL(mpool);
	if (impool != 0)
		ani_mempool_free(impool, ptr, size);
	NODEOS_LEAVE;
}

/*
 * Return statistics for a mempool.
 * Return memory use statistics for the given mempool in the provided buffer.
 * Returns zero on success or an error code otherwise.
 */
an_error_t
an_mempool_stats(an_mempool_t *mpool, an_mempoolstats_t *stats)
{
	ani_mempool_t *impool;
	ani_error_t err;

	NODEOS_ENTER;

	impool = NODEOS_INTERN_MEMPOOL(mpool);
	if (impool == 0)
		err = ani_error_set(AN_ERR_INVALIDARG,
				    "mempool parameter is null");
	else {
		/*
		 * Get the stats and copy them out to the user's memory.
		 *
		 * N.B. we can just copy stats directly into user buffer
		 * in our implementation.
		 */
		err = ani_mempool_stats(impool, stats);
	}

	NODEOS_LEAVE;
	return NODEOS_EXTERN_ERROR(err);
}

