/*
 * Copyright (c) 2000 The University of Utah and the Flux Group.
 * All rights reserved.
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */

/*
 * The packet buffer (pbuf) API
 */
#include <nodeos/an_pbuf.h>
#include <nodeos/an_error.h>

#include <nodeos/impl/ani_nodeos.h>
#include <nodeos/impl/ani_flow.h>
#include <nodeos/impl/ani_mempool.h>

#include "os.h"

an_pbuf_t *
an_pbuf_create(void *mem, int flags, int offset, int size)
{
	ani_flow_t *iflow = ani_flow_current();
	ani_pbuf_t *ipbuf = 0;
	size_t bsize;
	ani_mempool_t *cpool;
	int didalloc = 0;
	ani_error_t err;

	NODEOS_ENTER;

	/*
	 * Translate size-related flags.
	 * XXX we only have one buffer size so this is easy!
	 */
	switch (flags & AN_PBUF_SIZEMASK) {
	case AN_PBUF_DEFAULT:
	case AN_PBUF_ETH:
	case AN_PBUF_SOCK:
	case AN_PBUF_CLOCK:
	case AN_PBUF_STDIO:
		bsize = NODEOS_CHAN_BUFSIZE;
		break;
	default:
		err = ani_error_set(AN_ERR_INVALIDARG, "invalid flags");
		goto done;
	}

	/*
	 * Constrain the active region
	 */
	if (offset < 0 || size < 0 || offset + size > bsize) {
		err = ani_error_set(AN_ERR_INVALIDARG, "invalid offset/size");
		goto done;
	}

	/*
	 * Allocate memory for the object if needed.
	 */
	cpool = ani_flow_mempool(iflow);
	if (mem == 0) {
		mem = ani_mempool_alloc(cpool, sizeof(an_pbuf_t));
		didalloc = 1;
	}

	ipbuf = NODEOS_INTERN_PBUF(mem);
	if (ipbuf == 0)
		err = ani_error_set(AN_ERR_INVALIDARG, "invalid mem argument");
	else
		err = ani_pbuf_init(ipbuf, iflow, 0, bsize, offset, size);

	if (err != 0 && didalloc)
		ani_mempool_free(cpool, mem, sizeof(an_pbuf_t));

 done:
	NODEOS_LEAVE;
	return ani_error_occured(err) ? 0 : NODEOS_EXTERN_PBUF(ipbuf);
}

an_error_t
an_pbuf_destroy(an_pbuf_t *pbuf)
{
	ani_error_t err;
	ani_pbuf_t *ipbuf;

	NODEOS_ENTER;
	ipbuf = NODEOS_INTERN_PBUF(pbuf);
	if (ipbuf == 0)
		err = ani_error_set(AN_ERR_INVALIDARG, "invalid pbuf");
	else {
		ani_pbuf_deinit(ipbuf);
		err = ani_error_clear();
	}
	NODEOS_LEAVE;
	return NODEOS_EXTERN_ERROR(err);
}

an_error_t
an_pbuf_return(an_pbuf_t *pbuf)
{
	ani_pbuf_t *ipbuf;
	ani_error_t err;

	NODEOS_ENTER;
	ipbuf = NODEOS_INTERN_PBUF(pbuf);
	if (ipbuf == 0)
		err = ani_error_set(AN_ERR_INVALIDARG, "invalid pbuf");
	else {
		ani_pbuf_return(ipbuf);
		err = ani_error_clear();
	}
	NODEOS_LEAVE;
	return NODEOS_EXTERN_ERROR(err);
}

an_pbuf_t *
an_pbuf_clone(an_pbuf_t *pbuf, void *newpbufmem)
{
	ani_flow_t *iflow = ani_flow_current();
	int didalloc = 0;
	ani_pbuf_t *ipbuf = 0, *inewpbuf = 0;
	ani_mempool_t *cpool = 0;
	ani_error_t err;

	NODEOS_ENTER;
	ipbuf = NODEOS_INTERN_PBUF(pbuf);
	if (ipbuf == 0) {
		err = ani_error_set(AN_ERR_INVALIDARG, "invalid pbuf");
		goto done;
	}

	cpool = ani_flow_mempool(iflow);
	if (newpbufmem == 0) {
		newpbufmem = ani_mempool_alloc(cpool, sizeof(an_pbuf_t));
		didalloc = 1;
	}

	inewpbuf = NODEOS_INTERN_PBUF(newpbufmem);
	if (inewpbuf == 0) {
		err = ani_error_set(AN_ERR_INVALIDARG, "no memory for pbuf");
		goto done;
	}

	err = ani_pbuf_clone(ipbuf, inewpbuf);

 done:
	if (err != 0 && didalloc)
		ani_mempool_free(cpool, newpbufmem, sizeof(an_pbuf_t));
	NODEOS_LEAVE;
	return ani_error_occured(err) ? 0 : NODEOS_EXTERN_PBUF(inewpbuf);
}

an_error_t
an_pbuf_setoffset(an_pbuf_t *pbuf, int noffset, void **ndata)
{
	void *data;
	ani_pbuf_t *ipbuf;
	ani_error_t err;

	NODEOS_ENTER;
	ipbuf = NODEOS_INTERN_PBUF(pbuf);
	err = ani_pbuf_setoffset(ipbuf, noffset, &data);

	/*
	 * Copy out the new data pointer
	 *
	 * N.B. can do a direct copy in our implementation.
	 */
	if (!ani_error_occured(err) && ndata != 0)
		*ndata = data;

	NODEOS_LEAVE;
	return NODEOS_EXTERN_ERROR(err);
}

an_error_t
an_pbuf_setsize(an_pbuf_t *pbuf, int nsize, void **ndata)
{
	void *data;
	ani_pbuf_t *ipbuf;
	ani_error_t err;

	NODEOS_ENTER;
	ipbuf = NODEOS_INTERN_PBUF(pbuf);
	err = ani_pbuf_setsize(ipbuf, nsize, &data);

	/*
	 * Copy out the new data pointer
	 *
	 * N.B. can do a direct copy in our implementation.
	 */
	if (!ani_error_occured(err) && ndata != 0)
		*ndata = data;

	NODEOS_LEAVE;
	return NODEOS_EXTERN_ERROR(err);
}
