/*
 * Copyright (c) 1999, 2000, 2001 The University of Utah and the Flux Group.
 * All rights reserved.
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */

/*
 * A Click composite channel.
 *
 * TODO: Locking!
 */
#include <nodeos/config.h>

#ifdef CHAN_CLICK

#include <stdio.h>
#include <string.h>
#include <assert.h>

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netinet/in_systm.h>
#include <netinet/if_ether.h>
#include <netinet/ip.h>
#include <netinet/udp.h>

#include <nodeos/an_info.h>

#include <nodeos/impl/an_impl.h>
#include <nodeos/impl/an_error.h>
#include <nodeos/impl/ani_nodeos.h>
#include <nodeos/impl/ani_ilock.h>
#include <nodeos/impl/ani_thread.h>
#include <nodeos/impl/ani_pbuf.h>

#include <nodeos/impl/ani_flow.h>
#include <nodeos/impl/ani_chan.h>
#include <nodeos/impl/ani_demuxkey.h>
#include <nodeos/impl/ani_threadpool.h>
#include <nodeos/impl/ani_time.h>
#include "ani_debug.h"

static void chan_click_init(void);
static int chan_click_matches(an_chanproto_t *proto);
static ani_error_t chan_click_create(ani_chan_t *chan);
static void chan_click_destroy(ani_chan_t *chan);
static ani_error_t chan_click_send(ani_chan_t *chan, ani_pbuf_t *pbuf);
static ani_error_t chan_click_enable(ani_chan_t *chan);
static void chan_click_disable(ani_chan_t *chan);

struct chanops chan_click_ops = {
	chan_click_init,	chan_click_matches,
	chan_click_create,	chan_click_destroy,	chan_click_send,
	chan_click_enable,	chan_click_disable,	0
};

/*
 * A table of configurations we support
 */
struct click_config {
	const char     *proto;		/* The protocol spec to match on */
	int		ctype;		/* in or out. Temporary */
	char	       *config;	/* The graph */
	int	      (*rawdemux)(ani_chan_t *chan);  /* The rawdemux setup */
	int	      (*chandemux)(ani_chan_t *chan); /* The demux setup */
	char	     *(*graphsetup)(ani_chan_t *chan);/* Localize the graph */
};

/*
 * Per-graph structure. For each graph we construct, we create one of
 * these. Think of these as tuples (graph,rawchan) that can be shared
 * amongst channels, as long as the proto and the rawchan are equal.
 * For example, you would demux different UDP/IP streams from the same
 * source into different channels based on port number by using the same
 * combo of eth input channel and udp/ip graph, and then demux the resulting
 * UDP packets. 
 *
 * The lock is lower in the locking hierarchy than the channel lock.
 */
struct clickgraph {
	ani_ilock_t	     lock;
	struct clickgraph   *next;
	struct click_config *configp;
	ani_demuxkeylist_t   demux;	/* list of demuxers for this graph */
	void		    *router;
	an_chanrecvfunc_t    rfunc;	/* Click input function */
	void		    *rfuncarg;	/* Click C++ element */
	ani_chan_t	    *rawchan;
	ani_threadpool_t    *tpool;
	int		     enabled;
	int		     count;
};
static struct clickgraph	*graphs;	/* Head of list */

/*
 * Each channel gets one of these suckers.
 */
struct clickdata {
	struct clickgraph	*graph;
};

/* forward decls */
static struct click_config	*find_click_config(char *proto, int ctype);
static void			chan_click_intr(ani_chan_t *, ani_pbuf_t *);
	
/*
 * Initialize.
 */
static void
chan_click_init(void)
{
}

static int
chan_click_matches(an_chanproto_t *proto)
{
	return 0;
}

static ani_error_t
chan_click_create(ani_chan_t *chan)
{
	ani_chan_t	  *rawchan = chan->component.rawchan;
	an_chanrecvfunc_t rfunc;	         /* Click input function */
	void		  *rfuncarg;		 /* Click C++ element */
	ani_threadpool_t  *tpool = 0;
	ani_mempool_t     *cpool;
	an_cpuspec_t	  cspec;
	void		  *router;
	struct click_config *configp;
	ani_error_t	  err;
	ani_chan_t	  *input, *output;
	void		 (*deliver)(ani_chan_t *, ani_pbuf_t *);
	struct clickgraph *thisgraph;
	struct clickdata  *thisdata;
	char		  *graphstr;

	/*
	 * Make sure we can handle this protocol! 
	 */
	if ((configp = find_click_config(chan->proto.data,
					 ani_chan_type(chan))) == NULL) {
		return ani_error_set(AN_ERR_INVALIDARG,
				     "Cannot support protocol spec!");
	}

	err = ani_error_clear();

	/*
	 * If its an inchannel, look to see if the graph has already
	 * been instantiated for the given rawchan. Output graphs are
	 * currently unique.
	 */
	/* LOCK ME! */
	thisgraph = 0;
	if (ani_chan_type(chan) == CHAN_IN) {
		thisgraph = graphs;
		while (thisgraph) {
			if (thisgraph->configp == configp &&
			    thisgraph->rawchan == rawchan)
				break;
			thisgraph = thisgraph->next;
		}		;
	}
	if (thisgraph)
		goto skip;

	/*
	 * Customize the graph.
	 */
	if (configp->graphsetup)
		graphstr = configp->graphsetup(chan);
	else
		graphstr = configp->config;

	/*
	 * XXX create a threadpool with a single thread to serialize packets
	 * and events for the Click graph.
	 */
	cpool = ani_flow_mempool(chan->flow);
	tpool = ani_mempool_alloc(cpool, sizeof *tpool);
	if (!tpool)
		return ani_error_set(AN_ERR_NOMEMORY,
				     "Could not allocate threadpool");
	cspec.ttype = AN_TH_IMPLICIT;
	cspec.threads.im.maxthreads = 1;
	cspec.threads.im.stacksize = 0;
	err = ani_threadpool_init(tpool, chan->flow, &cspec, 0);
	if (ani_error_occured(err)) {
		ani_mempool_free(cpool, tpool, sizeof *tpool);
		return ani_error_set(AN_ERR_NOMEMORY,
				     "could not init threadpool");
	}

	/*
	 * Allocate per-graph thingie.
	 */
	thisgraph = ani_mempool_alloc(ani_mempool_current(),
				      sizeof(*thisgraph));
	if (thisgraph == 0) {
		err = ani_error_set(ENOMEM, "no memory for click graph data");
		goto done;
	}
	memset(thisgraph, 0, sizeof(*thisgraph));
	
	/*
	 * Ask Click to instantiate the graph. We need to exchange some
	 * stuff, and this seems like the best way to do it.
	 */
	if (ani_chan_type(chan) == CHAN_IN) {
		input   = rawchan;		/* Incoming Eth */
		output  = (ani_chan_t *)thisgraph;
		deliver = chan_click_intr;
	}
	else {
		input   = chan;
		output  = rawchan;		/* Outgoing Eth */
		deliver = (void (*)(ani_chan_t *, ani_pbuf_t *))ani_chan_send;
	}
	err = ani_clickchan_instantiate(graphstr,
					input,  /* Input side of graph */
					output,	  /* Output side of graph */
					tpool,	  /* Router needs this */
					deliver,  /* Delivery function */
					&rfunc,	   /* Injection into graph */
					&rfuncarg, /* Element */
					&router);  /* Router, for dealloc */

	if (err) {
		ani_mempool_free(cpool, tpool, sizeof *tpool);
		goto done;
	}

	/*
	 * Call the demux setup function, which will append to the
	 * demux for the raw channel.
	 */
	if (configp->rawdemux)
		configp->rawdemux(rawchan);

	chan->component.data = router;
	if (ani_chan_type(chan) == CHAN_IN) {
		/*
		 * The raw channel callback enters the Click graph.
		 */
		rawchan->tdata.in.func = rfunc;	
		rawchan->tdata.in.farg = rfuncarg;
		rawchan->tpool = tpool;
		rawchan->maxthreads = 1;
	}
	thisgraph->tpool    = tpool;
	thisgraph->rawchan  = rawchan;
	thisgraph->rfunc    = rfunc;
	thisgraph->rfuncarg = rfuncarg;
	thisgraph->router   = router;
	thisgraph->configp  = configp;
	thisgraph->enabled  = 0;
	thisgraph->count    = 0;
	thisgraph->next	    = graphs;		/* XXX Lock Me! */
	graphs              = thisgraph;
 skip:
	/*
	 * Allocate per-channel thingie.
	 */
	thisdata = ani_mempool_alloc(ani_mempool_current(),
				     sizeof(struct clickdata));
	if (thisdata == 0) {
		err = ani_error_set(ENOMEM, "no memory for click chan data");
		goto done;
	}
	
	thisdata->graph = thisgraph;
	chan->data = thisdata;
	
	if (ani_chan_type(chan) == CHAN_OUT) {
		/*
		 * This channel enters the Click graph using the single
		 * thread in the threadpool created above. We stash the
		 * the callback here.
		 */
		chan->tdata.in.func = thisgraph->rfunc;	
		chan->tdata.in.farg = thisgraph->rfuncarg;
		chan->tpool         = thisgraph->tpool;
		chan->maxthreads    = 1;

		/*
		 * Augment the demux filter with the address and demux info
		 * as necessary. This is the second level demux for packets
		 * coming out of the graph. 
		 */
		if (thisgraph->configp->chandemux)
			thisgraph->configp->chandemux(chan);
	}
	thisgraph->count++;

 done:
	return err;
}

/*
 * Called with the channel locked
 */
static void
chan_click_destroy(ani_chan_t *chan)
{
	struct clickdata	*thisdata  = (struct clickdata *) chan->data;
	struct clickgraph	*thisgraph = thisdata->graph;
	ani_threadpool_t	*tpool     = thisgraph->tpool;

	/* Locking */
	thisgraph->count--;
	if (thisgraph->count)
		return
	   
	assert_ani_threadpool_valid(tpool);
	ani_threadpool_disable(tpool, 1);
	ani_threadpool_deinit(tpool, 1);
	ani_mempool_free(ani_flow_mempool(ani_flow_current()),
			 tpool, sizeof *tpool);
}

static ani_error_t
chan_click_enable(ani_chan_t *chan)
{
	struct clickdata	*thisdata  = (struct clickdata *) chan->data;
	struct clickgraph	*thisgraph = thisdata->graph;
	ani_error_t		err;
	
	assert(ani_chan_type(chan) == CHAN_IN);

	err = ani_demuxkeylist_add(&thisgraph->demux, chan->tdata.in.demux);
	if (ani_error_occured(err))
		return err;

	/*
	 * We want to enable the underlying channel only once!
	 */
	if (! thisgraph->enabled)
		ani_chan_enable(chan->component.rawchan);
	
	thisgraph->enabled = 1;

	return ani_error_clear();
}

static void
chan_click_disable(ani_chan_t *chan)
{
	struct clickdata	*thisdata  = (struct clickdata *) chan->data;
	struct clickgraph	*thisgraph = thisdata->graph;

	assert(ani_chan_type(chan) == CHAN_IN);

	ani_demuxkeylist_remove(&thisgraph->demux, chan->tdata.in.demux);

	/*
	 * Turn off the packets at the spiggot? Maybe not the right thing
	 * to do, if we want to share the eth chan.
	return ani_chan_disable(chan->component.rawchan);
	 */
}

/*
 * Send a packet into the Click graph by starting a thread to enter
 * the graph with the chan/pbuf.
 *
 * Called with the channel locked.
 */
static ani_error_t
chan_click_send(ani_chan_t *chan, ani_pbuf_t *pbuf)
{
	ani_chan_thread_start(chan, pbuf, 0);
	return ani_error_clear();
}

/*
 * The receive interrupt function, which is explicitly called from
 * the Click toinchan element when it has a pbuf ready to hand off
 * from the graph. We arrange things so that this is called with
 * the graph, although its not prototyped that way.
 *
 * NB: We are assuming that packets coming out of the graph are ours,
 *     and that we do not have to copy them.
 *
 */
static void
chan_click_intr(ani_chan_t *chan, ani_pbuf_t *inpbuf)
{
	struct clickgraph	*thisgraph = (struct clickgraph *) chan;
	ani_error_t		err;
	ani_pbuf_t		*outpbuf;
	int			msize;

	assert(ani_pbuf_size(inpbuf) >=
	       sizeof(struct ip) + sizeof(struct udphdr));
	/*
	 * Need to strip the udp/ip protocol headers off. We could have the
	 * click graph do it, but this seems more flexible. 
	 */
	err = ani_pbuf_setoffset(outpbuf, ani_pbuf_offset(outpbuf) +
				 sizeof(struct ip) + sizeof(struct udphdr), 0);

	msize = ani_pbuf_size(inpbuf);
	chan = ani_demuxkeylist_apply(&thisgraph->demux,
				      ani_pbuf_data(inpbuf), &msize);
	if (chan == 0) {
		ani_pbuf_return(inpbuf);
		return;
	}

	ani_chan_lock(chan);
	assert((chan->flags & CHAN_DYING) == 0);

	/*
	 * Swap packet into a buffer for the target channel,
	 * dropping the packet if no target buffer is available.
	 * Return the original pbuf from whence it came.
	 */
	err = ani_chan_pbuf_alloc(chan, NODEOS_CHAN_BUFSIZE, &outpbuf);
	if (ani_error_occured(err)) {
		ani_chan_unlock(chan);
		ani_pbuf_return(inpbuf);
		return;
	}
	ani_pbuf_swap(inpbuf, outpbuf);

	ani_chan_thread_start(chan, outpbuf, msize);

	ani_chan_unlock(chan);
	ani_pbuf_return(inpbuf);
}

/*
 * A trivial UDP/IP graph. Note, we do not do ARP here since there is
 * no device output channel. This is a serious problem. Must be fixed.
 *
 * The eth header is assumed to be stripped, hence the classifier looks
 * at ip header offsets. The point of the classifier to ensure that we
 * got UDP/IP packets by checking the proto field of the ip header. This
 * is probably not necessary.
 */
static char click_ipudp_graph[] =
	"in  :: FromInChan;"
	"out :: ToInChan;"
	"cl  :: Classifier(9/11, -); "
	"ci  :: CheckIPHeader;"
	"ck  :: CheckUDPHeader;"
	"ipr :: IPReassembler;"
	"in -> cl;"
	"cl [0] -> "
		"MarkIPHeader(0) ->"
		"ci ->"
		"ipr ->"
		"ck ->"
		"out;"
	"cl  [1] -> Print(BadClass)  -> Discard;"
	"ck  [1] -> Print(BadUdpHdr) -> Discard;"
	"ci  [1] -> Print(BadIPHdr)  -> Discard;"
	"ipr [1] -> Print(BadIPReas) -> Discard;"
       ;

/*
 * The raw demux function adds to the raw input channel for the above graph.
 */
static int
click_ipudp_rawdemux(ani_chan_t *chan)
{
	unsigned short	val_s = htons(0x0800);
	unsigned char	val_c = 17;

	/* Shoot me */
	ani_demuxkey_append(chan->tdata.in.demux,
			    ani_demuxkeyseg_tuple(12, 2, &val_s, 0));
	ani_demuxkey_append(chan->tdata.in.demux,
			    ani_demuxkeyseg_tuple(23, 1, &val_c, 0));

	return 0;
}

/*
 * This demux adds to the channel demux (after the graph).
 *
 * XXX Assume no IP options since the demuxer cannot deal with it. 
 */
static int
click_ipudp_chandemux(ani_chan_t *chan)
{
	unsigned int	val_l;
	unsigned short	val_s;
	struct _hdr {
		struct ip	ip;
		struct udphdr	udp;
	} *hdr = (struct _hdr *) 0;

	if (chan->addr.udp4.saddr) {
		val_l = chan->addr.udp4.saddr;
		ani_demuxkey_append(chan->tdata.in.demux,
				ani_demuxkeyseg_tuple((int)&hdr->ip.ip_src,
						      sizeof(val_l),
						      &val_l, 0));
	}

	if (chan->addr.udp4.sport) {
		val_s = chan->addr.udp4.sport;
		ani_demuxkey_append(chan->tdata.in.demux,
				ani_demuxkeyseg_tuple((int)&hdr->udp.uh_sport,
						      sizeof(val_s),
						      &val_s, 0));
	}

	if (chan->addr.udp4.daddr) {
		val_l = chan->addr.udp4.daddr;
		ani_demuxkey_append(chan->tdata.in.demux,
				ani_demuxkeyseg_tuple((int)&hdr->ip.ip_src,
						      sizeof(val_l),
						      &val_l, 0));
	}

	if (chan->addr.udp4.dport) {
		val_s = chan->addr.udp4.dport;
		ani_demuxkey_append(chan->tdata.in.demux,
				ani_demuxkeyseg_tuple((int)&hdr->udp.uh_dport,
						      sizeof(val_s),
						      &val_s, 0));
	}
	return 0;
}

/*
 * This is an UDP/IP output graph. The tricky part of this is parameterizing
 * the graph with the information from the address specification. I'm going
 * to be dead simple right now, and sprintf the dumb thing.
 */
static char click_udpip_graph[] =
	"in  :: FromInChan;"
	"out :: ToInChan;"
	"fr  :: IPFragmenter(1500);"
	"in  -> "
		"UDPIPEncap(%s, %d, %s, %d, 1) ->"
		"fr ->"
		"out;"
	"fr[1] -> Print(FRAG) -> Discard;"
       ;

/*
 * Parameterize the graph.
 */
static char *
click_udpip_setup(ani_chan_t *chan)
{
	char	buf[1024], saddr[20], daddr[20], *bp;
	int	sport = ntohs(chan->addr.udp4.sport);
	int	dport = ntohs(chan->addr.udp4.dport);
        struct in_addr sa, da;

	sa.s_addr = chan->addr.udp4.saddr;
	da.s_addr = chan->addr.udp4.daddr;

	strcpy(saddr, inet_ntoa(sa));
	strcpy(daddr, inet_ntoa(da));

	sprintf(buf, click_udpip_graph, saddr, sport, daddr, dport);

	bp = ani_mempool_alloc(ani_mempool_current(), strlen(buf) + 1);
	if (!bp)
		return 0;

	strcpy(bp, buf);
	return bp;
}

/*
 * The table. I'm not happy about the demux thing at all!
 */
static struct click_config click_configs[] = {
	{ "udp/ip", CHAN_IN,
	  click_ipudp_graph, click_ipudp_rawdemux, click_ipudp_chandemux, 0 },
	{ "udp/ip", CHAN_OUT,
	  click_udpip_graph, 0, 0, click_udpip_setup },
	{ 0, 0, 0, 0, 0 },
};

/*
 * Figure out which builtin graph to use for a given proto spec.
 * Eventually we want to parameterize these things somehow.
 */
static struct click_config *
find_click_config(char *proto, int ctype)
{
	struct click_config	*cp;
	
	/*
	 * Okay, lets not get fancy yet.
	 */
	for (cp = click_configs; cp->proto; cp++)
		if (strncmp(proto, cp->proto, strlen(cp->proto)) == 0 &&
		    cp->ctype == ctype)
			break;

	return cp->proto ? cp : 0;
}

/*
 * XXX this should be merged with the standard channel routines at some point.
 */
ani_error_t
ani_clickchan_init(ani_chan_t *chan, int ctype, ani_flow_t *flow,
		   ani_threadpool_t *tpool,
		   an_chanproto_t *proto, an_chanaddr_t *addr,
		   ani_demuxkey_t *demux, an_cpuspec_t *cpu,
		   ani_pbuf_t **pbufs, int npbufs, an_netspec_t *net,
		   an_chanrecvfunc_t func, void *farg,
		   /* Click stuff */
		   ani_chan_t *rawchan)
{
	struct chanops *ops;
	ani_error_t err;

	assert(ctype != CHAN_CUT);

	memset(chan, 0, sizeof *chan);

	/*
	 * Find the click ops.
	 * XXX Not based on the protocol string since we have no easy
	 * way of knowing if we can support the protocol. 
	 */
	ops = &chan_click_ops;

	/*
	 * XXX set this now so that functions we call know who to charge
	 * for memory.  We officially link into the flow after everything
	 * has been setup.
	 */
	chan->flow = flow;

	/*
	 * XXX should we hold a ref to the pool here?
	 * Don't think so.  Threadpool cannot go away as long as the flow
	 * is alive.  If the flow has been destroyed, so have all its channels.
	 */
	chan->tpool = tpool;

	/*
	 * Record the resource specifications.
	 * It is up to the entry layer to perform any cross-address space
	 * copyins, we assume the specs passed to us are directly accessible.
	 */

	/*
	 * XXX right now we just store the raw, user supplied form of the
	 * address and protocol specs.  In the future we might want to reduce
	 * and compact this info sometime.
	 */
	if (addr != 0)
		memcpy(&chan->addr, addr, sizeof chan->addr);
	if (proto != 0)
		memcpy(&chan->proto, proto, sizeof chan->proto);
	/*
	 * XXX for CPU we only care about max threads right now.
	 */
	if (cpu != 0) {
		if (cpu->ttype != AN_TH_IMPLICIT)
			return ani_error_set(EINVAL,
					     "invalid to provide cpu spec "
					     "with explicit threads");
		chan->maxthreads = cpu->threads.im.maxthreads;
	}

	chan->ops = ops;
	chan->flags = ctype | CHAN_ISCLICK;
	chan->refs = 1;
	ani_ilock_init(&chan->lock);

	/*
	 * Do channel-type specific setup
	 */
	if (ctype == CHAN_IN) {
		/*
		 * XXX make sure we always have a demux object for the
		 * benefit of channel types that need to augment the demux.
		 *
		 * XXX this memory never gets freed right now.
		 */
		if (demux == 0) {
			demux = ani_mempool_alloc(ani_chan_mempool(chan),
						  sizeof *demux);
			if (demux == 0) {
				ani_ilock_deinit(&chan->lock);
				return ani_error_set(ENOMEM,
						     "ran out of memory while "
						     "allocating demux data");
			}
			err = ani_demuxkey_init(demux);
			assert(!ani_error_occured(err));
		}
		chan->tdata.in.demux = demux;
		chan->tdata.in.func = func;
		chan->tdata.in.farg = farg;

		if (npbufs) {
			err = ani_chan_pbuf_add(chan, pbufs, npbufs);
			if (ani_error_occured(err)) {
				ani_ilock_deinit(&chan->lock);
				return err;
			}
		}
	}

	/*
	 * Setup stuff that reflects its a Click component channel.
	 * Eventually, the component channels will consist of more
	 * than a single in/out chan, so maybe there will be lists
	 * of component in and out chans.
	 *
	 * This must be done before the channel-protocol specific call
	 * since it references the rawchan.
	 */
	chan->component.rawchan  = rawchan;
	rawchan->flags |= CHAN_WITHCLICK;

	/*
	 * Do channel-protocol specific setup.
	 * This will ask Click to create the graph.
	 */
	ani_chan_lock(chan);
	err = chan->ops->create(chan);
	if (ani_error_occured(err)) {
		ani_chan_unlock(chan);
		ani_ilock_deinit(&chan->lock);
		return err;
	}

	if (ctype == CHAN_IN) {
		/*
		 * Setup the inchan's recv function.
		 *
		 * Note that the inchan holds a reference to the cutchan to
		 * keep it alive since the inchan receive function accesses the
		 * cutchan.
		 */
		assert(ani_chan_type(rawchan) == CHAN_IN);
		ani_chan_reference(chan);
	}

	/*
	 * What about the rawchan flow? I'm leaving things as they are.
	 * Mike ...
	 */

	/*
	 * Finally link to the flow.
	 * We drop the channel lock first to adhere to our locking order.
	 */
	ani_chan_unlock(chan);
	ani_flow_addchannel(flow, chan);

	return ani_error_clear();
}
#endif /* CHAN_CLICK */
