/*
 * Copyright (c) 1999, 2000, 2001 The University of Utah and the Flux Group.
 * All rights reserved.
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */

/*
 * A "socket" channel driver.
 *
 * Currently we recognize only "socket:if/ipv4/udp".
 */
#include <nodeos/config.h>

#ifdef CHAN_SOCKET

#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <assert.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#ifdef OSKIT
#include <oskit/startup.h>
#include <oskit/com/listener.h>
#include <oskit/dev/dev.h>
#include <oskit/net/socket.h>
#include <oskit/dev/timer.h>
#endif

#include <nodeos/impl/an_error.h>
#include <nodeos/impl/ani_flow.h>
#include <nodeos/impl/ani_chan.h>
#include <nodeos/impl/ani_demuxkey.h>
#include <nodeos/impl/ani_thread.h>
#include <nodeos/impl/ani_threadpool.h>
#include "ani_debug.h"

static void chan_socket_init(void);
static int chan_socket_matches(an_chanproto_t *proto);
static ani_error_t chan_socket_create(ani_chan_t *chan);
static void chan_socket_destroy(ani_chan_t *chan);
static ani_error_t chan_socket_send(ani_chan_t *chan, ani_pbuf_t *pbuf);
static ani_error_t chan_socket_enable(ani_chan_t *chan);
static void chan_socket_disable(ani_chan_t *chan);

static char chan_socket_prefix[] = "socket";	/* XXX */

struct chanops chan_socket_ops = {
	chan_socket_init,	chan_socket_matches,
	chan_socket_create,	chan_socket_destroy,	chan_socket_send,
	chan_socket_enable,	chan_socket_disable,	0
};

/*
 * One structure per socket.
 * Sockets may be shared by multiple channels, with each channel having
 * a different demux for inChans.
 */
struct socketdata {
	struct socketdata  *next;
	short		   stype; /* SOCK_DGRAM or SOCK_STREAM */
	short		   ctype; /* inChan or outChan */
	an_udp4addrspec_t	   addr;  /* IP addr and UDP/TCP port */
	int		   fd;	  /* socket file descriptor */
	ani_demuxkeylist_t demux; /* list of demuxers for this socket */
	int		   refs;  /* refcount */
	int		   flags; /* flags */
	ani_thread_t	   thread;/* service thread */
	ani_icond_t	   cond;  /* synch with service thread */
	ani_ilock_t	   lock;  /* protects the demux list */
};

#define ANI_SD_THREAD_DYING	0x01
#define ANI_SD_THREAD_DEAD	0x02

static struct socketdata *socketlist;
static ani_ilock_t socketlock;

/*
 * Channel specific data (chan->data) is a pointer to a socketdata
 * structure (inChan) or a socket file descriptor (outChan).
 */

static ani_error_t chan_socketdata_find(int stype, an_udp4addrspec_t *caddr,
					int ctype,
					struct socketdata **out_sdata);
static void chan_socketdata_release(struct socketdata *sdata);
static void chan_socketdata_intr(void *arg, void *arg2, void *arg3);

static void
chan_socket_init(void)
{
	ani_ilock_init(&socketlock);

	/* XXX Put the following into os_init_socket()? */
#if defined(OSKIT) && !defined(OSKIT_UNIX)
	ani_ilock_lock(&socketlock);
	start_network_router_pthreads();
	ani_ilock_unlock(&socketlock);
#endif
}

static int
chan_socket_matches(an_chanproto_t *proto)
{
	int len;

	len = strlen(chan_socket_prefix);
	if (proto->len < len)
		return 0;

	return (strncmp(proto->data, chan_socket_prefix, len) == 0);
}

static ani_error_t
chan_socket_create(ani_chan_t *chan)
{
	struct socketdata *sdata;
	ani_error_t err;
	int stype, len;
	char *cp;

	assert(ani_chan_type(chan) != CHAN_CUT);

	cp = ani_chan_name(chan);
	len = ani_chan_namelen(chan);
	if (strncmp(cp, "socket:if/ipv4/udp", len) == 0)
		stype = SOCK_DGRAM;
#if 0
	else if (strncmp(cp, "socket:if/ipv4/tcp", len) == 0)
		stype = SOCK_STREAM;
#endif
	else
		return ani_error_set(EINVAL,
				     "invalid proto spec for socket channel");

	/*
	 * Find/create a socket of the appropriate type.
	 * Returns the socket data locked.
	 */
	err = chan_socketdata_find(stype, &chan->addr.udp4,
				   ani_chan_type(chan), &sdata);
	if (ani_error_occured(err))
		return err;

	ani_ilock_unlock(&sdata->lock);

	chan->data = sdata;
	return ani_error_clear();
}

static void
chan_socket_destroy(ani_chan_t *chan)
{
	struct socketdata *sdata;

	assert(chan->data);
	sdata = chan->data;

	ani_ilock_lock(&sdata->lock);
	chan_socketdata_release(sdata);
	/* lock got destroyed, no need to unlock. */
}

static ani_error_t
chan_socket_enable(ani_chan_t *chan)
{
	struct socketdata *sdata;
	ani_error_t err;

	assert(chan->data);

	if (ani_chan_type(chan) == CHAN_IN) {
		sdata = chan->data;
		ani_ilock_lock(&sdata->lock);
		err = ani_demuxkeylist_add(&sdata->demux, chan->tdata.in.demux);
		ani_ilock_unlock(&sdata->lock);
	} else
		err = ani_error_clear();

	return err;
}

static void
chan_socket_disable(ani_chan_t *chan)
{
	struct socketdata *sdata;

	assert(chan->data);

	if (ani_chan_type(chan) == CHAN_IN) {
		sdata = chan->data;
		ani_ilock_lock(&sdata->lock);
		ani_demuxkeylist_remove(&sdata->demux, chan->tdata.in.demux);
		ani_ilock_unlock(&sdata->lock);
	}
}

static ani_error_t
chan_socket_send(ani_chan_t *chan, ani_pbuf_t *pbuf)
{
	struct socketdata *sdata;
	int ierr;

	sdata = chan->data;
	ani_ilock_lock(&sdata->lock);
	DBG(DBG_SOCKFD, {
		ani_dprintf("write chan=%p fd=%d, pbuf=%p (data=%p, size=%d)\n",
			    chan, sdata->fd, pbuf,
			    ani_pbuf_data(pbuf), ani_pbuf_size(pbuf));
	});
	ani_pbuf_setbusy(pbuf);
	ierr = write(sdata->fd, ani_pbuf_data(pbuf), ani_pbuf_size(pbuf));
	ani_pbuf_clrbusy(pbuf);
	ani_ilock_unlock(&sdata->lock);
	if (ierr < 0) {
		CHANSTATS(chan->stats.dropnobufs++);	/* XXX */
		return ani_error_set(errno, "socket write failed");
	}
	return ani_error_clear();
}

static inline int
ipv4_addrcmp(an_udp4addrspec_t *a, an_udp4addrspec_t *b)
{
	if (a->sport == b->sport && a->saddr == b->saddr &&
	    a->dport == b->dport && a->daddr == b->daddr)
		return 0;

	return 1;
}

/*
 * Lookup or create a socket corresponding to the desired type/address.
 * Returns 0 on success with the socketdata struct locked.
 */
static ani_error_t
chan_socketdata_find(int stype, an_udp4addrspec_t *caddr, int ctype,
		     struct socketdata **out_sdata)
{
	struct socketdata *sdata;
	ani_error_t err;
	int sbsize;
	union {
		struct sockaddr gen;
		struct sockaddr_in ipv4;
	} saddr;

	/*
	 * Look for an existing socket
	 */
	ani_ilock_lock(&socketlock);
	for (sdata = socketlist; sdata != 0; sdata = sdata->next)
		if (sdata->ctype == ctype && sdata->stype == stype &&
		    ipv4_addrcmp(&sdata->addr, caddr) == 0) {
			ani_ilock_lock(&sdata->lock);
			ani_ilock_unlock(&socketlock);
			sdata->refs++;
			*out_sdata = sdata;
			return ani_error_clear();
		}
	ani_ilock_unlock(&socketlock);

	/*
	 * Otherwise, create a new one
	 * XXX who should we charge this to?
	 */
	sdata = ani_mempool_alloc(ani_mempool_rootpool(), sizeof *sdata);
	if (sdata == 0)
		return ani_error_set(ENOMEM,
				     "no memory for socket channel data");

	sdata->fd = socket(PF_INET, stype, 0);
	if (sdata->fd < 0) {
		ani_mempool_free(ani_mempool_rootpool(), sdata, sizeof *sdata);
		return ani_error_set(errno, "socket call failed");
	}
	DBG(DBG_SOCKFD, {
		ani_dprintf("created socket fd=%d\n",
			    sdata->fd);
	});

	/*
	 * Make socket buffer size as large as possible.
	 */
	sbsize = 32 * 1024;
	if (setsockopt(sdata->fd, SOL_SOCKET,
		       ctype == CHAN_IN ? SO_RCVBUF : SO_SNDBUF,
		       &sbsize, sizeof sbsize) < 0) {
		err = ani_error_set(errno, "setsockopt call failed");
		DBG(DBG_SOCKFD, {
			ani_dprintf("error, closing fd=%d\n",
				    sdata->fd);
		});
		close(sdata->fd);
		ani_mempool_free(ani_mempool_rootpool(), sdata, sizeof *sdata);
		return err;
	}

	/*
	 * We initialize and lock the sdata lock so we can use it to
	 * synchronize with any reader thread.  We set the ref count now
	 * so that we can use socketdata_release to cleanup on errors.
	 */
	ani_ilock_init(&sdata->lock);
	ani_ilock_lock(&sdata->lock);
	ani_icond_init(&sdata->cond);

	/*
	 * Convert NodeOS address info to BSD-style sockaddr
	 */
#if defined(__FreeBSD__) || defined(OSKIT)
	saddr.ipv4.sin_len = sizeof saddr.ipv4;
#endif
	saddr.ipv4.sin_family = AF_INET;
	saddr.ipv4.sin_port = caddr->dport;
	saddr.ipv4.sin_addr.s_addr = caddr->daddr;

	if (ctype == CHAN_IN) {
		DBG(DBG_SOCKFD, {
			ani_dprintf("bind fd=%d addr=%#x:%p\n",
				    sdata->fd, caddr->daddr, caddr->dport);
		});
		if (bind(sdata->fd, &saddr.gen, sizeof saddr.gen) < 0) {
			err = ani_error_set(errno, "bind call failed");
			goto error;
		}

		err = ani_rootflow_startthread(&sdata->thread,
					       chan_socketdata_intr, sdata,
					       0, 0);
		if (ani_error_occured(err))
			goto error;
	} else {
		DBG(DBG_SOCKFD, {
			ani_dprintf("connect fd=%d addr=%#x:%p\n",
				    sdata->fd, caddr->daddr, caddr->dport);
		});
		if (connect(sdata->fd, &saddr.gen, sizeof saddr.gen) < 0) {
			err = ani_error_set(errno, "connect call failed");
			goto error;
		}
	}

	/*
	 * Initialize the rest of the socket data
	 *
	 * Note that we acquire the socketlist lock out of order here.
	 * This is safe since no one else could hold our lock yet and cause
	 * deadlock.
	 */
	sdata->stype = stype;
	sdata->ctype = ctype;
	sdata->addr = *caddr;
	ani_demuxkeylist_init(&sdata->demux);
	sdata->refs = 1;
	sdata->flags = 0;
	ani_ilock_lock(&socketlock);
	sdata->next = socketlist;
	socketlist = sdata;
	ani_ilock_unlock(&socketlock);

	*out_sdata = sdata;
	return ani_error_clear();

error:
	DBG(DBG_SOCKFD, {
		ani_dprintf("error, closing fd=%d\n",
			    sdata->fd);
	});
	close(sdata->fd);
	ani_icond_deinit(&sdata->cond);
	ani_ilock_unlock(&sdata->lock);
	ani_ilock_deinit(&sdata->lock);
	ani_mempool_free(ani_mempool_rootpool(), sdata, sizeof *sdata);
	return err;
}

/*
 * Drop a reference to a socketdata struct.
 * Called with the sdata locked.  If no more refs, the sdata is destroyed.
 * Otherwise, the lock is dropped.
 */
static void
chan_socketdata_release(struct socketdata *sdata)
{
	struct socketdata **sdp;

	if (--sdata->refs == 0) {
		/*
		 * If we cannot grab the list lock, do the funky lock dance.
		 * Drop our lock to get the global lock then reaquire ours.
		 * Then check to make sure someone didn't latch on to our
		 * sdata in the meantime.
		 */
		if (ani_ilock_trylock(&socketlock)) {
			ani_ilock_unlock(&sdata->lock);
			ani_ilock_lock(&socketlock);
			ani_ilock_lock(&sdata->lock);
			if (sdata->refs != 0) {
				ani_ilock_unlock(&sdata->lock);
				ani_ilock_unlock(&socketlock);
				return;
			}
		}

		/*
		 * Locate and remove from the list
		 */
		for (sdp = &socketlist; *sdp != 0; sdp = &(*sdp)->next)
			if (*sdp == sdata)
				break;
		assert(*sdp != 0);
		*sdp = sdata->next;
		ani_ilock_unlock(&socketlock);

		/*
		 * For input sockets, wait for the service thread to terminate.
		 */
		if (sdata->ctype == CHAN_IN) {
			sdata->flags |= ANI_SD_THREAD_DYING;
			while ((sdata->flags & ANI_SD_THREAD_DEAD) == 0)
				ani_icond_wait(&sdata->cond, &sdata->lock);
		}
		DBG(DBG_SOCKFD, {
			ani_dprintf("release, closing fd=%d\n",
				    sdata->fd);
		});
		close(sdata->fd);
		ani_icond_deinit(&sdata->cond);
		ani_ilock_unlock(&sdata->lock);
		ani_ilock_deinit(&sdata->lock);
		ani_mempool_free(ani_mempool_rootpool(), sdata, sizeof *sdata);
	} else
		ani_ilock_unlock(&sdata->lock);
}

/*
 * We wake up the receiver thread peroidically to see if the socket needs
 * to be shutdown (i.e., all using channels have been destroyed).
 * As always, there seems to be no way to do this that works on all platforms.
 * For most systems we use select to wait and only do the receive if there
 * is something to read.  For the OSKit, the select implementation is dubious
 * so we set a socket timeout instead.  It would be nice (and the cleanest)
 * to use a socket timeout all around, but Linux and Solaris don't implement
 * it and it doesn't seem to work with the threads package in FreeBSD 3.4.
 */
#if !defined(OSKIT) || defined(OSKIT_UNIX)
#define USE_SELECT
#endif

static struct timeval timeo;
#ifdef USE_SELECT
static fd_set sofdset;
#endif

/*
 * Receiver thread for inChan sockets.
 * Applies the demux filters for the socket to determine a channel to
 * deliver to and fires up a channel thread to deal with it.
 *
 * XXX for TCP support, do listen and accept here.
 */
static void
chan_socketdata_intr(void *arg, void *arg2, void *arg3)
{
	struct socketdata *sdata = arg;
	void *bdata;
	ssize_t bsize, size;
	ani_error_t err;
	int s, msize;
	ani_pbuf_t *rbuf, *pbuf;
	ani_chan_t *chan;

	/*
	 * Locking forces synchronization with our creator
	 */
	ani_ilock_lock(&sdata->lock);

	/*
	 * Set a receive timeout so we break out of the recvfrom once in
	 * a while and can check for socket death.
	 */
	s = sdata->fd;
#if !defined(USE_SELECT)
	{
		int ierr;

		ierr = setsockopt(s, SOL_SOCKET, SO_RCVTIMEO,
				  &timeo, sizeof timeo);
		assert(ierr == 0);
	}
#endif
	ani_ilock_unlock(&sdata->lock);

	/*
	 * Create an initial read buffer.  We swap this buffer with
	 * empty ones from the socket's channels when we hand off packets.
	 */
	rbuf = ani_mempool_alloc(ani_mempool_rootpool(), sizeof *rbuf);
	assert(rbuf != 0);
	err = ani_pbuf_init(rbuf, ani_flow_current(), 0,
			    NODEOS_CHAN_BUFSIZE, 0, NODEOS_CHAN_BUFSIZE);
	assert(!ani_error_occured(err));
	bdata = ani_pbuf_data(rbuf);
	bsize = ani_pbuf_size(rbuf);

	while (1) {
		chan = 0;
		
		DBG(DBG_SOCKFD, {
			ani_dprintf("intr, blocking on fd=%d\n", s);
		});
#ifdef USE_SELECT
		assert(s < sizeof(sofdset)*8);
		FD_ZERO(&sofdset); 
		FD_SET(s, &sofdset);
		timeo.tv_sec = 0;
		timeo.tv_usec = 500000;	/* 500ms */
		
		/*
		 * Only one fd in the set, so set the nfds parameter
		 * to +1 that fd.
		 */
		size = select(s+1, &sofdset, 0, 0, &timeo);

		assert(size == 0 || size == 1 || size == -1);
		if (size == 1)
			size = recvfrom(s, bdata, bsize, 0, 0, 0);
		else if (size == 0)
			errno = EAGAIN;
#else
		size = recvfrom(s, bdata, bsize, 0, 0, 0);
#endif
		ani_ilock_lock(&sdata->lock);

		DBG(DBG_SOCKFD, {
			ani_dprintf("intr, woke up on fd=%d (dead: %s, timeout: %s, error: %s)\n",
				    s,
				    (sdata->flags&ANI_SD_THREAD_DYING)?"yes":"no",
				    (size <= 0 && errno == EAGAIN)?"yes":"no",
				    (size <= 0 && errno != EAGAIN)?"yes":"no");
		});

		/*
		 * The socket is being destroyed, exit now.
		 */
		if (sdata->flags & ANI_SD_THREAD_DYING) {
			ani_ilock_unlock(&sdata->lock);
			errno = EAGAIN;
			break;
		}

		/*
		 * Socket read got an error or EOF.
		 * If error was a periodic timeout, ignore it and keep reading.
		 * Otherwise, exit now.
		 */
		if (size <= 0) {
			ani_ilock_unlock(&sdata->lock);
			if (errno == EAGAIN)
				continue;
			break;
		}

		/*
		 * Valid packet.  Locate the target channel for this data.
		 * If none, just drop it and continue.
		 */
		msize = (int)size;
		chan = ani_demuxkeylist_apply(&sdata->demux, bdata, &msize);
		ani_ilock_unlock(&sdata->lock);
		if (chan == 0)
			continue;

		/*
		 * Allocate a packet buffer from the channel.
		 * If none are available, just drop the packet.
		 */
		ani_chan_lock(chan);
		assert((chan->flags & CHAN_DYING) == 0);
		err = ani_chan_pbuf_alloc(chan, NODEOS_CHAN_BUFSIZE, &pbuf);
		if (ani_error_occured(err)) {
			ani_chan_unlock(chan);
			continue;
		}

		/*
		 * Swap the buffer obtained from the channel with our buffer.
		 * The channel gets the full buffer in exchange for one of its
		 * empty ones that we will use for the next read.
		 */
		ani_pbuf_setsize(rbuf, size, 0);
		ani_pbuf_swap(rbuf, pbuf);

		/*
		 * Finally, fire up a handler thread.
		 */
		ani_chan_thread_start(chan, pbuf, msize);

		ani_chan_unlock(chan);

		/*
		 * Reset the buffer pointer/size since we have a new
		 * physical buffer.
		 */
		bdata = ani_pbuf_data(rbuf);
		bsize = ani_pbuf_size(rbuf);

		/*
		 * Yield here so that the packet threads have a chance to
		 * do their thing and return the buffer to the inchan.
		 * Otherwise we might just use up all the available buffers
		 * and have to drop the rest of the packets.
		 */
		ani_thread_yield();
	}

	/*
	 * We have broken out of the recvfrom.  Most likely because the
	 * socket was closed.  Free the read buffer, mark that we are dead,
	 * inform any interested party, and exit.
	 */
	ani_pbuf_deinit(rbuf);
	ani_ilock_lock(&sdata->lock);
	sdata->flags |= ANI_SD_THREAD_DEAD;
	if (sdata->flags & ANI_SD_THREAD_DYING)
		ani_icond_broadcast(&sdata->cond);
	else
		/* unexpected exit */
		printf("0x%x/%d: receive thread died: %s\n",
		       (unsigned int)ntohl(sdata->addr.daddr),
		       ntohs(sdata->addr.dport),
		       errno ? strerror(errno) : "EOF");
	ani_ilock_unlock(&sdata->lock);
}

#endif /* CHAN_SOCKET */
