/*
 * Copyright (c) 2000, 2001 The University of Utah and the Flux Group.
 * All rights reserved.
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */

/*
 * Stdio channel driver.
 * Supports reading raw packets from stdin (for anetd).
 * Could also transmit on stdout.
 *
 * Note that we don't actually use stdio routines.
 */
#include <nodeos/config.h>

#ifdef CHAN_STDIO

#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <assert.h>
#include <sys/types.h>
#include <sys/socket.h>
#ifdef OSKIT_UNIX
#include <termios.h>
#endif

#include <nodeos/impl/an_error.h>
#include <nodeos/impl/ani_flow.h>
#include <nodeos/impl/ani_chan.h>
#include <nodeos/impl/ani_demuxkey.h>
#include <nodeos/impl/ani_thread.h>
#include <nodeos/impl/ani_threadpool.h>
#include "ani_debug.h"
#include "os.h"

static void chan_stdio_init(void);
static int chan_stdio_matches(an_chanproto_t *proto);
static ani_error_t chan_stdio_create(ani_chan_t *chan);
static void chan_stdio_destroy(ani_chan_t *chan);
static ani_error_t chan_stdio_send(ani_chan_t *chan, ani_pbuf_t *pbuf);
static ani_error_t chan_stdio_enable(ani_chan_t *chan);
static void chan_stdio_disable(ani_chan_t *chan);

static char chan_stdio_prefix[] = "stdio";	/* XXX */

struct chanops chan_stdio_ops = {
	chan_stdio_init,	chan_stdio_matches,
	chan_stdio_create,	chan_stdio_destroy,	chan_stdio_send,
	chan_stdio_enable,	chan_stdio_disable,	0
};

struct stdiodata {
	ani_demuxkeylist_t demux; /* list of demuxers for this socket */
	int		   refs;  /* refcount */
	int		   flags; /* flags */
	ani_thread_t	   thread;/* service thread */
	ani_icond_t	   cond;  /* synch with service thread */
	ani_ilock_t	   lock;  /* protects the demux list */
};
#define ANI_SD_THREAD_DYING	0x01
#define ANI_SD_THREAD_DEAD	0x02

static struct stdiodata stdindata;
static void chan_stdio_intr(void *arg, void *arg2, void *arg3);

static void
chan_stdio_init(void)
{
	ani_ilock_init(&stdindata.lock);
	ani_icond_init(&stdindata.cond);
#ifdef OSKIT_UNIX
	/*
	 * XXX OSKit assumes stdin is coming from a console tty device
	 * so we have to turn off tty processing.
	 */
	{
		struct termios t;

		if (tcgetattr(STDIN_FILENO, &t) == 0) {
			t.c_oflag &= ~OPOST;
			t.c_lflag &= ~(ICANON|ECHO);
			tcsetattr(STDIN_FILENO, TCSANOW, &t);
		}
	}
#endif
}

static int
chan_stdio_matches(an_chanproto_t *proto)
{
	int len;

	len = strlen(chan_stdio_prefix);
	if (proto->len < len)
		return 0;

	return (strncmp(proto->data, chan_stdio_prefix, len) == 0);
}

static ani_error_t
chan_stdio_create(ani_chan_t *chan)
{
	ani_error_t err;
	char *cp;
	int len;

	if (ani_chan_type(chan) != CHAN_IN)
		return ani_error_set(EINVAL, "only support stdio in channels");

	cp = ani_chan_name(chan);
	len = ani_chan_namelen(chan);
	if (strncmp(cp, "stdio:anep", len) != 0)
		return ani_error_set(EINVAL, "only support ANEP stdio channels");

	if (++stdindata.refs == 1) {
		err = ani_rootflow_startthread(&stdindata.thread,
					       chan_stdio_intr, 0, 0, 0);
		if (ani_error_occured(err))
			return err;
	}

	return ani_error_clear();
}

static void
chan_stdio_destroy(ani_chan_t *chan)
{
	assert_ani_chan_valid(chan);
	assert(ani_chan_type(chan) == CHAN_IN);

	ani_ilock_lock(&stdindata.lock);
	if (--stdindata.refs == 0) {
		stdindata.flags |= ANI_SD_THREAD_DYING;
		while ((stdindata.flags & ANI_SD_THREAD_DEAD) == 0)
			ani_icond_wait(&stdindata.cond, &stdindata.lock);
		stdindata.flags = 0;
		ani_thread_deinit(&stdindata.thread);
	}
	ani_ilock_unlock(&stdindata.lock);
}

static ani_error_t
chan_stdio_enable(ani_chan_t *chan)
{
	ani_error_t err;

	assert_ani_chan_valid(chan);
	assert(ani_chan_type(chan) == CHAN_IN);

	ani_ilock_lock(&stdindata.lock);
	err = ani_demuxkeylist_add(&stdindata.demux, chan->tdata.in.demux);
	ani_ilock_unlock(&stdindata.lock);

	return err;
}

static void
chan_stdio_disable(ani_chan_t *chan)
{
	assert_ani_chan_valid(chan);
	assert(ani_chan_type(chan) == CHAN_IN);

	ani_ilock_lock(&stdindata.lock);
	ani_demuxkeylist_remove(&stdindata.demux, chan->tdata.in.demux);
	ani_ilock_unlock(&stdindata.lock);
}

static ani_error_t
chan_stdio_send(ani_chan_t *chan, ani_pbuf_t *pbuf)
{
	panic("chan_stdio_send called!");
	return ani_error_set(ENODEV, "chan_stdio_send not implemented");
}

/*
 * We wake up the receiver thread peroidically to see if the receiving thread
 * needs to be shutdown (i.e., all using channels have been destroyed).
 */
static struct timeval timeo;
static fd_set fdset;

/*
 * Loop reading til we get the specified number of bytes or til all channels
 * have died.  Returns zero if the indicated data has been read, -1 on EOF or
 * an error ow.
 */
static int
chan_stdio_readbytes(void *buf, size_t nbytes)
{
	int fd = STDIN_FILENO;
	size_t size;
	int err = 0;

	/*
	 * Do all the same crap we do for a socket channel.
	 * Set a select timeout so we check periodically for channel death.
	 *
	 * XXX this should get broken out into fd_chan_* for both stdin
	 * and sockets.
	 */
	while (nbytes > 0) {
		/*
		 * Must reinitialize fdset for each call
		 */
		FD_ZERO(&fdset); 
		FD_SET(fd, &fdset);

		/*
		 * and the timeout (for Linux)
		 */
		timeo.tv_sec = 0;
		timeo.tv_usec = 500000;	/* 500ms */
		
		/*
		 * Only one fd in the set, so set the nfds parameter
		 * to +1 that fd.
		 */
		switch (select(fd+1, &fdset, 0, 0, &timeo)) {
		case 1:
			/*
			 * Data is available, read as much as was requested
			 */
			size = read(fd, buf, nbytes);
			err = size < 0 ? errno : 0;
			break;
		case 0:
			/*
			 * We got a timeout, go check for channel death
			 */
			size = 0;
			err = EAGAIN;
			break;
		default:
			/*
			 * We got an error, we will exit for sure
			 */
			size = -1;
			err = errno;
			break;
		}

		/*
		 * Check to see if any channels still exist.
		 * We do this without locking since the DYING flag won't
		 * be cleared while we are still alive.
		 */
		if (stdindata.flags & ANI_SD_THREAD_DYING) {
			err = ECHILD; /* XXX */
			break;
		}

		/*
		 * Got an error or EOF.
		 * If error was a periodic timeout, ignore it and keep reading.
		 * Otherwise, return an error so we can exit.
		 */
		if (size <= 0) {
			if (err == EAGAIN)
				continue;
			err = -1;
			break;
		}

		nbytes -= size;
		buf += size;
	}

	return err;
}

/*
 * XXX
 */
struct anephdr {
	char ah_version;
	char ah_flags;           
	short ah_typeid;
	short ah_hlen;
	short ah_plen;
};

/*
 * Receiver thread for stdio inChan.
 * Applies the demux filters to determine a channel to
 * deliver to and fires up a channel thread to deal with it.
 */
static void
chan_stdio_intr(void *arg, void *arg2, void *arg3)
{
	void *bdata;
	ssize_t bsize, size;
	ani_error_t err;
	int ierr, msize;
	ani_pbuf_t *rbuf, *pbuf;
	ani_chan_t *chan;

	/*
	 * Locking forces synchronization with our creator
	 */
	ani_ilock_lock(&stdindata.lock);
	ani_ilock_unlock(&stdindata.lock);

	/*
	 * Create an initial read buffer.  We swap this buffer with
	 * empty ones from stdin's channels when we hand off packets.
	 */
	rbuf = ani_mempool_alloc(ani_mempool_rootpool(), sizeof *rbuf);
	assert(rbuf != 0);
	err = ani_pbuf_init(rbuf, ani_flow_current(), 0,
			    NODEOS_CHAN_BUFSIZE, 0, NODEOS_CHAN_BUFSIZE);
	assert(!ani_error_occured(err));
	bdata = ani_pbuf_data(rbuf);
	bsize = ani_pbuf_size(rbuf);

	while (1) {
		chan = 0;
		
		/*
		 * Read enough of the packet to find the ANEP total length.
		 */
		ierr = chan_stdio_readbytes(bdata, sizeof(struct anephdr));
		if (ierr != 0)
			break;

		size = ntohs(((struct anephdr *)bdata)->ah_plen);
		assert(size >= sizeof(struct anephdr));
		assert(size <= bsize);

		/*
		 * Now read the rest of the packet
		 */
		ierr = chan_stdio_readbytes(bdata + sizeof(struct anephdr),
					   size - sizeof(struct anephdr));
		if (ierr != 0)
			break;

		/*
		 * Valid packet.  Locate the target channel for this data.
		 * If none, just drop it and continue.
		 */
		msize = (int)size;
		ani_ilock_lock(&stdindata.lock);
		chan = ani_demuxkeylist_apply(&stdindata.demux, bdata, &msize);
		ani_ilock_unlock(&stdindata.lock);
		if (chan == 0)
			continue;

		/*
		 * Allocate a packet buffer from the channel.
		 * If none are available, just drop the packet.
		 */
		ani_chan_lock(chan);
		assert((chan->flags & CHAN_DYING) == 0);
		err = ani_chan_pbuf_alloc(chan, NODEOS_CHAN_BUFSIZE, &pbuf);
		if (ani_error_occured(err)) {
			ani_chan_unlock(chan);
			continue;
		}

		/*
		 * Swap the buffer obtained from the channel with our buffer.
		 * The channel gets the full buffer in exchange for one of its
		 * empty ones that we will use for the next read.
		 */
		ani_pbuf_setsize(rbuf, size, 0);
		ani_pbuf_swap(rbuf, pbuf);

		/*
		 * Finally, fire up a handler thread.
		 */
		ani_chan_thread_start(chan, pbuf, msize);

		ani_chan_unlock(chan);

		/*
		 * Reset the buffer pointer/size since we have a new
		 * physical buffer.
		 */
		bdata = ani_pbuf_data(rbuf);
		bsize = ani_pbuf_size(rbuf);
	}

	/*
	 * We have broken out of the read.  Most likely because stdin
	 * was closed.  Free the read buffer, mark that we are dead,
	 * inform any interested party, and exit.
	 */
	ani_pbuf_deinit(rbuf);
	ani_ilock_lock(&stdindata.lock);
	stdindata.flags |= ANI_SD_THREAD_DEAD;
	if (stdindata.flags & ANI_SD_THREAD_DYING)
		ani_icond_broadcast(&stdindata.cond);
	else
		/* unexpected exit */
		printf("stdin: receive thread died: %s\n",
		       ierr == -1 ? "EOF" : strerror(ierr));
	ani_ilock_unlock(&stdindata.lock);
}

#endif
