/*
 * Copyright (c) 1999, 2000, 2001 The University of Utah and the Flux Group.
 * All rights reserved.
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */
#include <string.h>

#include <nodeos/impl/ani_chan.h>
#include <nodeos/impl/ani_flow.h>
#include <nodeos/impl/ani_demuxkey.h>
#include <nodeos/impl/ani_mempool.h>
#include <nodeos/impl/ani_threadpool.h>
#include <nodeos/impl/ani_if.h>
#include <nodeos/impl/ani_error.h>
#include <nodeos/impl/an_pbuf.h>
#include <nodeos/config.h>

#include "os.h"

#ifdef CHAN_CLOCK
extern struct chanops chan_clock_ops;
#endif
#ifdef CHAN_ETH
extern struct chanops chan_eth_ops;
#endif
#ifdef CHAN_IF
extern struct chanops chan_if_ops;
#endif
#ifdef CHAN_SOCKET
extern struct chanops chan_socket_ops;
#endif
#ifdef CHAN_STDIO
extern struct chanops chan_stdio_ops;
#endif
#ifdef CHAN_CLICK
extern struct chanops chan_click_ops;
#endif

/*
 * Protocol-specific processing modules.
 */
static struct chanops *allchans[] =
{
#ifdef CHAN_CLOCK
	&chan_clock_ops,
#endif
#ifdef CHAN_ETH
	&chan_eth_ops,
#endif
#ifdef CHAN_IF
	&chan_if_ops,
#endif
#ifdef CHAN_SOCKET
	&chan_socket_ops,
#endif
#ifdef CHAN_STDIO
	&chan_stdio_ops,
#endif
#ifdef CHAN_CLICK
	&chan_click_ops,
#endif
	0
};

void
ani_chan_setup(void)
{
	struct chanops **ops;

	for (ops = allchans; *ops != 0; ops++)
		if ((*ops)->init != 0)
			(*ops)->init();

	ani_if_setup();
}

/*
 * ``Creates'' a channel object.
 * It acutally initializes an existing piece of memory (``chan'') to make it
 * a channel object.  The ideal here is that we allocate no memory, the
 * caller provides us with pointers to everything we need.  This includes
 * not only the channel struct, but the proto, addr, and demux structs as
 * well; i.e., anything that is passed by reference.
 *
 * XXX we still allocate channel protocol-specific data at this level.
 */
ani_error_t
ani_chan_init(ani_chan_t *chan, int ctype, ani_flow_t *flow,
	      ani_threadpool_t *tpool,
	      an_chanproto_t *proto, an_chanaddr_t *addr,
	      ani_demuxkey_t *demux, an_cpuspec_t *cpu,
	      ani_pbuf_t **pbufs, int npbufs, an_netspec_t *net,
	      an_chanrecvfunc_t func, void *farg)
{
	struct chanops **ops;
	ani_error_t err;

	assert(ctype != CHAN_CUT);

	/*
	 * See if we know how to handle the desired protocol processing.
	 */
	for (ops = allchans; *ops != 0; ops++)
		if ((*ops)->matches(proto))
			break;
	
	if (*ops == 0)
		return ani_error_set(EINVAL,
				     "No channel available for protocol %s",
				     proto->data);

	memset(chan, 0, sizeof *chan);

	/*
	 * XXX set this now so that functions we call know who to charge
	 * for memory.  We officially link into the flow after everything
	 * has been setup.
	 */
	chan->flow = flow;

	/*
	 * XXX should we hold a ref to the pool here?
	 * Don't think so.  Threadpool cannot go away as long as the flow
	 * is alive.  If the flow has been destroyed, so have all its channels.
	 */
	chan->tpool = tpool;

	/*
	 * Record the resource specifications.
	 * It is up to the entry layer to perform any cross-address space
	 * copyins, we assume the specs passed to us are directly accessible.
	 */

	/*
	 * XXX right now we just store the raw, user supplied form of the
	 * address and protocol specs.  In the future we might want to reduce
	 * and compact this info sometime.
	 */
	if (addr != 0)
		memcpy(&chan->addr, addr, sizeof chan->addr);
	if (proto != 0)
		memcpy(&chan->proto, proto, sizeof chan->proto);
	/*
	 * XXX for CPU we only care about max threads right now.
	 */
	if (cpu != 0) {
		if (cpu->ttype != AN_TH_IMPLICIT)
			return ani_error_set(EINVAL,
					     "invalid to provide cpu spec "
					     "with explicit threads");
		chan->maxthreads = cpu->threads.im.maxthreads;
	}

	chan->ops = *ops;
	chan->flags = ctype | CHAN_ISRAW;
	chan->refs = 1;
	ani_ilock_init(&chan->lock);

	/*
	 * Do channel-type specific setup
	 */
	if (ctype == CHAN_IN) {
		/*
		 * XXX make sure we always have a demux object for the
		 * benefit of channel types that need to augment the demux.
		 *
		 * XXX this memory never gets freed right now.
		 */
		if (demux == 0) {
			demux = ani_mempool_alloc(ani_chan_mempool(chan),
						  sizeof *demux);
			if (demux == 0) {
				ani_ilock_deinit(&chan->lock);
				return ani_error_set(ENOMEM,
						     "ran out of memory while "
						     "allocating demux data");
			}
			err = ani_demuxkey_init(demux);
			assert(!ani_error_occured(err));
		}

		/*
		 * Associate the demuxkey with the channel.
		 */
		err = ani_demuxkey_addchannel(demux, chan);
		if (ani_error_occured(err)) {
			ani_ilock_deinit(&chan->lock);
			return err;
		}
		chan->tdata.in.demux = demux;

		ani_pbuflist_init(&chan->tdata.in.rbufs);
		ani_pbuflist_init(&chan->tdata.in.bbufs);
		ani_pbuflist_init(&chan->tdata.in.qbufs);
		chan->tdata.in.func = func;
		chan->tdata.in.farg = farg;

		if (npbufs) {
			err = ani_chan_pbuf_add(chan, pbufs, npbufs);
			if (ani_error_occured(err)) {
				ani_demuxkey_removechannel(demux, chan);
				ani_ilock_deinit(&chan->lock);
				return err;
			}
		}
	} else if (ctype == CHAN_OUT) {
		chan->tdata.out.mbs = net ? net->bandwidth : 0;
		/* XXX ensure mbits/s fits in the flow's allocation */
	}

	/*
	 * Do channel-protocol specific setup
	 */
	ani_chan_lock(chan);
	err = chan->ops->create(chan);
	if (ani_error_occured(err)) {
		ani_chan_unlock(chan);
		ani_demuxkey_removechannel(demux, chan);
		ani_ilock_deinit(&chan->lock);
		return err;
	}

	/*
	 * Finally link to the flow.
	 * We drop the channel lock first to adhere to our locking order.
	 */
	ani_chan_unlock(chan);
	ani_flow_addchannel(flow, chan);

	return ani_error_clear();
}

static void
ani_chan_deinit_internal(ani_chan_t *chan)
{
	assert_ani_chan_valid(chan);
	assert(ani_chan_islocked(chan));

	/*
	 * Disable the channel immediately to avoid races with incoming
	 * packet interrupt handlers (and so that those handlers don't have
	 * to check for a race).
	 */
	if (ani_chan_type(chan) == CHAN_IN)
		chan->ops->disable(chan);

	/*
	 * Note that the channel is dying.  Needed in case threads block
	 * on the channel lock while we are destroying the channel.
	 */
	chan->flags |= CHAN_DYING;

	/*
	 * Once the channel is disabled and marked as dying,
	 * we can free up in-channel packet buffers and disassociate
	 * any demux.
	 */
	if (ani_chan_type(chan) == CHAN_IN) {
		ani_chan_pbuf_removeall(chan);
		ani_demuxkey_removechannel(chan->tdata.in.demux, chan);
	}

	/*
	 * For cut "pseudo" channels, we destroy the constituent channels.
	 * Otherwise we call the channel-protocol specific shutdown.
	 */
	if (ani_chan_type(chan) == CHAN_CUT) {
		ani_chan_t *cchan;

		cchan = chan->tdata.cut.inchan;
		assert(chan != cchan);
		ani_chan_lock(cchan);
		ani_chan_deinit_internal(cchan);
		ani_chan_unlock(cchan);
		ani_chan_release(cchan);

		cchan = chan->tdata.cut.outchan;
		assert(chan != cchan);
		ani_chan_lock(cchan);
		ani_chan_deinit_internal(cchan);
		ani_chan_unlock(cchan);
		ani_chan_release(cchan);
	} else
		chan->ops->destroy(chan);
}

ani_error_t
ani_chan_deinit(ani_chan_t *chan)
{
	assert_ani_chan_valid(chan);

	/*
	 * XXX hack, don't allow explicit destruction of cut-channel
	 * components.
	 */
	if (chan->flags & CHAN_WITHCUT)
		return ani_error_set(EINVAL,
				     "cannot destroy a cut-channel component");
	
	/*
	 * Remove from the flow before we lock the channel
	 */
	ani_flow_removechannel(chan->flow, chan);

	ani_chan_lock(chan);
	ani_chan_deinit_internal(chan);
	ani_chan_unlock(chan);
	ani_chan_release(chan);

	return ani_error_clear();
}

void
ani_chan_cleanup(ani_chan_t *chan)
{
	assert((chan->flags & CHAN_DYING) != 0);
	assert(ani_chan_islocked(chan) == 0);

	/*
	 * XXX if this is the inchan side of a cutchan, we need to
	 * release our reference now.
	 *
	 * XXX this is wrong for Click-based cutchans where the inchan
	 * receive func argument is not the cutchan itself.
	 */
	if ((chan->flags & CHAN_WITHCUT) && ani_chan_type(chan) == CHAN_IN) {
		ani_chan_t *cchan = (ani_chan_t *)chan->tdata.in.farg;

		assert_ani_chan_valid(cchan);
		ani_chan_release(cchan);
	}

	ani_ilock_deinit(&chan->lock);
}

ani_error_t
ani_chan_enable(ani_chan_t *chan)
{
	ani_error_t err;

	assert_ani_chan_valid(chan);

	ani_chan_lock(chan);
	if (ani_chan_type(chan) == CHAN_IN && chan->tdata.in.func != 0)
		err = chan->ops->enable(chan);
	else
		err = ani_error_set(EINVAL,
				    "not an InChannel or "
				    "no receive packet callback defined");
	ani_chan_unlock(chan);

	return err;
}

void
ani_chan_disable(ani_chan_t *chan)
{
	assert_ani_chan_valid(chan);

	ani_chan_lock(chan);
	chan->ops->disable(chan);
	ani_chan_unlock(chan);
}

ani_error_t
ani_chan_send(ani_chan_t *chan, ani_pbuf_t *pbuf)
{
	ani_error_t err;

	assert_ani_chan_valid(chan);

	if (ani_chan_type(chan) != CHAN_OUT)
		return ani_error_set(EINVAL, "not an outchan");
	if (chan->flags & CHAN_DYING)
		return ani_error_set(EBADF, /* XXX */
				     "channel is dying");
	if (ani_pbuf_isbusy(pbuf))
		return ani_error_set(EBUSY, "pbuf already in use");

	ani_chan_lock(chan);
	if (chan->ops->send) {
		CHANSTATS(unsigned int psize = ani_pbuf_size(pbuf));
		err = chan->ops->send(chan, pbuf);
		if (!ani_error_occured(err)) {
			CHANSTATS(chan->stats.outpackets++);
			CHANSTATS(chan->stats.outbytes += psize);
		} else {
			CHANSTATS(chan->stats.droppackets++);
			CHANSTATS(chan->stats.dropbytes += psize);
		}
	} else
		err = ani_error_set(EINVAL, "cannot send on channel type");
	ani_chan_unlock(chan);

	return err;
}

/*
 * Pass on (requeue) a pbuf on an inchan.
 * On success, the caller still has a pbuf but its contents are undefined.
 * On failure, the pbuf is unaffected.
 */
ani_error_t
ani_chan_handoff(ani_chan_t *chan, ani_pbuf_t *pbuf)
{
	ani_error_t err;
	ani_pbuf_t *npbuf;
	int msize;

	assert_ani_chan_valid(chan);

	if (ani_chan_type(chan) != CHAN_IN)
		return ani_error_set(EINVAL, "not an inchan");
	if (chan->flags & CHAN_DYING)
		return ani_error_set(EBADF, /* XXX */
				     "channel is dying");
	if (ani_pbuf_isbusy(pbuf))
		return ani_error_set(EBUSY, "pbuf already in use");

	ani_chan_lock(chan);

	/*
	 * Apply the channel's demux to ensure the packet "belongs"
	 * and to compute the offset.
	 */
	msize = ani_demuxkey_match(chan->tdata.in.demux,
				   ani_pbuf_data(pbuf), ani_pbuf_size(pbuf));
	if (msize < 0) {
		ani_chan_unlock(chan);
		return ani_error_set(EINVAL, "square peg, round hole");
	}

	/*
	 * Swap packet into a buffer for the target channel.
	 */
	err = ani_chan_pbuf_alloc(chan, NODEOS_CHAN_BUFSIZE, &npbuf);
	if (ani_error_occured(err)) {
		ani_chan_unlock(chan);
		return err;
	}
	ani_pbuf_swap(pbuf, npbuf);

	/*
	 * Queue the packet, firing off a channel thread if necessary.
	 */
	ani_chan_thread_start(chan, npbuf, msize);

	ani_chan_unlock(chan);

	return err;

}

an_chanstats_t *
ani_chan_stats(ani_chan_t *chan)
{
	assert_ani_chan_valid(chan);

	if (ani_chan_type(chan) == CHAN_CUT) {
		/* XXX copy stats out of constituent channels */
		chan->stats.inpackets =
			chan->tdata.cut.inchan->stats.inpackets;
		chan->stats.inbytes =
			chan->tdata.cut.inchan->stats.inbytes;

		/*
		 * Packets can be dropped on input, output or as the
		 * result of protocol processing.
		 */
		chan->stats.droppackets +=
			chan->tdata.cut.inchan->stats.droppackets +
			chan->tdata.cut.outchan->stats.droppackets;
		chan->stats.dropbytes +=
			chan->tdata.cut.inchan->stats.dropbytes +
			chan->tdata.cut.outchan->stats.dropbytes;
		chan->stats.dropnobufs =
			chan->tdata.cut.inchan->stats.dropnobufs +
			chan->tdata.cut.outchan->stats.dropnobufs;
		chan->stats.dropfailedbuf =
			chan->tdata.cut.inchan->stats.dropfailedbuf;
		chan->stats.inbufcopies =
			chan->tdata.cut.inchan->stats.inbufcopies;
		chan->stats.inbufqueued =
			chan->tdata.cut.inchan->stats.inbufqueued;
		chan->stats.maxthreads =
			chan->tdata.cut.inchan->stats.maxthreads;
		chan->stats.outpackets =
			chan->tdata.cut.outchan->stats.outpackets;
		chan->stats.outbytes =
			chan->tdata.cut.outchan->stats.outbytes;
		chan->stats.outbufcopies =
			chan->tdata.cut.outchan->stats.outbufcopies;
	}
#ifdef CHAN_CLICK
	/* XXX hack for "composed" click output channels */
	else if (ani_chan_type(chan) == CHAN_OUT && chan->stats.inpackets != 0) {
		chan->stats.inpackets = 0;
		chan->stats.inbytes = 0;
		chan->stats.inbufcopies = 0;
		chan->stats.inbufqueued = 0;
		chan->stats.maxthreads = 0;
	}
#endif
	return &chan->stats;
}

void
ani_chan_thread_startintr(ani_chan_t *chan, ani_pbuf_t *pbuf, int msize)
{
	int inkern;
	void *pktarg;

	assert_ani_chan_valid(chan);
	assert(chan->tdata.in.func != 0);

	/*
	 * If channel is at its thread limit, queue the packet for later
	 */
	if (chan->curthreads >= ani_chan_maxthreads(chan)) {
		pbuf->work.farg[2] = (void *)msize; /* XXX */
		ani_pbuflist_addtail(&chan->tdata.in.qbufs, pbuf);
		CHANSTATS(chan->stats.inbufqueued++);
		return;
	}

	/*
	 * XXX special case for in side of a cut-through channel.
	 *
	 * The handler function for a cut-channel is run in the NodeOS
	 * and is passed an internal pbuffer object.
	 *
	 * The handler function for an in-channel is run in the EE and
	 * is passed the external handle for a packet.
	 */
	if (chan->flags & (CHAN_WITHCUT|CHAN_WITHCLICK)) {
		inkern = 1;
		pktarg = pbuf;
	} else {
		inkern = 0;
		pktarg = NODEOS_EXTERN_PBUF(pbuf);
	}

	ani_threadwork_set(&pbuf->work, inkern, chan,
			   (ani_thread_func_t)chan->tdata.in.func,
			   pktarg, chan->tdata.in.farg, (void *)msize);
	ani_threadpool_thread_startintr(chan->tpool, &pbuf->work);
	CHANSTATS(chan->stats.inpackets++);
	CHANSTATS(chan->stats.inbytes += ani_pbuf_size(pbuf));
	ani_chan_reference(chan);
	chan->curthreads++;
#ifdef ANI_CHANNEL_STATS
	if (chan->curthreads > chan->stats.maxthreads)
		chan->stats.maxthreads = chan->curthreads;
#endif
}

void
ani_chan_thread_start(ani_chan_t *chan, ani_pbuf_t *pbuf, int msize)
{
	ani_error_t err;
	int inkern;
	void *pktarg;

	assert_ani_chan_valid(chan);
	assert(ani_chan_islocked(chan));
	assert(chan->tdata.in.func != 0);

	/*
	 * If channel is at its thread limit, queue the packet for later
	 */
	if (chan->curthreads >= ani_chan_maxthreads(chan)) {
		ani_os_intr_disable();
		pbuf->work.farg[2] = (void *)msize; /* XXX */
		ani_pbuflist_addtail(&chan->tdata.in.qbufs, pbuf);
		CHANSTATS(chan->stats.inbufqueued++);
		ani_os_intr_enable();
		return;
	}

	/*
	 * XXX special case for in side of a cut-through channel.
	 *
	 * The handler function for a cut-channel is run in the NodeOS
	 * and is passed an internal pbuffer object.
	 *
	 * The handler function for an in-channel is run in the EE and
	 * is passed the external handle for a packet.
	 */
	if (chan->flags & (CHAN_WITHCUT|CHAN_WITHCLICK)) {
		inkern = 1;
		pktarg = pbuf;
	} else {
		inkern = 0;
		pktarg = NODEOS_EXTERN_PBUF(pbuf);
	}

	ani_threadwork_set(&pbuf->work, inkern, chan,
			   (ani_thread_func_t)chan->tdata.in.func,
			   pktarg, chan->tdata.in.farg, (void *)msize);
	err = ani_threadpool_thread_schedule(chan->tpool, &pbuf->work, 0);
	assert(!ani_error_occured(err));
	CHANSTATS(chan->stats.inpackets++);
	CHANSTATS(chan->stats.inbytes += ani_pbuf_size(pbuf));
	ani_chan_reference(chan);
	chan->curthreads++;
#ifdef ANI_CHANNEL_STATS
	if (chan->curthreads > chan->stats.maxthreads)
		chan->stats.maxthreads = chan->curthreads;
#endif
}

/*
 * Called from thread_loop when a thread is done processing a channel
 * packet.  If there were pending packets for this channel, put one of
 * them on the worklist for this thread to pick up on return.
 */
void
ani_chan_thread_exit(ani_chan_t *chan)
{
	ani_pbuf_t *pbuf;

	assert_ani_chan_valid(chan);

	ani_chan_lock(chan);

	ani_os_intr_disable();
	pbuf = ani_pbuflist_remhead(&chan->tdata.in.qbufs);
	ani_os_intr_enable();
	if (pbuf != 0) {
		ani_error_t err;
		int inkern;
		void *pktarg;
		int msize;

		/*
		 * XXX special case for in side of a cut-through channel.
		 *
		 * The handler function for a cut-channel is run in the NodeOS
		 * and is passed an internal pbuffer object.
		 *
		 * The handler function for an in-channel is run in the EE and
		 * is passed the external handle for a packet.
		 */
		if (chan->flags & (CHAN_WITHCUT|CHAN_WITHCUT)) {
			inkern = 1;
			pktarg = pbuf;
		} else {
			inkern = 0;
			pktarg = NODEOS_EXTERN_PBUF(pbuf);
		}

		msize = (int)pbuf->work.farg[2]; /* XXX */
		ani_threadwork_set(&pbuf->work, inkern, chan, 
				   (ani_thread_func_t)chan->tdata.in.func,
				   pktarg, chan->tdata.in.farg, (void *)msize);
		err = ani_threadpool_thread_schedule(chan->tpool,
						     &pbuf->work, 1);
		if (!ani_error_occured(err)) {
			CHANSTATS(chan->stats.inpackets++);
			CHANSTATS(chan->stats.inbytes += ani_pbuf_size(pbuf));
		}
		ani_chan_unlock(chan);
	} else {
		--chan->curthreads;
		ani_chan_unlock(chan);
		ani_chan_release(chan);
	}
}

/*
 * Channel packet buffer handling.
 */

/*
 * Get a free buffer off this channel's buffer list.
 * Returns ENOMEM if there are no buffers available.
 *
 * The channel should either be locked, interrupts should be disabled,
 * or both (it depends on the execution environment of the specific channel
 * type's "interrupt" function).
 */
ani_error_t
ani_chan_pbuf_alloc(ani_chan_t *chan, size_t size, ani_pbuf_t **out_pbuf)
{
	ani_pbuf_t *pbuf;
	ani_error_t err;

	assert_ani_chan_valid(chan);
	assert(ani_chan_type(chan) == CHAN_IN);
	assert(size >= 0);
	assert(out_pbuf != 0);

	/*
	 * Grab a non-busy buffer of the appropriate size.
	 */
	pbuf = ani_pbuflist_remhead(&chan->tdata.in.rbufs);
	if (pbuf == 0) {
		CHANSTATS(chan->stats.droppackets++);
		CHANSTATS(chan->stats.dropbytes += size);
		CHANSTATS(chan->stats.dropnobufs++);
		return ani_error_set(ENOMEM,
				     "no pbuf available on the channel");
	}
	assert(!ani_pbuf_isbusy(pbuf));

	ani_pbuf_setmax(pbuf);
	assert(ani_pbuf_bufsize(pbuf) >= size);

	err = ani_pbuf_setsize(pbuf, size, 0);
	assert(!ani_error_occured(err));

	*out_pbuf = pbuf;
	return ani_error_clear();
}

/*
 * Place a buffer back on the channel's buffer list.
 * Channel must be locked by caller.
 */
void
ani_chan_pbuf_free(ani_chan_t *chan, ani_pbuf_t *pbuf)
{
	assert_ani_chan_valid(chan);
	assert(ani_chan_type(chan) == CHAN_IN);
	assert(pbuf != 0);
	assert(ani_pbuf_chan(pbuf) == chan);

	/*
	 * If the channel is in the process of dying just free the buffer.
	 * If the pbuf is busy put it at the end of the busy list,
	 * else put it at the end of the free list.
	 */
	if (chan->flags & CHAN_DYING)
		ani_pbuf_deinit(pbuf);
	else {
		ani_os_intr_disable();
		if (ani_pbuf_isbusy(pbuf)) {
			ani_pbuflist_addtail(&chan->tdata.in.bbufs, pbuf);
		} else {
			ani_pbuflist_addtail(&chan->tdata.in.rbufs, pbuf);
		}
		ani_os_intr_enable();
	}
}

/*
 * Associate a list of pbufs with a channel.
 * Channel is unlocked on call and return.
 *
 * This is so that this routine can be used to add buffering to an existing
 * channel.  In that case, we don't want to keep an active channel locked
 * while we allocate memory.
 */
ani_error_t
ani_chan_pbuf_add(ani_chan_t *chan, ani_pbuf_t *pbufs[], int npbufs)
{
	ani_pbuflist_t fpb;
	ani_error_t err;
	int i;

	assert_ani_chan_valid(chan);
	assert(ani_chan_type(chan) == CHAN_IN);
	assert(pbufs != 0 || npbufs == 0);

	err = ani_error_clear();

	/*
	 * Common case of a single buffer.
	 * We can handle a busy buffer here.
	 */
	if (npbufs == 1) {
		assert(!ani_pbuf_onlist(pbufs[0]));
		ani_chan_lock(chan);
		if (chan->flags & CHAN_DYING)
			err = ani_error_set(EBADF, /* XXX */
					    "channel is dying");
		else {
			ani_pbuf_setchan(pbufs[0], chan);
			ani_chan_pbuf_free(chan, pbufs[0]);
		}
		ani_chan_unlock(chan);
		return err;
	}

	/*
	 * Chain the new bufs together.
	 *
	 * XXX for now, they can't be busy.  Accepting busy buffers
	 * requires that we atomically (w.r.t. interrupts) split up the
	 * busy and non-busy buffers and put them on the respective
	 * bbufs and rbufs queues.  Don't want to go there right now.
	 */
	ani_pbuflist_init(&fpb);
	for (i = 0; i < npbufs; i++) {
		assert(pbufs[i] != 0);
		assert(!ani_pbuf_onlist(pbufs[i]));
		assert(ani_pbuf_isbusy(pbufs[i]) == 0);
		ani_pbuflist_addtail(&fpb, pbufs[i]);
		ani_pbuf_setchan(pbufs[i], chan);
	}

	/*
	 * Now add them to the channel's list.
	 * This is done is a separate step to avoid holding the channel
	 * lock (and blocking interrupts) for too long.
	 */
	ani_chan_lock(chan);
	if (chan->flags & CHAN_DYING) {
		ani_pbuf_t *pb;

		ani_chan_unlock(chan);
		while ((pb = ani_pbuflist_remhead(&fpb)) != 0)
			ani_pbuf_setchan(pb, 0);
		err = ani_error_set(EBADF, /* XXX */
				    "channel is dying");
	} else {
		ani_os_intr_disable();
		ani_pbuflist_append(&chan->tdata.in.rbufs, &fpb);
		ani_os_intr_enable();
		ani_chan_unlock(chan);
	}
	return err;
}

/*
 * Remove and free all buffers associated with a channel.
 * Called during channel destroy.  Channel should be locked on call and return.
 *
 * @deallocation: all buffers on the channel list
 */
void
ani_chan_pbuf_removeall(ani_chan_t *chan)
{
	ani_pbuf_t *pb;

	assert_ani_chan_valid(chan);
	assert(ani_chan_type(chan) == CHAN_IN);
	assert(ani_chan_islocked(chan));
	assert((chan->flags & CHAN_DYING) != 0);

	/*
	 * Loop through, freeing all the buffers.
	 * Time spent holding the lock is not an issue here
	 * since the channel is dying anyway.  Also, since the
	 * channel has been disabled, there is no need to protect
	 * against interrupts when accessing the queues.
	 */
	ani_os_intr_disable();
	while ((pb = ani_pbuflist_remhead(&chan->tdata.in.qbufs)) != 0)
		ani_pbuf_deinit(pb);
	while ((pb = ani_pbuflist_remhead(&chan->tdata.in.rbufs)) != 0)
		ani_pbuf_deinit(pb);
	ani_os_intr_enable();
}


/*
 * Cut-through channel support
 */

/*
 * Create a cutchan given an existing in and out chan.
 */
ani_error_t
ani_cutchan_init(ani_chan_t *chan, ani_flow_t *flow, ani_threadpool_t *tpool,
		 void *cdata, an_chanproto_t *proto,
		 ani_chan_t *inchan, ani_chan_t *outchan,
		 an_chanrecvfunc_t rfunc, void *rfuncarg)
{
	ani_error_t err;
	void cut_ani_chan_ipforward(an_pbuf_t *anpbuf, void *arg, int msize);

	memset(chan, 0, sizeof *chan);

	/*
	 * XXX set this now so that functions we call know who to charge
	 * for memory.  We officially link into the flow after everything
	 * has been setup.
	 */
	chan->flow = flow;

	/*
	 * XXX should we hold a ref to the pool here?
	 * Don't think so.  Threadpool cannot go away as long as the flow
	 * is alive.  If the flow has been destroyed, so have all its channels.
	 */
	chan->tpool = tpool;

	if (proto != 0) {
		NODEOS_COPYIN(&chan->proto, proto, sizeof chan->proto, &err);
		if (ani_error_occured(err))
			return err;
	}

	chan->flags = CHAN_CUT;
	chan->tdata.cut.inchan = inchan;
	chan->tdata.cut.outchan = outchan;
	chan->refs = 1;
	chan->data = cdata;
	ani_ilock_init(&chan->lock);

	ani_chan_lock(chan);

#ifdef CHAN_ETH
	/* XXX hack IP forwarding cut channel */
	if (proto != 0 && proto->len == 13 &&
	    strncmp(proto->data, "eth/ip|ip/eth", proto->len) == 0)
		rfunc = cut_ani_chan_ipforward;
#endif

	/* XXX handle protocol processing */

	/*
	 * Setup the inchan's recv function.
	 *
	 * Note that the inchan holds a reference to the cutchan to
	 * keep it alive since the inchan receive function accesses the
	 * cutchan.
	 *
	 * XXX we replace the inchan's original threadpool with the one
	 * passed in.  The Click cutchan implementation uses this to ensure
	 * a single-threaded environment by creating a single-thread pool.
	 */
	assert(ani_chan_type(inchan) == CHAN_IN);
	inchan->flags |= CHAN_WITHCUT;
	inchan->tdata.in.func = rfunc;
	inchan->tdata.in.farg = rfuncarg;
	inchan->tpool = chan->tpool;
	ani_chan_reference(chan);

	/*
	 * Mark the outchan as part of the cutchan.
	 */
	assert(ani_chan_type(outchan) == CHAN_OUT);
	outchan->flags |= CHAN_WITHCUT;

	/*
	 * Finally link to the flow.
	 * We drop the channel lock first to adhere to our locking order.
	 * Note that we remove the component in/out channels from the flow's
	 * list so they are not accessible by the EE any longer.  This is
	 * more of a hack than anything, but it simplifies the semantics
	 * a little--don't have to worry about what happens if someone
	 * tweaks the in/out channel while the cutchan is active.
	 *
	 * XXX argh!  Have to leave the channels pointing to the flow so
	 * that foo_current() calls work; i.e., for resource accounting. 
	 */
	ani_chan_unlock(chan);
	ani_flow_removechannel(flow, inchan);
	inchan->flow = flow;
	ani_flow_removechannel(flow, outchan);
	outchan->flow = flow;
	ani_flow_addchannel(flow, chan);

	return ani_error_clear();
}

/*
 * Process an incoming packet on the inchan-side of a cutchan.
 */
void
ani_cutchan_receive(an_pbuf_t *anpbuf, void *arg, int msize)
{
	ani_chan_t *chan, *inchan, *outchan;
	ani_pbuf_t *pbuf;

	chan = arg;
	assert_ani_chan_valid(chan);
	pbuf = (ani_pbuf_t *)anpbuf;
	assert(pbuf != 0);
	assert(!ani_pbuf_isbusy(pbuf));

	ani_chan_lock(chan);
	if (chan->flags & CHAN_DYING) {
		ani_pbuf_deinit(pbuf);
		ani_chan_unlock(chan);
		return;
	}

	assert(ani_chan_type(chan) == CHAN_CUT);
	inchan = chan->tdata.cut.inchan;
	assert_ani_chan_valid(inchan);
	outchan = chan->tdata.cut.outchan;
	assert_ani_chan_valid(outchan);

	/* XXX do cutchan protocol processing */

#if 0
	{
		unsigned int *ip = ani_pbuf_data(pbuf);

		printf("%p: %d: %#x/%#x/%#x/%#x\n", chan, ani_pbuf_size(pbuf),
		       ntohl(ip[0]), ntohl(ip[1]), ntohl(ip[2]), ntohl(ip[3]));
	}
#endif

#ifdef CHAN_ETH
	/*
	 * Rewrite the ethernet header.
	 * XXX done with outchan unlocked, but should not be a problem.
	 */
	assert(outchan->ops == &chan_eth_ops);
	assert(ani_pbuf_size(pbuf) >= 14);
	memcpy(ani_pbuf_data(pbuf), &outchan->addr.eth, 14);
#endif

	/* Pump it back out */
	(void) ani_chan_send(outchan, pbuf);

	ani_chan_lock(inchan);
	ani_chan_pbuf_free(inchan, pbuf);
	ani_chan_unlock(inchan);

	ani_chan_unlock(chan);
}

#ifdef CHAN_ETH
#include <netinet/in.h>	/* XXX htons */

/*
 * Forward no-option IP packets.
 * Verifies the checksum, decrements the TTL and recomputes the checksum.
 */
void
cut_ani_chan_ipforward(an_pbuf_t *anpbuf, void *arg, int msize)
{
	ani_chan_t *chan, *inchan, *outchan;
	ani_pbuf_t *pbuf;
	void *pdata;
	int psize;
	struct iph {
		unsigned char vhl, tos;
		unsigned short len, id, off;
		unsigned char ttl, proto;
		unsigned short sum;
		unsigned long saddr, daddr;
	} *iph;

	chan = arg;
	assert_ani_chan_valid(chan);
	pbuf = (ani_pbuf_t *)anpbuf;
	assert(pbuf != 0);
	assert(!ani_pbuf_isbusy(pbuf));

	ani_chan_lock(chan);
	if (chan->flags & CHAN_DYING) {
		ani_pbuf_deinit(pbuf);
		ani_chan_unlock(chan);
		return;
	}

	assert(ani_chan_type(chan) == CHAN_CUT);
	inchan = chan->tdata.cut.inchan;
	assert_ani_chan_valid(inchan);
	outchan = chan->tdata.cut.outchan;
	assert_ani_chan_valid(outchan);

	pdata = ani_pbuf_data(pbuf);
	assert(pdata != 0);
	psize = ani_pbuf_size(pbuf);
	assert(psize > 14);

	if (psize < 14 + sizeof *iph) {
	bad:
		CHANSTATS(chan->stats.droppackets++);
		CHANSTATS(chan->stats.dropbytes += psize);
		CHANSTATS(chan->stats.dropbadpacket++);
		goto done;
	}
	iph = pdata + 14;

#if 0
	printf("%p: %d: vhl=%2x, len=%d, ttl=%d, s=%#lx, d=%#lx\n",
	       chan, psize - 14,
	       iph->vhl, ntohs(iph->len), iph->ttl,
	       ntohl(iph->saddr), ntohl(iph->daddr));
#endif

	/* Verify the checksum */
	{
		unsigned short *usp = (unsigned short *)iph;
		unsigned long cs;

		cs = *usp++;  cs += *usp++;
		cs += *usp++; cs += *usp++;
		cs += *usp++; cs += *usp++;
		cs += *usp++; cs += *usp++;
		cs += *usp++; cs += *usp;
		cs = (cs >> 16) + (cs & 0xffff);
		cs += (cs >> 16);
		if (cs != 0xffff)
			goto bad;
	}

	/* Decrement the TTL */
	if (--iph->ttl == 0)
		goto bad;

	/* Recompute the checksum (from Scout) */
	{
		unsigned long cs = iph->sum;

		cs += htons(0x0100);
		cs += (cs >> 16);
		iph->sum = cs;
	}

	/*
	 * Rewrite the ethernet header.
	 * XXX done with outchan unlocked, but should not be a problem.
	 */
	memcpy(pdata, &outchan->addr.eth, 14);

	/* Pump it back out */
	(void) ani_chan_send(outchan, pbuf);

 done:
	ani_chan_lock(inchan);
	ani_chan_pbuf_free(inchan, pbuf);
	ani_chan_unlock(inchan);

	ani_chan_unlock(chan);
}
#endif
