/*
 * Copyright (c) 1999, 2000 The University of Utah and the Flux Group.
 * All rights reserved.
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */

/*
 * Credential manipulation.
 *
 * XXX total hack
 */
#include <stdlib.h>
#include <string.h>
#include <assert.h>

#include <nodeos/impl/an_error.h>
#include <nodeos/impl/ani_resource.h>
#include <nodeos/impl/ani_cred.h>
#include <nodeos/impl/ani_flow.h>

ani_cred_t *ani_root_credential;

void
ani_cred_setup(ani_cred_t *root_cred, ani_resource_t *res)
{
	ani_error_t err;

	err = ani_cred_init(root_cred, 0, res);
	assert(!ani_error_occured(err));
	assert_ani_cred_valid(root_cred);

	ani_root_credential = root_cred;
}

ani_error_t
ani_cred_init(ani_cred_t *cred, void *auth, ani_resource_t *res)
{
	if (cred == 0 || auth != 0)
		return ani_error_set(EINVAL,
				     "invalid argument to ani_cred_init");

	memset(cred, 0, sizeof *cred);
	if (res != 0) {
		cred->limits = *res;
		assert_ani_cred_valid(cred);
	}

	return ani_error_clear();
}

void
ani_cred_deinit(ani_cred_t *cred)
{
	assert_ani_cred_valid(cred);
}

ani_cred_t *
ani_cred_current(void)
{
	ani_cred_t *cred;
	cred = ani_flow_current()->cred;
	assert_ani_cred_valid(cred);
	return cred;
}

int
ani_cred_check(ani_cred_t *cred, int op, ...)
{
	assert_ani_cred_valid(cred);
	return CCHECK_OK;
}


/* amount of resource available in the given cred */
#define credres_avail(c,f) \
	((c)->limits.f - (c)->current.f)

/* return 0 if adding amount from res doesn't exceed the cred limit */
#define credres_check(c,r,f) \
	((c) != ani_root_credential && (r)->f + (c)->current.f > (c)->limits.f)

/* add the amount indicated by res to the cred current use */
#define credres_add(c,r,f) \
	if ((c) != ani_root_credential) { \
		(c)->current.f += (r)->f; \
		assert((c)->current.f <= (c)->limits.f); \
	}

/* subtract the amount indicated by res to the cred current use */
#define credres_sub(c,r,f) \
	if ((c) != ani_root_credential) { \
		(c)->current.f -= (r)->f; \
		assert((c)->current.f >= 0); \
	}

ani_error_t
ani_cred_resource_get(ani_cred_t *cred, 
		      unsigned int *nthreads, unsigned int *cpurate,
		      unsigned int *memtotal, unsigned int *netrate)
{
	assert_ani_cred_valid(cred);

	*nthreads = cred->limits.threads;
	*cpurate = cred->limits.cpurate;
	*memtotal = cred->limits.memtotal;
	*netrate = cred->limits.netrate;

	return ani_error_clear();
}

ani_error_t
ani_cred_resource_xfer(ani_cred_t *srccred, ani_cred_t *dstcred,
		       unsigned int nthreads, unsigned int cpurate,
		       unsigned int memtotal, unsigned int netrate)
{
	assert_ani_cred_valid(srccred);
	assert_ani_cred_valid(dstcred);

	if (credres_avail(srccred, threads) < nthreads ||
	    credres_avail(srccred, cpurate) < cpurate ||
	    credres_avail(srccred, memtotal) < memtotal ||
	    credres_avail(srccred, netrate) < nthreads)
		return ani_error_set(EINVAL,
				"cpurate/netrate parameter > cred available");

	/* XXX all we care about right now */
	dstcred->limits.threads += nthreads;
	srccred->limits.threads -= nthreads;
	dstcred->limits.cpurate += cpurate;
	srccred->limits.cpurate -= cpurate;
	dstcred->limits.memtotal += memtotal;
	srccred->limits.memtotal -= memtotal;
	dstcred->limits.netrate += netrate;
	srccred->limits.netrate -= netrate;

	/* XXX */
	if (dstcred->limits.timeslice == 0)
		dstcred->limits.timeslice = srccred->limits.timeslice;
	if (dstcred->limits.stacksize == 0)
		dstcred->limits.stacksize = srccred->limits.stacksize;

	return ani_error_clear();
}

ani_error_t
ani_cred_resource_alloc(ani_cred_t *cred, ani_resource_t *res)
{
	ani_resource_t nres;

	assert_ani_cred_valid(cred);

	nres = *res;
	if (res->threads == RESOURCE_ALL)
		nres.threads = credres_avail(cred, threads);
	else if (credres_check(cred, res, threads))
		return ani_error_set(EPERM,
				     "Failed thread limit permission check");
	if (res->cpurate == RESOURCE_ALL)
		nres.cpurate = credres_avail(cred, cpurate);
	else if (credres_check(cred, res, cpurate))
		return ani_error_set(EPERM,
				     "Failed cpurate limit permission check");
	if (res->memtotal == RESOURCE_ALL)
		nres.memtotal = credres_avail(cred, memtotal);
	else if (credres_check(cred, res, memtotal))
		return ani_error_set(EPERM,
				     "Failed memory limit permission check");
	if (res->netrate == RESOURCE_ALL)
		nres.netrate = credres_avail(cred, netrate);
	else if (credres_check(cred, res, netrate))
		return ani_error_set(EPERM,
				     "Failed netrate limit permission check");
	*res = nres;

	credres_add(cred, res, threads);
	credres_add(cred, res, cpurate);
	credres_add(cred, res, memtotal);
	credres_add(cred, res, netrate);
	assert_ani_cred_valid(cred);

	return ani_error_clear();
}

void
ani_cred_resource_free(ani_cred_t *cred, ani_resource_t *res)
{
	assert_ani_cred_valid(cred);
	credres_sub(cred, res, threads);
	credres_sub(cred, res, cpurate);
	credres_sub(cred, res, memtotal);
	credres_sub(cred, res, netrate);
	assert_ani_cred_valid(cred);
}

