/*
 * Copyright (c) 1999, 2000, 2001 The University of Utah and the Flux Group.
 * All rights reserved.
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */
#include <assert.h>
#include <stdio.h> /* xxx snprintf() */

#include <nodeos/an_demuxkey.h>
#include <nodeos/impl/an_error.h>
#include <nodeos/impl/ani_chan.h>
#include <nodeos/impl/ani_mempool.h>
#include "ani_debug.h"

/* for various inet header matching funcs */
#include <netinet/in.h>
#include <netinet/in_systm.h>
#include <netinet/ip.h>

#define __FAVOR_BSD
#include <netinet/udp.h>
#include <netinet/tcp.h>
#undef __FAVOR_BSD

static ani_demuxkeyseg_t *segment_create(int off, int len,
					 void *key, void *mask);
static int segment_match(void *seg, void *pdata, int len);
static int segment_tostring(ani_demuxkeyseg_t *seg, char *buf, int buflen);

void
ani_demuxkeylist_init(ani_demuxkeylist_t *list)
{
	list->head = 0;
}

/*
 * Add a demuxkey to a key list.
 * The list is assumed to be locked by the caller.
 */
ani_error_t
ani_demuxkeylist_add(ani_demuxkeylist_t *list, ani_demuxkey_t *dkey)
{
	ani_demuxkey_t *dk, **dkp;

	for (dkp = &list->head; (dk = *dkp) != 0; dkp = &dk->next)
		if (dkey->prio > dk->prio)
			break;

	dkey->next = dk;
	*dkp = dkey;

	return ani_error_clear();
}

/*
 * Remove a demuxkey from a list.
 * The list is assumed to be locked by the caller.
 */
void
ani_demuxkeylist_remove(ani_demuxkeylist_t *list, ani_demuxkey_t *dkey)
{
	ani_demuxkey_t **dkp;

	for (dkp = &list->head; *dkp != 0; dkp = &(*dkp)->next)
		if (*dkp == dkey) {
			*dkp = dkey->next;
			break;
		}
}

/*
 * Apply all demuxkeys in the indicated list to the given packet data
 * and select the best match.  Returns the channel for the demuxkey selected
 * or zero if none matched.
 *
 * On call, *lenp contains the length of the data.
 * On successful return, *lenp indicates the length of the matched data.
 */
ani_chan_t *
ani_demuxkeylist_apply(ani_demuxkeylist_t *list, void *data, int *lenp)
{
	ani_demuxkey_t *dkey;
	int mlen;

	for (dkey = list->head; dkey != 0; dkey = dkey->next) {
		mlen = ani_demuxkey_match(dkey, data, *lenp);
		if (mlen >= 0) {
			assert(dkey->chan != 0);
			*lenp = mlen;
			return dkey->chan;
		}
	}

	return 0;
}

ani_error_t
ani_demuxkey_init(ani_demuxkey_t *dkey)
{
	assert(dkey != 0);

	memset(dkey, 0, sizeof *dkey);

	return ani_error_clear();
}

ani_error_t
ani_demuxkey_deinit(ani_demuxkey_t *dkey)
{
	ani_demuxkeyseg_t *seg;

	assert(dkey != 0);

	while (dkey->segs != 0) {
		seg = dkey->segs;
		dkey->segs = seg->next;
		ani_demuxkeyseg_destroy(seg);
	}

	return ani_error_clear();
}

/*
 * Create a copy of a demuxkey and all its segments.
 * Does not copy any chain or channel association.
 */
ani_error_t
ani_demuxkey_clone(ani_demuxkey_t *newdkey, ani_demuxkey_t *dkey)
{
	int seglen;
	ani_demuxkeyseg_t *seg, *nseg;

	assert(newdkey != 0);
	assert(dkey != 0);

	memset(newdkey, 0, sizeof *newdkey);
	newdkey->minpsize = dkey->minpsize;
	newdkey->prio = dkey->prio;
	for (seg = dkey->segs; seg != 0; seg = seg->next) {
		seglen = sizeof *seg + seg->dlen;
		nseg = ani_mempool_alloc(ani_mempool_current(), seglen);
		if (nseg == 0) {
			ani_demuxkey_deinit(newdkey);
			return ani_error_set(ENOMEM,
					     "no memory for demuxkey segment");
		}
		memcpy(nseg, seg, seglen);
	}

	return ani_error_clear();
}

ani_error_t
ani_demuxkey_addchannel(ani_demuxkey_t *dkey, ani_chan_t *chan)
{
	if (dkey->chan != 0)
		return ani_error_set(EBUSY,
				     "demuxkey already associated "
				     "with a channel");

	dkey->chan = chan;
	return ani_error_clear();
}

void
ani_demuxkey_removechannel(ani_demuxkey_t *dkey, ani_chan_t *chan)
{
	assert(dkey->chan == chan);
	dkey->chan = 0;
}

/*
 * Add a demuxkey segment to the front of the indicated key's segment list.
 * Returns an error if the segment is already attached to a key, zero ow.
 */
ani_error_t
ani_demuxkey_prepend(ani_demuxkey_t *dkey, ani_demuxkeyseg_t *seg)
{
	assert(dkey != 0);
	assert(seg != 0);

	/* XXX ensure segment not already attached to a list */

	/*
	 * Stick it on the front of the list
	 */
	seg->next = dkey->segs;
	dkey->segs = seg;

	/*
	 * Update minimum match packet size and demuxkey priority.
	 * XXX for priority we just use bytes matched by all segments.
	 */
	dkey->minpsize += (seg->offset + seg->mlen);
	dkey->prio += seg->mlen;

	return ani_error_clear();
}

/*
 * Add a demuxkey segment to the end of the indicated key's segment list.
 * Returns an error if the segment is already attached to a key, zero ow.
 */
ani_error_t
ani_demuxkey_append(ani_demuxkey_t *dkey, ani_demuxkeyseg_t *seg)
{
	ani_demuxkeyseg_t **prev;

	assert(dkey != 0);
	assert(seg != 0);

	/* XXX ensure segment not already attached to a list */

	/*
	 * Find the end of the list and stick it on.
	 */
	for (prev = &dkey->segs; *prev != 0; prev = &(*prev)->next)
		;
	seg->next = 0;
	*prev = seg;

	/*
	 * Update minimum match packet size and demuxkey priority.
	 * XXX for priority we just use bytes matched by all segments.
	 */
	dkey->minpsize += (seg->offset + seg->mlen);
	dkey->prio += seg->mlen;

	return ani_error_clear();
}

ani_error_t
ani_demuxkey_len(ani_demuxkey_t *dkey, int *len)
{
	*len = dkey->minpsize;

	return ani_error_clear();
}

/* 
 * Compare a given key to a given packet.  Returns -1 if they don't
 * match.  Otherwise, returns the offset in the pdata of the first byte
 * in the packet *after* the key.
 *
 * 'dkey' is the demux key to compare against 'pdata' for 'plen' bytes.
 */
int
ani_demuxkey_match(ani_demuxkey_t *dkey, void *pdata, int plen)
{
	ani_demuxkeyseg_t *seg;
	int poff;

	assert(dkey != 0);
	assert(pdata != 0);

	/*
	 * Null demuxkey matches the entire packet.
	 */
	if (dkey->segs == 0) {
		DBG(DBG_DEMUX_MATCH_DUMP,{
			ani_dprintf("DEMUX: <null>\nMATCHED\n");
		});
		return plen;
	}

	/*
	 * Packet not long enough for the minimum length of our filter,
	 * cannot match.
	 */
	if (plen < dkey->minpsize) {
		DBG(DBG_DEMUX_MATCH_DUMP,{
			ani_dprintf("DATA: <too short for key>\nFAILED\n");
		});
		return -1;
	}

	DBG(DBG_DEMUX_MATCH_DUMP,{
		    int j;
		    u_int8_t *cp = pdata;
		    
		    ani_dprintf("DATA: ");
		    for (j = 0; j < dkey->minpsize; j++)
			    ani_dprintf("%02x", cp[j]);
		    ani_dprintf("\n");
	});

	poff = 0;
	for (seg = dkey->segs; seg != 0; seg = seg->next) {
		int noff;

		poff += seg->offset;

		/*
		 * Packet not long enough to possibly match the next
		 * segment, fail.  Note that this test will only ever
		 * fail if we have segments which match variable lengths
		 * of data, a key consisting of completely static segments
		 * would have failed the composite minpsize above.
		 */
		if (poff + seg->mlen > plen) {
			DBG(DBG_DEMUX_MATCH_DUMP,{
				ani_dprintf("DATA: <too short for segment>\n"
					    "FAILED\n");
			});
			return -1;
		}

		/*
		 * Otherwise, call the segment match function with the
		 * appropriate part of the packet.
		 */
		noff = seg->func(seg->data, pdata+poff, plen-poff);
		if (noff == -1) {
			DBG(DBG_DEMUX_MATCH_DUMP,{
				ani_dprintf("FAILED\n");
			});
			return -1;
		}
		poff += noff;
	}

	DBG(DBG_DEMUX_MATCH_DUMP,{
		ani_dprintf("MATCHED\n");
	});

	return poff;
}

/*
 * Dump a representation of the given key into the given string buffer.
 */
ani_error_t
ani_demuxkey_tostring(ani_demuxkey_t *dkey, char *buf, int buflen)
{
	int rc;
	ani_demuxkeyseg_t *seg;

	assert(dkey);
	assert(buf);

	for (seg = dkey->segs, rc = 0; seg != 0; seg = seg->next, rc++)
		;

	rc = snprintf(buf, buflen, "nsegments=%d, maxoff=%d: ",
		      rc, dkey->minpsize);

	for (seg = dkey->segs; seg != 0; seg = seg->next) {
		if (rc >= buflen)
			return 0; /* XXX buffer too short */

		/*
		 * XXX gruesome hack alert...
		 * Needs serious work!
		 */
		if (seg->func == segment_match)
			rc += segment_tostring(seg, buf+rc, buflen-rc);
		else
			rc += snprintf(buf+rc, buflen-rc, "<fn=%p", seg->func);

		if (rc >= buflen)
			return 0; /* XXX buffer too short */

		rc += snprintf(buf+rc, buflen-rc, "@%d>", seg->offset);
	}

	return 0;
}

/*
 * Demuxkey segment constructors and destructor.
 *
 * XXX all do memory allocation.
 */

/*
 * Create a generic tuple segment.
 */
ani_demuxkeyseg_t *
ani_demuxkeyseg_tuple(int off, int len, void *key, void *mask)
{
	return segment_create(off, len, key, mask);
}

void
ani_demuxkeyseg_destroy(ani_demuxkeyseg_t *seg)
{
	/* XXX memory (de)allocation */
	ani_mempool_free(ani_mempool_current(), seg, sizeof *seg + seg->dlen);
}

/*
 * Create a demuxkey segment for matching ethernet headers.
 * Builds a tuple with key and possibly mask describing the header.
 */
ani_demuxkeyseg_t *
ani_demuxkeyseg_eth(int off, int flags, u_int8_t src[6], u_int8_t dst[6],
		    u_int16_t etype)
{
	an_ethaddrspec_t key, mask;
	int needmask;

	/*
	 * No fields wildcarded, initialize complete key and note
	 * that no mask is needed.
	 */
	if ((flags & AN_DEMUXKEY_ETH_ALLFLAGS) == 0) {
		needmask = 0;
		memcpy(&key.daddr, dst, sizeof key.daddr);
		memcpy(&key.saddr, src, sizeof key.saddr);
		key.type = etype;
	}

	/*
	 * Otherwise, some fields wildcarded and we need a mask.
	 * Initialize the key and mask values per field as necessary.
	 */
	else {
		needmask = 1;
		if ((flags & AN_DEMUXKEY_ETH_ANYDADDR) == 0) {
			memcpy(&key.daddr, dst, sizeof key.daddr);
			memset(&mask.daddr, ~0, sizeof mask.daddr);
		} else {
			memset(&key.daddr, 0, sizeof key.daddr);
			memset(&mask.daddr, 0, sizeof mask.daddr);
		}
		if ((flags & AN_DEMUXKEY_ETH_ANYSADDR) == 0) {
			memcpy(&key.saddr, src, sizeof key.saddr);
			memset(&mask.saddr, ~0, sizeof mask.saddr);
		} else {
			memset(&key.saddr, 0, sizeof key.saddr);
			memset(&mask.saddr, 0, sizeof mask.saddr);
		}
		if ((flags & AN_DEMUXKEY_ETH_ANYETYPE) == 0) {
			key.type = etype;
			mask.type = ~0;
		} else {
			key.type = 0;
			mask.type = 0;
		}
	}

	return segment_create(off, sizeof key, &key, needmask ? &mask : 0);
}

ani_demuxkeyseg_t *
ani_demuxkeyseg_udp(int off, int flags, u_int16_t sport, u_int16_t dport)
{
	struct udphdr key, mask;

	/*
	 * XXX always need a mask right now as there are fields in the UDP
	 * header we always wildcard.
	 */

	memset(&key, 0, sizeof key);
	memset(&mask, 0, sizeof mask);

	/*
	 * Initialize the key and mask values per field as necessary.
	 */
	if ((flags & AN_DEMUXKEY_UDP_ANYSPORT) == 0) {
		key.uh_sport = sport;
		mask.uh_sport = ~0;
	}
	if ((flags & AN_DEMUXKEY_UDP_ANYDPORT) == 0) {
		key.uh_dport = dport;
		mask.uh_dport = ~0;
	}

	return segment_create(off, sizeof key, &key, &mask);
}

/*
 * Support for IPv4 demuxkey
 */
struct ipv4seg {
	int flags;	 /* which fields to wildcard */
	u_int32_t saddr; /* desired source IP address */
	u_int32_t smask; /* mask for source IP address */
	u_int32_t daddr; /* desired destination IP address */
	u_int32_t dmask; /* mask for destination IP address */
	u_int8_t proto;	 /* desired IP protcol */
};

static int
ipv4_match(void *ddata, void *pdata, int plen)
{
	struct ipv4seg *ipseg = ddata;
	struct ip *ip = pdata;
	int iphlen;

	assert(ddata != 0);
	assert(ipseg != 0);
	assert(ip != 0);

	/* XXX should we checksum the header? */

	iphlen = ip->ip_hl << 2;

	DBG(DBG_DEMUX_MATCH_DUMP|DBG_IP_MATCH_DUMP, {
		int j;
		int len;
		u_int8_t *cp = pdata;

		ani_dprintf("APPLY IPv4 match:\n");
		printf(" K=0x400000000000000000%02x0000%08x%08x\n",
		       ipseg->proto, ipseg->saddr, ipseg->daddr);
		printf(" M=0xf00000000000000000%02x0000%08x%08x\n",
		  (ipseg->flags & AN_DEMUXKEY_IPV4_ANYPROTO)? 0 : (u_int8_t)~0,
		  (ipseg->flags & AN_DEMUXKEY_IPV4_ANYSADDR)? 0 : ipseg->smask,
		  (ipseg->flags & AN_DEMUXKEY_IPV4_ANYDADDR)? 0 : ipseg->dmask);
		ani_dprintf("TO:\n D=0x");
		len = plen;
		if (iphlen < len)
			len = iphlen;
		for (j = 0; j < len; j++)
			ani_dprintf("%02x", cp[j]);
		ani_dprintf("\n");
	});

	if (ip->ip_v != 4 || plen < iphlen ||
	    ((ipseg->flags & AN_DEMUXKEY_IPV4_ANYSADDR) == 0 &&
	     (ip->ip_src.s_addr & ipseg->smask) != ipseg->saddr) ||
	    ((ipseg->flags & AN_DEMUXKEY_IPV4_ANYDADDR) == 0 &&
	     (ip->ip_dst.s_addr & ipseg->dmask) != ipseg->daddr) ||
	    ((ipseg->flags & AN_DEMUXKEY_IPV4_ANYPROTO) == 0 &&
	     ip->ip_p != ipseg->proto)) {
		DBG(DBG_IP_MATCH_DUMP, {
			ani_dprintf("FAILED\n");
		})
		return -1;
	}

	DBG(DBG_IP_MATCH_DUMP, {
		ani_dprintf("MATCHED\n");
	})
	return iphlen;
}

/*
 * Create a demuxkey segment for matching IPv4 headers.
 */
ani_demuxkeyseg_t *
ani_demuxkeyseg_ipv4(int off, int flags, u_int32_t src, u_int32_t srcmask,
		     u_int32_t dst, u_int32_t dstmask, u_int8_t proto)
{
	ani_demuxkeyseg_t *seg;
	struct ipv4seg *ipseg;
	int alloclen;

	alloclen = sizeof *seg + sizeof *ipseg;

	/* XXX memory allocation */
	seg = ani_mempool_alloc(ani_mempool_current(), alloclen);
	if (seg == 0)
		return 0;

	memset(seg, 0, alloclen);
	seg->offset = off;
	seg->mlen = sizeof(struct ip);
	seg->func = ipv4_match;
	seg->dlen = sizeof *ipseg;

	ipseg = (struct ipv4seg *)seg->data;
	ipseg->flags = flags;
	ipseg->saddr = src;
	ipseg->smask = srcmask;
	ipseg->daddr = dst;
	ipseg->dmask = dstmask;
	ipseg->proto = proto;

	return seg;
}


/*
 * Support for TCP demuxkey
 */
struct tcpseg {
	int flags;	 /* which fields to wildcard */
	u_int16_t sport; /* desired source port */
	u_int16_t dport; /* desired destination port */
};

static int
tcp_match(void *ddata, void *pdata, int plen)
{
	struct tcpseg *tcpseg = ddata;
	struct tcphdr *tcp = pdata;
	int tcphlen;

	assert(ddata != 0);
	assert(tcpseg != 0);
	assert(tcp != 0);

	/* XXX should we checksum the packet? */

	tcphlen = tcp->th_off << 2;

	DBG(DBG_DEMUX_MATCH_DUMP, {
		int j;
		int len;
		u_int8_t *cp = pdata;

		ani_dprintf("APPLY TCP match TO");
		ani_dprintf("\n      D=0x");
		len = plen;
		if (tcphlen < len)
			len = tcphlen;
		for (j = 0; j < len; j++)
			ani_dprintf("%02x", cp[j]);
		ani_dprintf("\n");
	});

	if (plen < tcphlen ||
	    ((tcpseg->flags & AN_DEMUXKEY_TCP_ANYSPORT) == 0 &&
	     tcp->th_sport != tcpseg->sport) ||
	    ((tcpseg->flags & AN_DEMUXKEY_TCP_ANYDPORT) == 0 &&
	     tcp->th_dport != tcpseg->dport))
		return -1;

	return tcphlen;
}

/*
 * Create a demuxkey segment for matching TCP headers.
 */
ani_demuxkeyseg_t *
ani_demuxkeyseg_tcp(int off, int flags, u_int16_t sport, u_int16_t dport)
{
	ani_demuxkeyseg_t *seg;
	struct tcpseg *tcpseg;
	int alloclen;

	alloclen = sizeof *seg + sizeof *tcpseg;

	/* XXX memory allocation */
	seg = ani_mempool_alloc(ani_mempool_current(), alloclen);
	if (seg == 0)
		return 0;

	memset(seg, 0, alloclen);
	seg->offset = off;
	seg->mlen = sizeof(struct tcphdr);
	seg->func = tcp_match;
	seg->dlen = sizeof *tcpseg;

	tcpseg = (struct tcpseg *)seg->data;
	tcpseg->flags = flags;
	tcpseg->sport = sport;
	tcpseg->dport = dport;

	return seg;
}


/*
 * Support for ANEP demuxkey
 */
struct anephdr {
	u_int8_t ah_version;
	u_int8_t ah_flags;
	u_int16_t ah_typeid;
	u_int16_t ah_hlen;
	u_int16_t ah_plen;
};

struct anepseg {
	int flags;	  /* which fields to wildcard */
	u_int8_t version; /* desired version number */
	u_int16_t proto;  /* desired protocol ID */
};

static int
anep_match(void *ddata, void *pdata, int plen)
{
	struct anepseg *anepseg = ddata;
	struct anephdr *anep = pdata;
	int anephlen;

	assert(ddata != 0);
	assert(anepseg != 0);
	assert(anep != 0);

	anephlen = anep->ah_hlen;

	DBG(DBG_DEMUX_MATCH_DUMP, {
		int j;
		int len;
		u_int8_t *cp = pdata;

		ani_dprintf("APPLY ANEP match TO");
		ani_dprintf("\n      D=0x");
		len = plen;
		if (anephlen < len)
			len = anephlen;
		for (j = 0; j < len; j++)
			ani_dprintf("%02x", cp[j]);
		ani_dprintf("\n");
	});

	if (plen < anephlen ||
	    ((anepseg->flags & AN_DEMUXKEY_ANEP_ANYVERSION) == 0 &&
	     anep->ah_version != anepseg->version) ||
	    ((anepseg->flags & AN_DEMUXKEY_ANEP_ANYPROTOID) == 0 &&
	     anep->ah_typeid != anepseg->proto))
		return -1;

	return anephlen;
}

/*
 * Create a demuxkey segment for matching ANEP headers.
 */
ani_demuxkeyseg_t *
ani_demuxkeyseg_anep(int off, int flags, u_int8_t version, u_int16_t protoid)
{
	ani_demuxkeyseg_t *seg;
	struct anepseg *anepseg;
	int alloclen;

	alloclen = sizeof *seg + sizeof *anepseg;

	/* XXX memory allocation */
	seg = ani_mempool_alloc(ani_mempool_current(), alloclen);
	if (seg == 0)
		return 0;

	memset(seg, 0, alloclen);
	seg->offset = off;
	seg->mlen = sizeof(struct anephdr);
	seg->func = anep_match;
	seg->dlen = sizeof *anepseg;

	anepseg = (struct anepseg *)seg->data;
	anepseg->flags = flags;
	anepseg->version = version;
	anepseg->proto = protoid;

	return seg;
}


/*
 * Support for standard tuple segments, with and without mask.
 */
struct tupleseg {
	int len;
	u_int8_t *key;
	u_int8_t *mask;
	u_int8_t data[0];
};

#ifdef DEBUG
static void segment_dump(struct tupleseg *tseg);
#endif

static ani_demuxkeyseg_t *
segment_create(int off, int len, void *key, void *mask)
{
	ani_demuxkeyseg_t *seg;
	struct tupleseg *tseg;
	int alloclen;

	alloclen = sizeof *seg + sizeof *tseg;
	if (key) {
		alloclen += len;
		if (mask)
			alloclen += len;
	}

	/* XXX memory allocation */
	seg = ani_mempool_alloc(ani_mempool_current(), alloclen);
	if (seg == 0)
		return 0;

	memset(seg, 0, alloclen);
	seg->offset = off;
	seg->mlen = len;
	seg->func = segment_match;
	seg->dlen = alloclen - sizeof *seg;

	tseg = (struct tupleseg *)seg->data;	/* XXX dangerous cast */
	tseg->len = len;
	if (key != 0) {
		tseg->key = tseg->data;
		memcpy(tseg->key, key, len);
		if (mask != 0) {
			tseg->mask = tseg->data + len;
			memcpy(tseg->mask, mask, len);
		}
	}

	return seg;
}

static int
segment_match(void *ddata, void *pdata, int plen)
{
	struct tupleseg *tseg = ddata;
	u_int8_t *dp, *kp, *mp;
	int len;

	assert(tseg != 0);
	assert(pdata != 0);
	assert(plen >= tseg->len);

	/* no key, cannot match */
	if (tseg->key == 0)
		return -1;

	DBG(DBG_DEMUX_MATCH_DUMP, {
		int j;
		u_int8_t *cp = pdata;

		ani_dprintf("APPLY tuple:\n");
		if (tseg->key) {
			printf(" K=0x");
			for (j = 0; j < tseg->len; j++)
				ani_dprintf("%02x", tseg->key[j]);
			if (tseg->mask) {
				ani_dprintf("\n");
				printf(" M=0x");
				for (j = 0; j < tseg->len; j++)
					ani_dprintf("%02x", tseg->mask[j]);
			}
		}
		ani_dprintf("\nTO:\n D=0x");
		len = plen;
		if (len > tseg->len)
			len = tseg->len;
		for (j = 0; j < len; j++)
			ani_dprintf("%02x", cp[j]);
		ani_dprintf("\n");
	});

	/*
	 * No mask, must match exactly.
	 */
	if (tseg->mask == 0) {
		if (memcmp(pdata, tseg->key, tseg->len) == 0)
			return tseg->len;
		return -1;
	}

	/*
	 * Otherwise, perform mask and compare.
	 * XXX needs to be optimized!
	 */
	dp = pdata;
	kp = tseg->key;
	mp = tseg->mask;
	len = tseg->len;
	while (len-- != 0)
		if ((*dp++ & *mp++) != *kp++)
			return -1;

	return tseg->len;
}

static int
segment_tostring(ani_demuxkeyseg_t *seg, char *buf, int buflen)
{
	struct tupleseg *tseg;
	int rc, b;

	tseg = (struct tupleseg *)seg->data;	/* XXX dangerous cast */
	if (tseg->key == 0)
		return 0;

	rc = snprintf(buf, buflen, "<0x");
	for (b = 0; b < tseg->len; b++) {
		if (rc >= buflen)
			return 0;
		rc += snprintf(buf+rc, buflen-rc, "%02x", tseg->key[b]);
	}
	if (tseg->mask != 0) {
		rc += snprintf(buf+rc, buflen-rc, "/0x");
		for (b = 0; b < tseg->len; b++) {
			if (rc >= buflen)
				return 0;
			rc += snprintf(buf+rc, buflen-rc, "%02x",
				       tseg->mask[b]);
		}
	}
	rc += snprintf(buf+rc, buflen-rc, ">");

	return rc;
}

#ifdef DEBUG
static void
segment_dump(struct tupleseg *tseg)
{
	int j;

	if (tseg->key) {
		printf("key=0x");
		for (j = 0; j < tseg->len; j++)
			ani_dprintf("%02x", tseg->key[j]);
		if (tseg->mask) {
			printf(", mask=0x");
			for (j = 0; j < tseg->len; j++)
				ani_dprintf("%02x", tseg->mask[j]);
		}
	}
	ani_dprintf("\n");
}

void
ani_demuxkey_dump(ani_demuxkey_t *dkey)
{
	ani_demuxkeyseg_t *seg;

	ani_dprintf("KEY %p, minpsize=%d, prio=%d, chan=%p, next=%p:\n",
		    dkey, dkey->minpsize, dkey->prio, dkey->chan, dkey->next);
	for (seg = dkey->segs; seg != 0; seg = seg->next) {
		ani_dprintf(" SEG %p: off=%d, mlen=%d, ",
			    seg, seg->offset, seg->mlen);
		if (seg->func == segment_match)
			segment_dump((struct tupleseg *)seg->data);
		else
			ani_dprintf("func=%s\n",
				    seg->func == ipv4_match ? "ipv4_match" :
				    seg->func == tcp_match ? "tcp_match" :
				    seg->func == anep_match ? "anep_match" :
				    "unknown");
	}
}

void
ani_demuxkeylist_dump(ani_demuxkeylist_t *list)
{
	ani_demuxkey_t *dkey;

	ani_dprintf("KEYLIST %p:\n", list);
	for (dkey = list->head; dkey != 0; dkey = dkey->next)
		ani_demuxkey_dump(dkey);
}

#endif
