/*
 * Copyright (c) 1999, 2000 The University of Utah and the Flux Group.
 * All rights reserved.
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */

#include <stdlib.h>
#include <string.h>
#include <assert.h>

#include <nodeos/impl/ani_hnode.h>

void
hnode_init(struct hnode *node)
{
	memset(node, 0, sizeof *node);
}

void
hnode_attach(struct hnode *node, struct hnode *parent, struct hnode *list)
{
	if (parent) {
		assert(node->parent == 0);
		node->parent = parent;
		node->parentlink = parent->child;
		parent->child = node;
	}
	if (list) {
		assert(node->global == 0);
		node->global = list;
		node->globallink = list->globallink;
		list->globallink = node;
	}
}

void
hnode_detach(struct hnode *node, int flags)
{
	struct hnode **prevp;

	if ((flags & HNODE_PARENT) && node->parent) {
		for (prevp = &node->parent->child;
		     *prevp != 0; prevp = &(*prevp)->parentlink)
			if (*prevp == node)
				break;
		assert(*prevp);
		*prevp = node->parentlink;
		node->parent = 0;
	}
	if ((flags & HNODE_GLOBAL) && node->global) {
		for (prevp = &node->global->globallink;
		     *prevp != 0; prevp = &(*prevp)->globallink)
			if (*prevp == node)
				break;
		assert(*prevp);
		*prevp = node->globallink;
		node->global = 0;
	}
}

int
hnode_traverse(struct hnode *root, int (*func)(struct hnode *, void *),
	       void *arg, int flags, int reverse)
{
	struct hnode *cur, *next;
	int rv;

	assert((flags & (HNODE_GLOBAL|HNODE_PARENT)) != 0);

	if (flags & HNODE_GLOBAL) {
		for (cur = root->global->globallink; cur != 0; cur = next) {
			next = cur->globallink;
			rv = func(cur, arg);
			if (rv)
				return rv;
		}
		return 0;
	}

	/* flags & HNODE_PARENT */
	if (!reverse) {
		rv = func(root, arg);
		if (rv)
			return rv;
	}
	for (cur = root->child; cur != 0; cur = next) {
		next = cur->parentlink;
		rv = hnode_traverse(cur, func, arg, HNODE_PARENT, reverse);
		if (rv)
			return rv;
	}
	if (reverse) {
		rv = func(root, arg);
		if (rv)
			return rv;
	}

	return 0;
}

int
hnode_isattached(struct hnode *node, int flags)
{
	if ((flags & HNODE_PARENT) && node->parent)
		return 1;
	if ((flags & HNODE_GLOBAL) && node->global)
		return 1;
	return 0;
}

struct hnode *
hnode_parent(struct hnode *node)
{
	return node->parent;
}

struct hnode *
hnode_child(struct hnode *node)
{
	return node->child;
}

struct hnode *
hnode_sibling(struct hnode *node)
{
	return node->parentlink;
}

struct hnode *
hnode_listhead(struct hnode *node)
{
	return node->global;
}

struct hnode *
hnode_listnext(struct hnode *node)
{
	return node->globallink;
}

