/*
 * Copyright (c) 2000 The University of Utah and the Flux Group.
 * All rights reserved.
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */

#ifndef _NODEOS_SRC_OS_H_
#define _NODEOS_SRC_OS_H_

#include <nodeos/an_error.h>
#include <nodeos/impl/ani_resource.h>
#include <nodeos/impl/an_time.h>

/*
 * The host-os-specific interfaces.  There are five (somewhat)
 * separate implementations of these functions for: 
 *   Solaris, Linux, FreeBSD, OSKit and OSKit-unix.
 * All the relevant files are in the os/ subdirectory.
 */

/*
 * Basic initialization interfaces.
 */
void ani_os_init_basic(void);
void ani_os_init_mem(void);
void ani_os_init_devices(void);
void ani_os_init_net(void);
void ani_os_init_fs(void);

/*
 * Generic stuff
 */
void ani_os_time_current(an_timespec_t *curtime);
void ani_os_dump_state(void *state);

/*
 * Scheduling interfaces
 */
struct ani_thread;
struct ani_threadpool;
struct ani_threadcpu;
void ani_os_init_sched(void);
int ani_os_sched_setparam(struct ani_thread *thread,
			  struct ani_threadpool *tpool);
int ani_os_sched_setprio(struct ani_thread *thread, int nprio);
int ani_os_sched_getcpu(struct ani_thread *thread, struct ani_threadcpu *cpu);

/*
 * Gprof interfaces
 */
void ani_os_gprof_init(void);
void ani_os_gprof_start(const char *kernelname);
void ani_os_gprof_stop(void);

/*
 * Info interfaces
 */
const char *ani_os_gethostname(void);
int ani_os_sethostname(const char *name);
void ani_os_init_resources(ani_resource_t *res);

/*
 * Timer interfaces
 */
void *ani_os_timer_init(void (*func)(void *, void *), void *farg,
		    unsigned long *timerres);
void ani_os_timer_deinit(void *handle);
int ani_os_timer_start(void *handle, unsigned long timerres);

void panic(const char *fmt, ...);

/*
 * Processor interrupt handling
 */
#ifdef OSKIT
#include <oskit/dev/dev.h>

/*
 * (Linux) ethernet device packet received callbacks (netio_push) occur
 * at software interrupt level.
 *
 * Ethernet transmit done callbacks (the final bufio_release) occur
 * at hardware interrupt level.
 *
 * Timer callbacks (listener_notify) occur at hardware interrupt level.
 *
 * Right now we just block hardware interrupts for all of them.
 */
#define ani_os_intr_enable() \
	assert(osenv_intr_enabled() == 0), osenv_intr_enable()
#define ani_os_intr_disable() \
	assert(osenv_intr_enabled() != 0), osenv_intr_disable()
#define ani_os_intr_save_disable() \
	osenv_intr_save_disable()
#define ani_os_intr_enabled() \
	osenv_intr_enabled()
#else
/*
 * Currently there are no interrupts to worry about.
 * Timers and ethernet devices are polled by dedicated threads.
 */
#define ani_os_intr_enable()
#define ani_os_intr_disable()
#define ani_os_intr_save_disable()
#define ani_os_intr_enabled()
#endif
#endif /* _NODEOS_SRC_OS_H_ */
