/*
 * Copyright (c) 2000 The University of Utah and the Flux Group.
 * All rights reserved.
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */
#ifdef OSKIT_STRIDE_SCHED
#include <oskit/threads/sched_stride.h>	/* XXX force inclusion of stridesched */
#endif

#include <nodeos/an_thread.h>
#include <nodeos/impl/an_error.h>
#include <nodeos/impl/ani_thread.h>

#include "os.h"

/*
 * CPU scheduling support of OSKit and OSKit on Unix
 */

static unsigned int cpucycles, pt_scale, an_scale;

void
ani_os_init_sched(void)
{
	ani_resource_t res;

	ani_os_init_resources(&res);
	cpucycles = res.cpurate;
	assert(cpucycles != 0);

	/*
	 * For mapping to/from pthread/NodeOS priorities
	 */
	pt_scale = cpucycles / (PRIORITY_MAX - PRIORITY_MIN);
	assert(pt_scale != 0);
	an_scale = cpucycles / (AN_THREAD_PRIO_EXCLUSIVE - AN_THREAD_PRIO_IDLE);
	assert(an_scale != 0);
}

static inline unsigned int
anprio_to_cpurate(ani_thread_t *thread, int prio)
{
	int tp_scale =
		(thread->pool == 0) ? 1 :
		((thread->pool->params.rate > cpucycles) ? 1 :
		 cpucycles / thread->pool->params.rate);
	
	assert(cpucycles != 0);
	return ((prio - AN_THREAD_PRIO_IDLE) * an_scale) / tp_scale;
}

static inline int
cpurate_to_anprio(ani_thread_t *thread, unsigned int cpurate)
{
	int tp_scale =
		(thread->pool == 0) ? 1 :
		((thread->pool->params.rate > cpucycles) ? 1 :
		 cpucycles / thread->pool->params.rate);
	
	assert(cpucycles != 0);

	/*
	 * XXX possible due to really-high-priority root flow threads
	 */
	if (cpurate >= cpucycles)
		return AN_THREAD_PRIO_EXCLUSIVE;

	return (cpurate * tp_scale) / an_scale + AN_THREAD_PRIO_IDLE;
}

/*
 * Initialize the scheduling attributes for an embryonic thread.
 */
int
ani_os_sched_setparam(ani_thread_t *thread, ani_threadpool_t *tpool)
{
	int err;
	unsigned int cpurate;

	/*
	 * XXX threadpool cpurate should be shared by all active threads in
	 * the pool.  That would involve dynamic adjustment of existing
	 * thread's cpu rates.  For now we just divide up the cpurate
	 * among all possible threads in the pool.
	 */
	cpurate = tpool->res.cpurate / tpool->res.threads;
	assert(cpurate != 0);

#ifdef OSKIT_STRIDE_SCHED
	/*
	 * Scaled CPU rate is the number of tickets we assign.
	 */
	thread->pprio.tickets = cpurate;

	/*
	 * XXX set priority field as well since setschedparam contains
	 * a range check on it even though the value is ignored.
	 */
	thread->pprio.priority = PRIORITY_NORMAL;

	err = pthread_attr_setschedparam(&thread->attr, &thread->pprio);
	if (err == 0)
		err = pthread_attr_setschedpolicy(&thread->attr, SCHED_STRIDE);
#else
	/*
	 * Map CPU cycle rate onto pthread priorities
	 * XXX note hack for root flow real-high-priority threads
	 */
	if (cpurate >= cpucycles)
		thread->pprio.priority = PRIORITY_MAX;
	else
		thread->pprio.priority = PRIORITY_MIN + cpurate / pt_scale;

	err = pthread_attr_setschedparam(&thread->attr, &thread->pprio);
#endif

	/*
	 * XXX map cpurate onto a NodeOS priority value.
	 * In our implementation these priority values are not scheduler
	 * specific as we have only one scheduler in any instantiation
	 * of the NodeOS.
	 */
	if (err == 0)
		thread->nprio = cpurate_to_anprio(thread, cpurate);

	return err;
}

/*
 * Change the scheduling priority of an existing thread.
 * This routine should map standard NodeOS priority values into
 * implementation specific priorities.
 */
int
ani_os_sched_setprio(ani_thread_t *thread, int nprio)
{
	struct sched_param oparam;
	unsigned int cpurate;
	int err;

	if (nprio < AN_THREAD_PRIO_IDLE || nprio > AN_THREAD_PRIO_EXCLUSIVE)
		return EINVAL;

	oparam = thread->pprio;

	cpurate = anprio_to_cpurate(thread, nprio);
#ifdef OSKIT_STRIDE_SCHED
	thread->pprio.tickets = cpurate;
	thread->pprio.priority = PRIORITY_NORMAL; /* XXX for setschedparam */
#else
	assert(pt_scale != 0);
	thread->pprio.priority = cpurate / pt_scale;
#endif

	err = pthread_setschedparam(thread->id, -1, &thread->pprio);
	if (err != 0)
		thread->pprio = oparam;

	return err;
}

int
ani_os_sched_getcpu(ani_thread_t *thread, ani_threadcpu_t *cpu)
{
	oskit_u32_t val;
	long long cycles;
	
	if (oskit_pthread_cpucycles(thread->id, &cycles) == 0) {
		cpu->units = ANI_CPU_CYCLES;
		cpu->amount = cycles;
	} else {
		val = oskit_pthread_cputime(thread->id);
		assert((int)val >= 0);
		cpu->units = ANI_CPU_USECS;
		cpu->amount = (long long)val * 1000;
	}

	return 0;
}
