/*
 * Copyright (c) 1999, 2000 The University of Utah and the Flux Group.
 * All rights reserved.
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */

#include <assert.h>
#include <nodeos/impl/an_error.h>

#include <oskit/com/listener.h>
#include <oskit/dev/dev.h>
#include <oskit/dev/clock.h>
#include <oskit/dev/timer.h>
#include <oskit/threads/pthread.h>

#include <nodeos/impl/ani_flow.h>

#include "os.h"

#define TIMER_FREQ	100	/* XXX OSKit max resolution is 10ms */

oskit_listener_t *oskit_create_threaded_listener(
				oskit_listener_callback_t handler, void *arg);
void oskit_run_threaded_listener(oskit_listener_t *io);

struct ani_os_timer {
	oskit_timer_t *timer;
	ani_thread_t thread;
};

/*
 * Initialize a periodic timer for NodeOS use.
 *
 * Returns a handle to the timer created.  The maximum resolution
 * of the timer in microseconds is returned in the timerres param.
 */
void *
ani_os_timer_init(void (*func)(void *, void *), void *farg,
		  unsigned long *timerres)
{
	ani_error_t err;
	oskit_listener_t *l;
	struct ani_os_timer *timer;
	extern oskit_clock_t *oskit_system_clock;

	timer = osenv_mem_alloc(sizeof *timer, OSENV_PHYS_WIRED, 0);
	if (timer == 0)
		return 0;

	/*
	 * Create a dedicated listener thread for servicing timer interrupts
	 * in a full thread context.
	 */
	l = oskit_create_threaded_listener((oskit_listener_callback_t)func,
					   farg);
	if (l == 0)
		return 0;

	err = ani_rootflow_startthread(&timer->thread,
			(ani_thread_func_t)oskit_run_threaded_listener, l, 0, 0);
	if (ani_error_occured(err)) {
		oskit_listener_release(l);
		osenv_mem_free(timer, OSENV_PHYS_WIRED, sizeof *timer);
		return 0;
	}

	/*
	 * Create the timer and associate it with the listener
	 */
	if (oskit_clock_createtimer(oskit_system_clock, &timer->timer) != 0) {
		oskit_listener_release(l);
		osenv_mem_free(timer, OSENV_PHYS_WIRED, sizeof *timer);
		return 0;
	}

	oskit_timer_setlistener(timer->timer, l);
	oskit_listener_release(l);

	*timerres = 1000000 / TIMER_FREQ;
	return timer;
}

/*
 * Destroy a periodic timer.
 */
void
ani_os_timer_deinit(void *handle)
{
	struct ani_os_timer *timer = handle;

	ani_thread_kill(&timer->thread, 1);
	oskit_timer_release(timer->timer);
	osenv_mem_free(timer, OSENV_PHYS_WIRED, sizeof *timer);
}

/*
 * Start a periodic timer with the specified frequency.
 */
int
ani_os_timer_start(void *handle, unsigned long timerres)
{
	struct ani_os_timer *timer = handle;
	oskit_itimerspec_t ts;
	int err;

	assert(timerres < 1000000);

	ts.it_value.tv_sec = 0;
	ts.it_value.tv_nsec = timerres * 1000;
	ts.it_interval.tv_sec = 0;
	ts.it_interval.tv_nsec = timerres * 1000;
	err = oskit_timer_settime(timer->timer, 0, &ts);

	return err;
}
