/*
 * Copyright (c) 1999, 2000, 2001 The University of Utah and the Flux Group.
 * All rights reserved.
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */

#include <nodeos/an_thread.h>
#include <nodeos/impl/ani_thread.h>
#include <nodeos/impl/ani_threadpool.h>

#include "os.h"

/*
 * Use this is each pthread is a separate process and if the pthreads
 * priority framework doesn't work.
 */
#undef USE_NICE

#ifdef USE_NICE
#include <sys/time.h>
#include <sys/resource.h>
#endif

/*
 * CPU scheduling support of pthreads on Unix
 */

static unsigned int cpucycles, pt_scale, an_scale;
static int priority_min, priority_max, nopriorities;

void
ani_os_init_sched(void)
{
	ani_resource_t res;
	int policy;

	ani_os_init_resources(&res);
	cpucycles = res.cpurate;
	assert(cpucycles != 0);

	/*
	 * Get pthread priority range
	 */
	policy = sched_getscheduler(0);
	if (policy >= 0) {
		priority_min = sched_get_priority_min(policy);
		priority_max = sched_get_priority_max(policy);
	}
	assert(priority_min <= priority_max);

	/*
	 * If pthread priorities are not supported, use the old setpriority
	 * (aka nice) mechanism of 0 (high) to 20 (low).
	 */
	if (priority_min == priority_max) {
#ifdef USE_NICE
		priority_min = 0;
		priority_max = 20;
#else
		nopriorities = 1;
		return;
#endif
	}

	/*
	 * For mapping to/from pthread/NodeOS priorities
	 */
	pt_scale = cpucycles / (priority_max - priority_min);
	assert(pt_scale != 0);
	an_scale = cpucycles / (AN_THREAD_PRIO_EXCLUSIVE - AN_THREAD_PRIO_IDLE);
	assert(an_scale != 0);
}

static inline unsigned int
anprio_to_cpurate(int prio)
{
	assert(cpucycles != 0);
	return (prio - AN_THREAD_PRIO_IDLE) * an_scale;
}

static inline int
cpurate_to_anprio(unsigned int cpurate)
{
	assert(cpucycles != 0);

	/*
	 * XXX possible due to really-high-priority root flow threads
	 */
	if (cpurate >= cpucycles)
		return AN_THREAD_PRIO_EXCLUSIVE;

	return cpurate / an_scale + AN_THREAD_PRIO_IDLE;
}

/*
 * Initialize the scheduling attributes for an embryonic thread.
 */
int
ani_os_sched_setparam(ani_thread_t *thread, ani_threadpool_t *tpool)
{
	int err;
	unsigned int cpurate;

	if (nopriorities)
		return 0;

	/*
	 * XXX threadpool cpurate should be shared by all active threads in
	 * the pool.  That would involve dynamic adjustment of existing
	 * thread's cpu rates.  For now we just divide up the cpurate
	 * among all possible threads in the pool.
	 */
	cpurate = tpool->res.cpurate / tpool->res.threads;
	assert(cpurate != 0);

	/*
	 * Map CPU cycle rate onto pthread priorities
	 * XXX note hack for root flow real-high-priority threads
	 */
	if (cpurate >= cpucycles)
		thread->pprio.sched_priority = priority_max;
	else
		thread->pprio.sched_priority =
			priority_min + cpurate / pt_scale;


#ifdef USE_NICE
	assert(thread->pprio.sched_priority <= 20); 
	err = setpriority(PRIO_PROCESS, 0, 20 - thread->pprio.sched_priority);
#else
	err = pthread_attr_setschedparam(&thread->attr, &thread->pprio);
#endif

	/*
	 * XXX map cpurate onto a NodeOS priority value.
	 * In our implementation these priority values are not scheduler
	 * specific as we have only one scheduler in any instantiation
	 * of the NodeOS.
	 */
	if (err == 0)
		thread->nprio = cpurate_to_anprio(cpurate);

	return err;
}

/*
 * Change the scheduling priority of an existing thread.
 * This routine should map standard NodeOS priority values into
 * implementation specific priorities.
 */
int
ani_os_sched_setprio(ani_thread_t *thread, int nprio)
{
	struct sched_param oparam;
	unsigned int cpurate;
	int err;

	assert(cpucycles != 0);

	if (nopriorities)
		return 0;

	if (nprio < AN_THREAD_PRIO_IDLE || nprio > AN_THREAD_PRIO_EXCLUSIVE)
		return EINVAL;

	oparam = thread->pprio;

	cpurate = anprio_to_cpurate(nprio);
	thread->pprio.sched_priority = priority_min + cpurate / pt_scale;

#ifdef USE_NICE
	assert(thread->pprio.sched_priority <= 20); 
	err = setpriority(PRIO_PROCESS, 0, 20 - thread->pprio.sched_priority);
#else
	err = pthread_setschedparam(thread->id, -1, &thread->pprio);
#endif
	if (err != 0)
		thread->pprio = oparam;

	return err;
}

int
ani_os_sched_getcpu(ani_thread_t *thread, ani_threadcpu_t *cpu)
{
	return ENODEV;
}
