/*
 * Copyright (c) 2000 The University of Utah and the Flux Group.
 * All rights reserved.
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */

/*
 * UNIX timer layer.
 *
 * Can use one of three methods.  Some methods work better than others on
 * various UNIX systems.  Run the eventtest EE and look at the output.
 * You should get about 100 events per printf (i.e. per second) and the
 * time displayed should increase by one second each time.  If you only
 * get around 50 events per second or the second counter goes up by 2-3
 * seconds per printf, try another method.  The methods are:
 *
 * Use select of a null descriptor set and a timeout.  Works well for
 * Solaris.  Also works for Linux but select never seems to return at some
 * point during shutdown (as result of a cancelation?)  The select method
 * does not work well for FreeBSD 3.x.
 *
 * Use a cond_timedwait with the appropriate timeout value and a cond variable
 * that is never signalled.  Works well for FreeBSD 3.x and Linux.
 * Doesn't work well for Solaris.
 *
 * Use nanosleep.  Doesn't work well on anything so far.
 */
#include <unistd.h>
#include <stdlib.h>
#include <pthread.h>

#include <nodeos/impl/ani_flow.h>
#include "os.h"

#if defined(__linux__) || defined(__sun__)
#define USE_SELECT
#else
#define USE_CONDWAIT
#endif

#define	TIMER_FREQ	100	/* XXX most unixes have a 10ms timer */

struct ani_os_timer {
	void (*func)(void *, void *);
	void *arg;
	unsigned long res;
	volatile int isrunning;
	ani_thread_t thread;
#ifdef USE_CONDWAIT
	ani_ilock_t mutex;
	ani_icond_t cond;
#endif
};

/*
 * Initialize a periodic timer for NodeOS use.
 *
 * Returns a handle to the timer created.  The maximum resolution
 * of the timer in microseconds is returned in the timerres param.
 */
void *
ani_os_timer_init(void (*func)(void *, void *), void *farg,
		  unsigned long *timerres)
{
	struct ani_os_timer *timer;

	timer = malloc(sizeof *timer);
	if (timer != 0) {
		timer->func = func;
		timer->arg = farg;
		timer->res = 0;
#ifdef USE_CONDWAIT
		ani_ilock_init(&timer->mutex);
		ani_icond_init(&timer->cond);
#endif

		*timerres = 1000000 / TIMER_FREQ;
	}

	return timer;
}

/*
 * Destroy a periodic timer.
 */
void
ani_os_timer_deinit(void *handle)
{
	struct ani_os_timer *timer = handle;

	timer->isrunning = 0;
	ani_thread_kill(&timer->thread, 1);
}

/*
 * Start a periodic timer with the specified frequency.
 */
int
ani_os_timer_start(void *handle, unsigned long timerres)
{
	static void *timer_handler(void *arg);
	struct ani_os_timer *timer = handle;
	ani_error_t err;

	assert(timerres < 1000000);
	timer->res = timerres;
	timer->isrunning = 1;

	/*
	 * Create a thread to just do cond_timedwait and invoke the interrupt
	 * handler whenever it wakes up.
	 */
	err = ani_rootflow_startthread(&timer->thread,
				       (ani_thread_func_t)timer_handler,
				       timer, 0, 0);
	if (ani_error_occured(err)) {
		timer->isrunning = 0;
		return ani_error_code(err);
	}

	return err;
}

#ifdef USE_SELECT
static void *
timer_handler(void *arg)
{
	struct ani_os_timer *timer = arg;
	struct timeval tv;

	while (timer->isrunning) {
		tv.tv_sec = 0;
		tv.tv_usec = timer->res;
		(void) select(0, 0, 0, 0, &tv);
		timer->func(0, timer->arg);
	}

	return 0;
}
#endif

#ifdef USE_CONDWAIT
static void *
timer_handler(void *arg)
{
	struct ani_os_timer *timer = arg;
	struct timespec tspec;
	struct timeval tv;
	int err;

	ani_ilock_lock(&timer->mutex);
	gettimeofday(&tv, 0);
	tspec.tv_sec = tv.tv_sec;
	tspec.tv_nsec = tv.tv_usec * 1000;

	while (timer->isrunning) {
		tspec.tv_nsec += (timer->res * 1000);
		if (tspec.tv_nsec >= 1000000000) {
			tspec.tv_nsec -= 1000000000;
			tspec.tv_sec++;
		}
		do {
			err = ani_icond_timedwait(&timer->cond, &timer->mutex,
						  tspec.tv_sec, tspec.tv_nsec);
		} while (err != ETIMEDOUT);

		timer->func(0, timer->arg);
	}

	return 0;
}
#endif

#ifdef USE_NANOSLEEP
static void *
timer_handler(void *arg)
{
	struct ani_os_timer *timer = arg;
	struct timespec tspec;

	tspec.tv_sec = 0;
	tspec.tv_nsec = timer->res * 1000;

	while (timer->isrunning) {
		nanosleep(&tspec, 0);
		timer->func(0, timer->arg);
	}

	return 0;
}
#endif
