/*
 * Copyright (c) 2000 The University of Utah and the Flux Group.
 * All rights reserved.
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */

#include <assert.h>
#include <sys/types.h>
#ifdef OSKIT
#include <oskit/io/netio.h>
#include <oskit/io/bufio.h>
#endif

#include <nodeos/impl/ani_flow.h>
#include <nodeos/impl/ani_chan.h>
#include <nodeos/impl/ani_mempool.h>
#include <nodeos/impl/ani_pbuf.h>

#include "os.h"

/*
 * Packet buffer manipulation
 */

#ifndef OSKIT
/*
 * XXX sigh...even though we now control Linux skbuff allocation,
 * it still isn't enough.  The bufio interface doesn't have a buffer
 * size independent of the data size.  When we receive a bufio via the
 * push method, we can only discover its current size, not its max size.
 * At first glance that seems ok, since we know the underlying buffer
 * has to be NODEOS_CHAN_BUFSIZE since we allocated it.  But there is
 * no way to "reset" a bufio to its maximum size (before we put it back
 * on a channel's pbuf list) so we cannot use the bufio interfaces to
 * access the complete buffer even though we know it should be ok.
 * Until we come up with a new oskit netio buffer interface, we are still
 * stuck copying all but the max sized packet.
 */
#undef NODEOS_CHAN_COPYSIZE
#define NODEOS_CHAN_COPYSIZE	1513
#endif

#ifdef OSKIT
static void
pbuffer_skbufiofree(ani_pbuf_t *pbuf, void *arg)
{
	oskit_skbufio_metadata_unmap(pbuf->skbufio, pbuf->skb);
	oskit_skbufio_release(pbuf->skbufio);
	pbuf->skbufio = 0;
}
#else
/*
 * Allocate system-wide packet buffer memory.
 * Right now we just malloc every time.  Someday we might manage
 * a cache of memory.
 *
 * @allocation: "size" bytes charged to the given flow.
 */
static ani_pbuf_data_buffer_t *
pbuffer_memalloc(ani_flow_t *flow, size_t size)
{
	ani_pbuf_data_buffer_t *retval;
	
	assert_ani_flow_valid(flow);
	assert(size >= 0);

	/* XXX round up size or verify it is a valid size */

	retval = ani_mempool_alloc(ani_flow_mempool(flow),
				   sizeof(ani_pbuf_data_buffer_t) + size);
	retval->refs = 0;
	retval->off = 0;
	retval->size = 0;
	retval->bufsize = size;
	retval->bufdata = retval->bufmem;
	return retval;
}

/*
 * Free system-wide packet buffer memory.
 * Called with a pbuf when its reference count drops to zero.
 * Frees just the buffer, not the header.
 *
 * @deallocation: pbuf->bufsize bytes returned to pbuffer's flow.
 */
static void
pbuffer_memfree(ani_pbuf_t *pbuf, void *arg)
{
	assert(pbuf != 0);
	assert_ani_flow_valid(pbuf->flow);
	assert(pbuf->dbuffer != 0);
	assert(pbuf->dbuffer->bufsize > 0);

	pbuf->dbuffer->refs--;
	if (pbuf->dbuffer->refs == 0) {
		ani_mempool_free(ani_flow_mempool(pbuf->flow),
				 pbuf->dbuffer, pbuf->dbuffer->bufsize);
	}
	pbuf->dbuffer = 0;
}
#endif

/*
 * Initialize the given pbuf with the indicated info.
 */
void
ani_pbuf_init_external(ani_pbuf_t *pbuf, ani_flow_t *flow, ani_chan_t *chan,
#ifdef OSKIT
		       struct sk_buff *skb)
#else
		       ani_pbuf_data_buffer_t *dbuf,
		       ani_pbuf_free_t ffunc, void *farg)
#endif
{
	assert(pbuf != 0);
	assert_ani_flow_valid(flow);

	memset(pbuf, 0, sizeof *pbuf);

#ifdef OSKIT
	assert(skb != 0);

	pbuf->skbufio = skb_skbufio(skb);
	pbuf->skb = skb;
	pbuf->ffunc = pbuffer_skbufiofree;
#else
	assert(ffunc != 0);
	assert(dbuf != 0);
	assert(dbuf->off >= 0);
	assert(dbuf->off + dbuf->size <= dbuf->bufsize);

	dbuf->refs++;
	pbuf->dbuffer = dbuf;
	pbuf->ffunc = ffunc;
	pbuf->farg = farg;
#endif
	pbuf->refcount = 1;
	pbuf->flow = flow;
	pbuf->chan = chan;
}

/*
 * Allocate a buffer and associate it with the given pbuf object.
 */
ani_error_t
ani_pbuf_init(ani_pbuf_t *pbuf, ani_flow_t *flow, ani_chan_t *chan,
	      size_t bsize, size_t off, size_t size)
{
#ifdef OSKIT
	oskit_skbufio_t *skbufio;
	struct sk_buff *skb;
	oskit_error_t err;

	assert(off + size <= bsize);

	err = oskit_skbufio_create(size, off, bsize-(off+size), &skbufio);
	if (err != 0)
		return ani_error_set(err, "oskit_skbufio_create failed");
	err = oskit_skbufio_metadata_map(skbufio, SKB_VERSION_NUMBER, &skb);
	assert(err == 0);

	ani_pbuf_init_external(pbuf, flow, chan, skb);
#else
	ani_pbuf_data_buffer_t *bdata;

	bdata = pbuffer_memalloc(flow, bsize);
	bdata->size = size;
	if (bdata == 0) {
		ani_mempool_free(ani_flow_mempool(flow), pbuf, sizeof *pbuf);
		return ani_error_set(ENOMEM, "no memory for pbuffer");
	}

	ani_pbuf_init_external(pbuf, flow, chan, bdata, pbuffer_memfree, 0);
#endif

	return ani_error_clear();
}

void
ani_pbuf_deinit(ani_pbuf_t *pbuf)
{
	assert(pbuf != 0);

	/* XXX remove pbuf from any channel lists it is on */
	assert(!ani_pbuf_onlist(pbuf));

	if (ani_pbuf_isvalid(pbuf))
		ani_pbuf_release(pbuf);
}

/*
 * Create a new pbuffer referencing the same data buffer.
 */
ani_error_t
ani_pbuf_clone(ani_pbuf_t *pbuf, ani_pbuf_t *newpbuf)
{
	ani_error_t err;

	assert(pbuf != 0 && newpbuf != 0);
	assert_ani_flow_valid(pbuf->flow);
	//assert(pbuf->chan == 0);

#ifdef OSKIT
	assert(pbuf->skb != 0);
	oskit_skbufio_addref(pbuf->skbufio);
	ani_pbuf_init_external(newpbuf, pbuf->flow, 0, pbuf->skb);
	err = ani_error_clear();
#else
	ani_pbuf_init_external(newpbuf, pbuf->flow, 0,
			       pbuf->dbuffer, pbuffer_memfree, 0);
	err = ani_error_clear();
#endif

	return err;
}

/*
 * Create a new pbuffer, copying the data buffer.
 */
ani_error_t
ani_pbuf_copy(ani_pbuf_t *pbuf, ani_pbuf_t **newpbuf)
{
	ani_error_t err;

	assert(pbuf != 0 && newpbuf != 0);
	assert_ani_flow_valid(pbuf->flow);
	assert(pbuf->chan);

#ifdef OSKIT
	/*
	 * Get a large enough buffer from the channel.
	 * If it can't pay, we are hosed.
	 */
	err = ani_chan_pbuf_alloc(pbuf->chan, ani_pbuf_bufsize(pbuf), newpbuf);
	if (!ani_error_occured(err)) {
		ani_pbuf_t *npbuf = *newpbuf;

		/*
		 * XXX naked sk_buff field refs
		 */
		npbuf->skb->head = npbuf->skb->data + ani_pbuf_offset(pbuf);
		npbuf->skb->tail = npbuf->skb->head + ani_pbuf_size(pbuf);
		npbuf->skb->len = ani_pbuf_size(pbuf);
		memcpy(ani_pbuf_data(npbuf), ani_pbuf_data(pbuf),
		       ani_pbuf_size(pbuf));
		err = ani_error_clear();
	} else {
		/*
		 * XXX undo failure stats
		 */
		CHANSTATS(pbuf->chan->stats.droppackets--);
		CHANSTATS(pbuf->chan->stats.dropbytes -=
			  ani_pbuf_bufsize(pbuf));
		CHANSTATS(pbuf->chan->stats.dropnobufs--);
	}
#else
	err = ani_error_set(AN_ERR_NOTSUPPORTED,
			    "ani_pbuf_copy hasn't been written yet.");
#endif

	return err;
}

/*
 * Set the offset of the active region of the pbuffer.
 * Implicitly changes the size of the active region of the buffer.
 */
ani_error_t
ani_pbuf_setoffset(ani_pbuf_t *pbuf, size_t newoff, void **out_data)
{
	size_t offset, eoffset;

	assert(pbuf != 0);
	assert(ani_pbuf_isvalid(pbuf));

	offset = ani_pbuf_offset(pbuf);
	if (newoff != offset) {
		eoffset = offset + ani_pbuf_size(pbuf);
		if (newoff < 0 || newoff > eoffset)
			return ani_error_set(EINVAL,
					     "invalid pbuf offset specified");

#ifdef OSKIT
		newoff -= offset;
		if ((int) newoff < 0)
			skb_push(pbuf->skb, -newoff);
		else
			skb_pull(pbuf->skb, newoff);
#else
		pbuf->dbuffer->off = newoff;
		pbuf->dbuffer->size = eoffset - newoff;
#endif
	}

	if (out_data)
		*out_data = ani_pbuf_data(pbuf);
	return ani_error_clear();
}

/*
 * Set the size of the active region of the pbuffer.
 */
ani_error_t
ani_pbuf_setsize(ani_pbuf_t *pbuf, size_t newsize, void **out_data)
{
	size_t size;

	assert(pbuf != 0);
	assert(ani_pbuf_isvalid(pbuf));

	size = ani_pbuf_size(pbuf);
	if (newsize != size) {
		if (newsize < 0 ||
		    ani_pbuf_offset(pbuf) + newsize > ani_pbuf_bufsize(pbuf))
			return ani_error_set(EINVAL,
					     "invalid pbuf size specified");

#ifdef OSKIT
		newsize -= size;
		if ((int) newsize > 0)
			skb_put(pbuf->skb, newsize);
		else
			skb_get(pbuf->skb, -newsize);
#else
		pbuf->dbuffer->size = newsize;
#endif
	}

	if (out_data)
		*out_data = ani_pbuf_data(pbuf);

	return 0;
}

/*
 * Open a buffer of to its max size.
 * Returns a pointer to the active region.
 */
void *
ani_pbuf_setmax(ani_pbuf_t *pbuf)
{
	assert(pbuf != 0);
	assert(ani_pbuf_isvalid(pbuf));

#ifdef OSKIT
	/* XXX naked sk_buff field refs */
	pbuf->skb->data = pbuf->skb->head;
	pbuf->skb->tail = pbuf->skb->end;
	pbuf->skb->len = pbuf->skb->end - pbuf->skb->head;
#else
	pbuf->dbuffer->off = 0;
	pbuf->dbuffer->size = pbuf->dbuffer->bufsize;
#endif

	return ani_pbuf_data(pbuf);
}

/*
 * Swap the memory buffers of two buffer objects.
 */
void
ani_pbuf_swap(ani_pbuf_t *pbuf1, ani_pbuf_t *pbuf2)
{
	ani_pbuf_t tmp;

	assert(pbuf1->refcount == 1);
	assert(pbuf2->refcount == 1);
	assert(ani_pbuf_bufsize(pbuf1) == ani_pbuf_bufsize(pbuf2));

	tmp = *pbuf1;
	*pbuf1 = *pbuf2;

	/*
	 * Retain flow and channel association
	 * XXX need to adjust accounting if buffer sizes don't match.
	 */
	pbuf1->flow = tmp.flow;
	pbuf1->chan = tmp.chan;
	tmp.flow = pbuf2->flow;
	tmp.chan = pbuf2->chan;

	*pbuf2 = tmp;
}

void
ani_pbuf_return(ani_pbuf_t *pbuf)
{
	assert(ani_pbuf_isvalid(pbuf));

	if (pbuf->chan != 0) {
		ani_chan_lock(pbuf->chan);
		ani_chan_pbuf_free(pbuf->chan, pbuf);
		ani_chan_unlock(pbuf->chan);
	} else
		ani_pbuf_deinit(pbuf);
}

#ifdef OSKIT
/*
 * Wrap an OSKit bufio with a NodeOS packet buffer.
 * The caller should have a reference to the bufio.  This routine will
 * take a bufio reference if successful.
 */
ani_error_t
ani_pbuf_frombufio(ani_flow_t *flow, ani_chan_t *chan, oskit_bufio_t *bufio,
		   size_t off, size_t size, ani_pbuf_t **out_pbuf)
{
	size_t bsize;
	ani_pbuf_t *pbuf;
	ani_error_t err;
	int ierr;
	oskit_skbufio_t *skbufio;
	struct sk_buff *skb;

	assert_ani_flow_valid(flow);
	assert(bufio != 0);
	assert(off >= 0);

	/*
	 * Extract the skbufio interface and map the skbuff meta info.
	 * This call grabs a reference to the skbufio.
	 */
	ierr = oskit_bufio_query(bufio, &oskit_skbufio_iid, (void **)&skbufio);
	if (ierr != 0) {
		CHANSTATS(chan->stats.droppackets++);
		CHANSTATS(chan->stats.dropfailedbuf++);
		return ani_error_set(EINVAL, "oskit_bufio_query failed");
	}
	ierr = oskit_skbufio_metadata_map(skbufio, SKB_VERSION_NUMBER, &skb);
	if (ierr != 0) {
		oskit_skbufio_release(skbufio);
		CHANSTATS(chan->stats.droppackets++);
		CHANSTATS(chan->stats.dropfailedbuf++);
		return ani_error_set(EINVAL,
				     "oskit_skbufio_metadata_map failed");
	}
	bsize = skb->end - skb->head;
	assert(bsize == NODEOS_CHAN_BUFSIZE);

	/*
	 * Get a large enough buffer from the channel.
	 * If it can't pay, we drop the packet.
	 */
	err = ani_chan_pbuf_alloc(chan, bsize, &pbuf);
	if (err) {
		oskit_skbufio_metadata_unmap(skbufio, skb);
		oskit_skbufio_release(skbufio);
		return err;
	}
	assert(pbuf->flow == flow);

	/*
	 * XXX free the actual data buffer as we cannot return it
	 * to the OSKit driver to use right now.
	 */
	ani_pbuf_release(pbuf);

	/*
	 * Pop the header if required and initialize the pbuf with the
	 * new buffer info.
	 */
	if (off != 0)
		skb_pull(skb, off);
	ani_pbuf_init_external(pbuf, flow, chan, skb);

	*out_pbuf = pbuf;
	return ani_error_clear();
}

void
ani_pbuf_senddone(void *arg)
{
	ani_pbuf_t *pbuf = arg;
	int intson;

	intson = ani_os_intr_save_disable();
	assert(pbuf->skb != 0);
	/*
	 * XXX hack.
	 * If we prepended a header in tobufio below, we must pop it off
	 * again so that caller does not have the pbuf mysteriously change
	 * size after a channel_send.
	 */
	if (pbuf->popoff != 0) {
		skb_pull(pbuf->skb, pbuf->popoff);
		pbuf->popoff = 0;
	}

	/*
	 * If the pbuf has been returned to the free list,
	 * put it there now.
	 */
	if (ani_pbuf_onlist(pbuf)) {
		ani_chan_t *chan = pbuf->chan;
#if 0
		/* make sure it is really on the bbufs list */
		ani_pbuf_t *pb;
		ani_queue_iterate(&chan->tdata.in.bbufs, pb, ani_pbuf_t *, link)
			if (pb == pbuf)
				break;
		assert(!ani_queue_end(&chan->tdata.in.bbufs, (ani_queue_t)pb));
#endif
		assert_ani_chan_valid(chan);
		ani_pbuflist_rempbuf(&chan->tdata.in.bbufs, pbuf);
		ani_pbuflist_addtail(&chan->tdata.in.rbufs, pbuf);
	}
	pbuf->isbusy = 0;
	ani_pbuf_release(pbuf);

	if (intson)
		ani_os_intr_enable();
}

ani_error_t
ani_pbuf_tobufio(ani_pbuf_t *pbuf, void *hdr, size_t hsize, oskit_netio_t *nio,
		 oskit_bufio_t **out_bufio, int *didcopy)
{
	oskit_skbufio_t *skbufio;
	struct sk_buff *skb;
	oskit_size_t psize;
	int ierr;

	assert(pbuf->skb != 0);

	/*
	 * Make sure there is room for the header in the skbuff.
	 * If not, we'll have to create a new skbuff.
	 */
	skb = pbuf->skb;
	if (skb_headroom(skb) >= hsize) {
		assert(pbuf->skbufio == skb_skbufio(skb));
		if (hsize != 0) {
			memcpy(skb_push(skb, hsize), hdr, hsize);
			pbuf->popoff = hsize;
		}
		skbufio = pbuf->skbufio;
		oskit_bufio_addref((oskit_bufio_t *)skbufio);
		*didcopy = 0;
	} else {
		/*
		 * Create a new skbufio and copy the data.
		 * XXX should be a netio function to return an skbufio?
		 */
		psize = ani_pbuf_size(pbuf);
		assert(hsize + psize <= NODEOS_CHAN_BUFSIZE);
		ierr = oskit_skbufio_create(0, 0, NODEOS_CHAN_BUFSIZE, &skbufio);
		if (ierr != 0)
			return ani_error_set(ierr, "oskit_skbufio_create failed");
		ierr = oskit_skbufio_metadata_map(skbufio, SKB_VERSION_NUMBER,
						  &skb);
		if (ierr != 0) {
			oskit_skbufio_release(skbufio);
			return ani_error_set(ierr,
					     "oskit_skbufio_metadata map failed");
		}
		if (hsize)
			memcpy(skb_put(skb, hsize), hdr, hsize);
		memcpy(skb_put(skb, psize), ani_pbuf_data(pbuf), psize);
		oskit_skbufio_metadata_unmap(skbufio, skb);
		*didcopy = 1;
	}

	/*
	 * Wrap the skbufio so we can get notification when the send is done.
	 * Takes a reference to the pbuf since our wrapper skbufio is part of
	 * it!
	 *
	 * Note that wrap_skbufio takes a reference to the bufio so we can drop
	 * the one we took (explicitly or implicitly via the create) after the
	 * call returns.
	 */
	ani_os_wrap_skbufio(&pbuf->sendio, skbufio, ani_pbuf_senddone, pbuf);
	oskit_skbufio_release(skbufio);
	ani_pbuf_reference(pbuf);
	*out_bufio = (oskit_bufio_t *)&pbuf->sendio;
	return ani_error_clear();
}

#else	/* OSKIT */

ani_error_t
ani_pbuf_tomem(ani_pbuf_t *pbuf, void *hdr, size_t hsize,
	       void **out_buf, size_t *out_size, int *didcopy)
{
	void *buf;

	/*
	 * See if we can prepend the header to the pbuf.
	 * As in the tobufio case above, we don't modify the pbuf offset.
	 */
	if (pbuf->dbuffer->off >= hsize) {
		buf = ani_pbuf_data(pbuf);
		if (hsize != 0) {
			buf -= hsize;
			memcpy(buf, hdr, hsize);
		}

		*out_buf = buf;
		*out_size = ani_pbuf_size(pbuf) + hsize;
		*didcopy = 0;
		return ani_error_clear();
	}

	/*
	 * Otherwise, copy the header and packet to a contiguous buffer.
	 */
	buf = ani_mempool_alloc(ani_flow_mempool(ani_flow_current()),
				ani_pbuf_size(pbuf) + hsize);
	if (buf == 0)
		return ani_error_set(ENOMEM, "no memory for pbuf");

	memcpy(buf, hdr, hsize);
	memcpy(buf+hsize, ani_pbuf_data(pbuf), ani_pbuf_size(pbuf));

	*out_buf = buf;
	*out_size = ani_pbuf_size(pbuf) + hsize;
	*didcopy = 1;
	return ani_error_clear();
}

void
ani_pbuf_relmem(void *buf, size_t bufsize)
{
	ani_mempool_free(ani_flow_mempool(ani_flow_current()), buf, bufsize);
}
#endif	/* OSKIT */

