/*
 * Copyright (c) 1999, 2000 The University of Utah and the Flux Group.
 * All rights reserved.
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */

/*
 *
 */
#include <stdlib.h>
#include <string.h>
#include <assert.h>

#include <nodeos/an_resource.h>
#include <nodeos/an_thread.h>

#include <nodeos/impl/an_error.h>
#include <nodeos/impl/an_impl.h>
#include <nodeos/impl/ani_thread.h>
#include <nodeos/impl/ani_nodeos.h>

#include <nodeos/impl/ani_threadpool.h>
#include <nodeos/impl/ani_flow.h>
#include <nodeos/impl/ani_chan.h>
#include <nodeos/impl/ani_mempool.h>
#include <nodeos/impl/ani_event.h>

#include "os.h"
#include "ani_debug.h"

static ani_error_t ani_threadpool_thread_dowork(ani_threadpool_t *tpool,
						ani_threadwork_t *work,
						int canqueue, int onlyqueue,
						ani_thread_t **out_thread);

ani_error_t
ani_threadpool_init(ani_threadpool_t *tpool, ani_flow_t *flow,
		    an_cpuspec_t *cspec, ani_resource_t *res)
{
	int i;
	int nthreads;
	ani_error_t err;

	err = ani_error_clear();
	memset(tpool, 0, sizeof *tpool);

	/*
	 * Attach to a flow.  Note that the threadpool will keep the flow
	 * alive.  This is necessary to ensure that threads have a context
	 * in which to free their resources.
	 */
	tpool->flow = flow;
	ani_flow_reference(flow);

	/*
	 * Record user-supplied specification and normalized resources.
	 */
	tpool->params = *cspec;
	if (res == 0)
		tpool->res = flow->res;
	else
		tpool->res = *res;

	ani_icond_init(&tpool->waiting);
	tpool->refs = 1;
	ani_ilock_init(&tpool->lock);

	/*
	 * Allocate and initialize threads.
	 */
	ani_threadpool_lock(tpool);
	if (cspec->ttype == AN_TH_IMPLICIT) {
		nthreads = ANI_THREADPOOL_CTHREADS;
		if (nthreads > cspec->threads.im.maxthreads)
			nthreads = cspec->threads.im.maxthreads;

		for (i = 0; i < nthreads; i++) {
			err = ani_thread_spawn(tpool, 0);
			if (ani_error_occured(err))
				break;
		}
	} else {
		ani_thread_t *thread;
		an_thread_t **xtlist;

		nthreads = cspec->threads.ex.nthreads;
		xtlist = cspec->threads.ex.threads;
		for (i = 0; i < nthreads; i++) {
			thread = NODEOS_INTERN_THREAD(xtlist[i]); /* XXX */
			if (thread == 0) {
				err = ani_error_set(EINVAL,
						    "invalid thread argument");
				break;
			}
			err = ani_thread_start(thread, tpool);
			if (ani_error_occured(err))
				break;
		}
	}

	/*
	 * Wait for the number of threads we did successfully start
	 * to reach the service loop.  We do this even if there was an
	 * error so that we can be sure that the threadpool_deinit will
	 * kill them all.
	 */
	tpool->reserve = i;
	while (tpool->freethreads != tpool->reserve) {
		tpool->flags |= TPOOL_WAITING;
		ani_icond_wait(&tpool->waiting, &tpool->lock);
	}
	ani_threadpool_unlock(tpool);

	if (ani_error_occured(err)) {
		ani_threadpool_disable(tpool, 1);
		ani_threadpool_deinit(tpool, 1);
	}

	return err;
}

/*
 * Finish off a threadpool.  Kill off any remaining active threads.
 * The pool should already have been disabled; i.e., there should be no
 * cached threads at this point.
 */
void
ani_threadpool_deinit(ani_threadpool_t *tpool, int waitfor)
{
	ani_thread_t *th, *me;

	assert_ani_threadpool_valid(tpool);
	assert((tpool->flags & TPOOL_DYING) != 0);
	assert(tpool->refs > tpool->curthreads+tpool->curosthreads);

	me = ani_thread_current();

	ani_threadpool_lock(tpool);

	DBG(DBG_THREADPOOLS, {
		ani_dprintf("TP%p: deinit: %d + %d threads\n",
			    tpool, tpool->curthreads, tpool->curosthreads);
	});

	/*
	 * Cancel all active pool and one-shot threads.
	 */
	for (th = tpool->threads; th != 0; th = th->next) {
		assert((th->flags & THREAD_ONESHOT) == 0);
		if (th != me)
			ani_thread_kill_locked(th, 0);
		else
			assert(waitfor == 0);
	}
	for (th = tpool->osthreads; th != 0; th = th->next) {
		assert((th->flags & THREAD_ONESHOT) != 0);
		if (th != me)
			ani_thread_kill_locked(th, 0);
		else
			assert(waitfor == 0);
	}

	/*
	 * Wait for them to finish
	 */
	if (waitfor) {
		DBG(DBG_THREADPOOLS, {
			ani_dprintf("TP%p: deinit: waiting for thread death\n",
				    tpool);
		});
		while (tpool->curthreads + tpool->curosthreads != 0) {
			tpool->flags |= TPOOL_WAITING;
			ani_icond_wait(&tpool->waiting, &tpool->lock);
		}
		DBG(DBG_THREADPOOLS, {
			ani_dprintf("TP%p: deinit: threads dead\n", tpool);
		});
	}

	ani_threadpool_unlock(tpool);

	/*
	 * Remove all unscheduled (future) events.
	 * We do this after killing all threads and after releasing the
	 * threadpool lock.  These are necessary to adhere to the locking
	 * protocol.
	 */
	ani_event_removeall(tpool);

	ani_threadpool_release(tpool);
}

/*
 * Permanently disable a threadpool.  Detach all the cached threads
 * and mark the pool as dead so no future activity is initiated.
 * Currently running threads are not killed.
 */
void
ani_threadpool_disable(ani_threadpool_t *tpool, int waitfor)
{
	assert_ani_threadpool_valid(tpool);

	ani_threadpool_lock(tpool);
	tpool->flags |= TPOOL_DYING;
	ani_threadpool_thread_setreserve(tpool, 0, waitfor);
	ani_threadpool_unlock(tpool);
}

void
ani_threadpool_cleanup(ani_threadpool_t *tpool)
{
	assert_ani_threadpool_valid(tpool);
	assert((tpool->flags & TPOOL_DYING) != 0);
	assert(ani_threadpool_islocked(tpool) == 0);
	assert(tpool->curthreads+tpool->curosthreads == 0);

	assert(tpool->flow);
	ani_flow_release(tpool->flow);
#ifdef ANI_THREADPOOL_STATS
	DBG(DBG_THREADPOOL_STATS, {
		static void ani_threadpool_printstats(ani_threadpool_t *);
		ani_threadpool_printstats(tpool);
	});
#endif
	ani_icond_deinit(&tpool->waiting);
	ani_ilock_deinit(&tpool->lock);
}

/*
 * Cancellation handler.  Active while a thread is waiting in thread_wait
 * for another thread to terminate.  If triggered, marks that the cancelled
 * thread is no longer interested in the target thread and signals the
 * target to make sure it doesn't get stuck waiting for a reply from the
 * cancelled thread.
 */
static void
ani_threadpool_thread_wait_ch(void *arg)
{
	ani_thread_t *thread = arg;
	ani_threadpool_t *tpool;

	DBG(DBG_THREADS|DBG_THREADPOOLS, {
		ani_dprintf("TP%p: T%p cancelled waiting for T%p\n",
			    tpool, ani_thread_current(), thread);
	});
	tpool = thread->pool;
	assert_ani_threadpool_valid(tpool);
	assert(ani_threadpool_islocked(tpool));

	thread->flags &= ~(THREAD_WAITEDON|THREAD_DONE);
	ani_icond_signal(&thread->waiting);

	ani_threadpool_unlock(tpool);
}

/*
 * Wait for the indicated thread to terminate or until the given time has
 * elapsed.  Threadpool should be locked on call.
 *
 * XXX implement timeout.
 */
ani_error_t
ani_threadpool_thread_wait(ani_threadpool_t *tpool, ani_thread_t *thread,
			   an_timespec_t *timo)
{
	assert_ani_threadpool_valid(tpool);
	assert(ani_threadpool_islocked(tpool));
	assert_ani_thread_valid(thread);

	/*
	 * There is a little synchronization dance here:
	 *
	 * 1. we wait til the target thread completes
	 *	(flags & THREAD_WAITEDON) == 0
	 * 2. it does its job, signals us, and then waits til we tell it "ok"
	 *	(flags & THREAD_DONE) == 0
	 * 3. we wakeup, extract whatever we need from the thread's state,
	 *	signal it, and go on about our business
	 * 4. the target wakesup, and goes back to process another request.
	 *
	 * Right now, the only state we look at in the target is the flags
	 * field to detect that the THREAD_WAITEDON bit has been cleared.
	 * But, alas, that is enough.  The target must wait for us to see
	 * the bit, otherwise we could block forever if someone quickly
	 * did a thread_run, got the same thread and reset the bit.
	 *
	 * XXX note that this will not work if more than one thread tried
	 * waiting on a single target.
	 */
	DBG(DBG_THREADS|DBG_THREADPOOLS, {
		ani_dprintf("TP%p: wait: T%p waiting for T%p (WAITEDON)\n",
			    tpool, ani_thread_current(), thread);
	});
	assert((thread->flags & THREAD_WAITEDON) == 0);
	thread->flags |= THREAD_WAITEDON;
	while ((thread->flags & THREAD_WAITEDON) != 0)
		ani_icond_wait_withhandler(&thread->waitedon, &tpool->lock,
					   ani_threadpool_thread_wait_ch,
					   thread);

	/*
	 * Allow it to go service another request or terminate.
	 */
	assert((thread->flags & THREAD_DONE) != 0);
	thread->flags &= ~THREAD_DONE;
	DBG(DBG_THREADS|DBG_THREADPOOLS, {
		ani_dprintf("TP%p: wait: T%p signalling DONE on T%p\n",
			    tpool, ani_thread_current(), thread);
	});
	ani_icond_signal(&thread->waiting);

	return ani_error_clear();
}

/*
 * Start a thread from the given pool running to do the indicated work.
 * If there are no threads available and the pool is at the thread limit,
 * returns an error.  If the pool is under limit but there are no cached
 * threads available, the call is delayed while a new thread is started.
 * That is, upon return from this call, an actual thread has been commited
 * to run the function.  If out_threadp is non-zero, the call returns a
 * pointer to the new thread.  The threadwork record may be reused upon
 * return from this call.  Returns zero if a thread was successfully
 * started, an error otherwise.
 */
ani_error_t
ani_threadpool_thread_start(ani_threadpool_t *tpool, ani_threadwork_t *work,
			    ani_thread_t **out_thread)
{
	ani_error_t err;

	assert_ani_threadpool_valid(tpool);

	ani_threadpool_lock(tpool);
	err = ani_threadpool_thread_dowork(tpool, work, 0, 0, out_thread);
	ani_threadpool_unlock(tpool);

	return err;
}

/*
 * Start a thread from the given pool to do the indcated work and wait
 * til it terminates or the timeout expires.  If the timeout expires,
 * the thread is killed and a distinguished error code is returned.
 * The threadwork recrod may be reused upon return from this call.
 * Returns zero on success, an error code otherwise.
 *
 * XXX implement timeout
 */
ani_error_t
ani_threadpool_thread_run(ani_threadpool_t *tpool, ani_threadwork_t *work,
			  an_timespec_t *timo)
{
	ani_thread_t *thread;
	ani_error_t err;

	assert_ani_threadpool_valid(tpool);

	ani_threadpool_lock(tpool);

	/*
	 * Fire off the thread
	 */
	err = ani_threadpool_thread_dowork(tpool, work, 0, 0, &thread);
	if (err) {
		ani_threadpool_unlock(tpool);
		return err;
	}

	/*
	 * Wait for the thread to finish.
	 */
	err = ani_threadpool_thread_wait(tpool, thread, timo);

	ani_threadpool_unlock(tpool);
	return ani_error_clear();
}

/*
 * Schedule work to be done in a thread from the indicated pool.
 * If no threads are currently available or queueonly is non-zero, the
 * request will be queued.  Returns zero on success, an error code otherwise.
 *
 * XXX odd semantics: work record must remain valid after return,
 * until a thread is started.
 */
ani_error_t
ani_threadpool_thread_schedule(ani_threadpool_t *tpool, ani_threadwork_t *work,
			       int queueonly)
{
	ani_error_t err;

	assert_ani_threadpool_valid(tpool);

	ani_threadpool_lock(tpool);
	err = ani_threadpool_thread_dowork(tpool, work, 1, queueonly, 0);
	ani_threadpool_unlock(tpool);

	return err;
}

/*
 * Start the given thread as a ``one-shot'' thread in the context of the
 * given pool running the indicated function.  Returns zero if a thread was
 * successfully started, an error otherwise.
 */
ani_error_t
ani_threadpool_osthread_start(ani_threadpool_t *tpool, ani_thread_t *thread,
			      int inkern, ani_thread_func_t func,
			      void *farg1, void *farg2, void *farg3)
{
	ani_error_t err;

	assert_ani_threadpool_valid(tpool);
	/* XXX not a valid thread yet: assert_ani_thread_valid(thread); */

	ani_threadpool_lock(tpool);

	if ((tpool->flags & TPOOL_DYING) == 0) {
		thread->flags |= THREAD_ONESHOT;
		ani_threadwork_set(&thread->work, inkern, 0,
				   func, farg1, farg2, farg3);
		err = ani_thread_start(thread, tpool);
	} else
		err = ani_error_set(EINVAL, "invalid threadpool");

	ani_threadpool_unlock(tpool);
	return err;
}

/*
 * Lightweight version of thread_schedule called at interrupt level.
 * Puts the given work record at the end of the pool work queue and does
 * a wakeup on any available thread.
 */
void
ani_threadpool_thread_startintr(ani_threadpool_t *tpool, ani_threadwork_t *work)
{
	assert_ani_threadpool_valid(tpool);

	/*
	 * Queue it up
	 */
	if (tpool->worktail == 0) {
		assert(tpool->workhead == 0);
		tpool->workhead = tpool->worktail = work;
	} else {
		assert(tpool->workhead != 0);
		tpool->worktail->next = work;
		tpool->worktail = work;
	}
	tpool->worklen++;

	TPOOLSTATS({
		if (tpool->worklen > tpool->stats.maxqueued)
			tpool->stats.maxqueued = tpool->worklen;
		tpool->stats.queued++;
	});

	/*
	 * Wake any available thread
	 */
	if (tpool->freethreads != 0) {
		ani_thread_t *thread;

		assert(tpool->freelist != 0);
		thread = tpool->freelist;
		tpool->freelist = thread->freenext;
		tpool->freethreads--;
		ani_icond_signal(&thread->waiting);
	}
}

/*
 * Start a thread in the given pool to perform the indicated work.
 * If canqueue is non-zero then, if no threads are available, the record
 * is put on the threadpool workqueue to be handled later.  If onlyqueue
 * is non-zero then the work is always queued and no thread start is
 * attempted.  If out_thread is non-zero, the thread chosen to process
 * the request is returned.
 */
static ani_error_t
ani_threadpool_thread_dowork(ani_threadpool_t *tpool, ani_threadwork_t *work,
			     int canqueue, int onlyqueue,
			     ani_thread_t **out_thread)
{
	ani_thread_t *thread;
	ani_error_t err;

	assert(work != 0);
	assert(work->func != 0);
	assert(ani_threadpool_islocked(tpool));

	if ((tpool->flags & TPOOL_DYING) != 0)
		return ani_error_set(EINVAL, "invalid threadpool");

	DBG(DBG_THREADS, {
		ani_dprintf("TP%p: dowork: scheduling call to %p(%p)\n",
			    tpool, work->func, work->farg);
	});
	TPOOLSTATS(tpool->stats.starts++);

	err = ani_error_clear();
	if (!onlyqueue) {
		/*
		 * If there is a free pool thread available now,
		 * give it the work record and wake it up.
		 */
		ani_os_intr_disable();
		if (tpool->freethreads != 0) {
			assert(tpool->freelist != 0);
			thread = tpool->freelist;
			tpool->freelist = thread->freenext;
			tpool->freethreads--;
			ani_os_intr_enable();
			thread->work = *work;
			ani_icond_signal(&thread->waiting);
			if (out_thread)
				*out_thread = thread;
			DBG(DBG_THREADS|DBG_THREADPOOLS, {
				ani_dprintf("TP%p: dowork: woke cached T%p\n",
					    tpool, thread);
			});
			return ani_error_clear();
		}
		ani_os_intr_enable();
		
		/*
		 * If we are not at the thread limit for the pool,
		 * start up a new thread and give it the work.
		 */
		err = ani_thread_spawn(tpool, &thread);
		if (!ani_error_occured(err)) {
			thread->work = *work;
			if (out_thread)
				*out_thread = thread;
			DBG(DBG_THREADS|DBG_THREADPOOLS, {
				ani_dprintf("TP%p: dowork: spawned T%p\n",
					    tpool, thread);
			});
			TPOOLSTATS(tpool->stats.creates++);

			return ani_error_clear();
		}
	}

	/*
	 * No threads are available.  If it is ok to queue the request,
	 * put it on the end of the work list for the pool.  Otherwise
	 * we return an error.
	 */
	if (canqueue) {
		ani_os_intr_disable();
		if (tpool->worktail == 0) {
			assert(tpool->workhead == 0);
			tpool->workhead = tpool->worktail = work;
		} else {
			assert(tpool->workhead != 0);
			tpool->worktail->next = work;
			tpool->worktail = work;
		}
		tpool->worklen++;

		TPOOLSTATS({
			if (tpool->worklen > tpool->stats.maxqueued)
				tpool->stats.maxqueued = tpool->worklen;
			tpool->stats.queued++;
		});
		DBG(DBG_THREADPOOLS, {
			ani_dprintf("TP%p: dowork: queued work\n", tpool);
		});
		ani_os_intr_enable();

		if (out_thread)
			*out_thread = 0;

		return ani_error_clear();
	}

	DBG(DBG_THREADPOOLS, {
		ani_dprintf("TP%p: dowork: failed\n", tpool);
	});
	TPOOLSTATS(tpool->stats.failed++);

	return err;
}

/*
 * Ensure there are the indicated number of threads in a pool's freelist now
 * and make that value the target freelist size.  Pool should be locked on
 * entry and exit.
 */
ani_error_t
ani_threadpool_thread_setreserve(ani_threadpool_t *tpool, int nthreads,
				 int waitfor)
{
	int delta;
	ani_error_t err;
	ani_thread_t *thread;

	assert(ani_threadpool_islocked(tpool));

	delta = nthreads - tpool->freethreads;

	DBG(DBG_THREADPOOLS, {
		ani_dprintf("TP%p: setreserve %d (%+d)\n",
			    tpool, nthreads, delta);
	});
	
	/*
	 * Not enough free threads currently, make more.
	 *
	 * XXX note that we might fail after creating some threads.
	 * In this case we return an error without updating the reserve count.
	 * The extra threads we did spawn will die when they start up since
	 * we haven't updated the reserve.
	 */
	if (delta > 0) {
		while (delta > 0) {
			err = ani_thread_spawn(tpool, 0);
			DBG(DBG_THREADPOOLS, {
				ani_dprintf("TP%p: setreserve: "
					    "create thread returns err=%d\n",
					    tpool, err);
			});
			if (ani_error_occured(err))
				return err;
			delta--;
		}
		tpool->reserve = nthreads;

		/*
		 * Wait for the thread count to hit the desired value.
		 * The threadpool had better be idle or this wait might
		 * not return.
		 */
		if (waitfor) {
			DBG(DBG_THREADPOOLS, {
				ani_dprintf("TP%p: setreserve: "
					    "waiting for %d new threads\n",
					    tpool, nthreads - tpool->freethreads);
			});
			while (tpool->freethreads < tpool->reserve) {
				tpool->flags |= TPOOL_WAITING;
				ani_icond_wait(&tpool->waiting, &tpool->lock);
			}
			DBG(DBG_THREADPOOLS, {
				ani_dprintf("TP%p: setreserve: wait done\n",
					    tpool);
			});
		}
	}

	/*
	 * Too many threads, pull off the appropriate number
	 * and wake them up so they will exit.
	 */
	if (delta < 0) {
		while (delta < 0) {
			ani_os_intr_disable();
			assert(tpool->freelist != 0);
			thread = tpool->freelist;
			tpool->freelist = thread->freenext;
			tpool->freethreads--;
			ani_os_intr_enable();
			tpool->freeexitcount++;
			assert(thread->work.func == 0);
			ani_icond_signal(&thread->waiting);
			delta++;
			DBG(DBG_THREADPOOLS, {
				ani_dprintf("TP%p: setreserve: wake thread %p\n",
					    tpool, thread);
			});
		}
		tpool->reserve = nthreads;

		/*
		 * Wait for the thread count to hit the desired value.
		 */
		if (waitfor) {
			DBG(DBG_THREADPOOLS, {
				ani_dprintf("TP%p: setreserve: "
					    "waiting for %d threads to die\n",
					    tpool, tpool->freeexitcount);
			});
			while (tpool->freeexitcount > 0) {
				tpool->flags |= TPOOL_WAITING;
				ani_icond_wait(&tpool->waiting, &tpool->lock);
			}
			DBG(DBG_THREADPOOLS, {
				ani_dprintf("TP%p: setreserve: wait done\n",
					    tpool);
			});
		}
	}

	return ani_error_clear();
}

/*
 * Called from the explicit thread_exit API function.
 * Return the thread to the threadpool reserve.
 */
void
ani_threadpool_thread_exit(ani_threadpool_t *tpool)
{
	ani_thread_t *thread = ani_thread_current();
	assert(tpool == thread->pool);

	siglongjmp(thread->jmpbuf, 1);
}

void
ani_threadpool_stats(ani_threadpool_t *tpool, ani_threadstats_t *stats)
{
	ani_thread_t *th;

	assert_ani_threadpool_valid(tpool);
	assert((tpool->flags & TPOOL_DYING) == 0);

	ani_threadpool_lock(tpool);

#ifdef ANI_THREADPOOL_STATS
	/*
	 * Update pool stats
	 */
	tpool->stats.curcpu.amount = 0;
	for (th = tpool->threads; th != 0; th = th->next) {
		ani_threadcpu_t tmp;
		ani_os_sched_getcpu(th, &tmp);
		tpool->stats.curcpu.units = tmp.units;
		tpool->stats.curcpu.amount += tmp.amount;
	}
	for (th = tpool->osthreads; th != 0; th = th->next) {
		ani_threadcpu_t tmp;
		ani_os_sched_getcpu(th, &tmp);
		tpool->stats.curcpu.units = tmp.units;
		tpool->stats.curcpu.amount += tmp.amount;
	}
	tpool->stats.active =
		(tpool->curthreads + tpool->curosthreads) - tpool->freethreads;
#endif

	memcpy(stats, &tpool->stats, sizeof *stats);
	ani_threadpool_unlock(tpool);
}

#ifdef ANI_THREADPOOL_STATS
static void
ani_threadpool_printstats(ani_threadpool_t *tpool)
{
	int cached;

	cached = tpool->stats.starts -
		(tpool->stats.creates+tpool->stats.queued+tpool->stats.failed);
	ani_dprintf("TP%p: %d thread starts\n", tpool, tpool->stats.starts);
	ani_dprintf("\t(%d cached, %d created, %d queued, %d failed)\n",
		    cached, tpool->stats.creates, tpool->stats.queued,
		    tpool->stats.failed);
	ani_dprintf("  Max: %d active threads, %d free threads, %d queued\n",
		    tpool->stats.maxactive, tpool->stats.maxcached,
		    tpool->stats.maxqueued);
}
#endif
